#include "ColorSpline.h"

#include <math.h>
#include "System.h"

/* Coefficients of basis matrix. */

#define CR00     -0.5
#define CR01      1.5
#define CR02     -1.5
#define CR03      0.5
#define CR10      1.0
#define CR11     -2.5
#define CR12      2.0
#define CR13     -0.5
#define CR20     -0.5
#define CR21      0.0
#define CR22      0.5
#define CR23      0.0
#define CR30      0.0
#define CR31      1.0
#define CR32      0.0
#define CR33      0.0

float Max(float a, float b)
{
	return a >= b ? a : b;
} /* Max */

float Min(float a, float b)
{
	return a <= b ? a : b;
} /* Min */

float Mod(float a, float b)
{
	return a - b * floor(a / b); /* fmodf() returns values < 0 */
}  /* Mod */

float Clamp(float n, float min, float max)
{
	if (n <= min)
		return min;
	if (n >= max)
		return max;
	return n;
}  /* Clamp */

float BoxStep(float a, float b, float x)
{
	if (x < a)
		return 0.0;
	if (x > b)
		return 1.0;
	return (x - a) / (b - a);
}  /* BoxStep */

float SmoothStep(float a, float b, float x)
{
	if (x < a)
		return 0.0;
	if (x > b)
		return 1.0;
	x = (x - a) / (b - a);	/* normalize to [0:1] */
	return (x * x * (3.0 - (x + x)));
}  /* SmoothStep */

float ValueSpline(float x, long nKnots, float *knot)
{
	long span;
	long nSpans;
	float c0, c1, c2, c3;	/* Coefficients of the cubic */

	if (nKnots < 1)			/* Less than one knot is illegal */
		return 0.0;

	if (nKnots == 1)		/* Only one knot is a constant value */
		return knot[0];

	if (nKnots == 2)		/* Two knots are linearly interpolated */
		return knot[0] + x * (knot[1] - knot[0]);

	if (nKnots == 3) {		/* Three knots uses a smooth interpolation */
		float xg = x * x;
		float x0 = 0.5 - x + xg * 0.5;
		float x1 = 0.5 + x - xg;
		float x2 = 1.0 - x0 - x1;
		
		return x0 * knot[0] + x1 * knot[1] + x2 * knot[2];
	}

	/* Four or more knots uses a cubic spline */
	
	nSpans = nKnots - 3;

	/* Find the appropriate 4-point span of the spline */

	if (x <= 0.0)
		x = 0.0;
	else if (x >= 1.0)
		x = nSpans;
	else
		x *= nSpans;
		
	span = x;
	x -= span;
	knot += span;

	/* Evaluate the span cubic at x using Horner's rule */

	c3 = CR00 * knot[0] + CR01 * knot[1] + CR02 * knot[2] + CR03 * knot[3];
	c2 = CR10 * knot[0] + CR11 * knot[1] + CR12 * knot[2] + CR13 * knot[3];
	c1 = CR20 * knot[0] + CR21 * knot[1] + CR22 * knot[2] + CR23 * knot[3];
	c0 = CR30 * knot[0] + CR31 * knot[1] + CR32 * knot[2] + CR33 * knot[3];

	return ((c3 * x + c2) * x + c1) * x + c0;
}  /* ValueSpline */

void ColorSpline(float x, long nKnots, ARGB_Real *knot, ARGB_Real *result)
{
	long span;
	long nSpans;
	float c0, c1, c2, c3;	/* Coefficients of the cubic */
	
	if (nKnots < 1) {		/* Less than one knot is illegal */
		result->a = 0.0;
		result->r = 0.0;
		result->g = 0.0;
		result->b = 0.0;
		return;
	}

	if (nKnots == 1) {		/* Only one knot is a constant value */
		*result = knot[0];
		return;
	}

	if (nKnots == 2) {		/* Two knots are linearly interpolated */
		result->a = knot[0].a + x * (knot[1].a - knot[0].a);
		result->r = knot[0].r + x * (knot[1].r - knot[0].r);
		result->g = knot[0].g + x * (knot[1].g - knot[0].g);
		result->b = knot[0].b + x * (knot[1].b - knot[0].b);
		return;
	}

	if (nKnots == 3) {		/* Three knots uses a smooth interpolation */
		float xg = x * x;
		float x0 = 0.5 - x + xg * 0.5;
		float x1 = 0.5 + x - xg;
		float x2 = 1.0 - x0 - x1;
		
		result->a = x0 * knot[0].a + x1 * knot[1].a + x2 * knot[2].a;
		result->r = x0 * knot[0].r + x1 * knot[1].r + x2 * knot[2].r;
		result->g = x0 * knot[0].g + x1 * knot[1].g + x2 * knot[2].g;
		result->b = x0 * knot[0].b + x1 * knot[1].b + x2 * knot[2].b;
		return;
	}

	/* Four or more knots uses a cubic spline */
	
	nSpans = nKnots - 3;

	/* Find the appropriate 4-point span of the spline */

	if (x <= 0.0) {
		x = 0.0;
		span = 0;
	} else if (x >= 1.0) {
		x = 1.0;
		span = nSpans - 1;
	} else {
		x *= nSpans;
		span = x;
		x -= span;
	}
	
	/* Get the first knot of the span */
	
	knot += span;

	/* Evaluate the span cubic at x using Horner's rule */

	c3 = CR00 * knot[0].a + CR01 * knot[1].a + CR02 * knot[2].a + CR03 * knot[3].a;
	c2 = CR10 * knot[0].a + CR11 * knot[1].a + CR12 * knot[2].a + CR13 * knot[3].a;
	c1 = CR20 * knot[0].a + CR21 * knot[1].a + CR22 * knot[2].a + CR23 * knot[3].a;
	c0 = CR30 * knot[0].a + CR31 * knot[1].a + CR32 * knot[2].a + CR33 * knot[3].a;
	result->a = ((c3 * x + c2) * x + c1) * x + c0;

	c3 = CR00 * knot[0].r + CR01 * knot[1].r + CR02 * knot[2].r + CR03 * knot[3].r;
	c2 = CR10 * knot[0].r + CR11 * knot[1].r + CR12 * knot[2].r + CR13 * knot[3].r;
	c1 = CR20 * knot[0].r + CR21 * knot[1].r + CR22 * knot[2].r + CR23 * knot[3].r;
	c0 = CR30 * knot[0].r + CR31 * knot[1].r + CR32 * knot[2].r + CR33 * knot[3].r;
	result->r = ((c3 * x + c2) * x + c1) * x + c0;

	c3 = CR00 * knot[0].g + CR01 * knot[1].g + CR02 * knot[2].g + CR03 * knot[3].g;
	c2 = CR10 * knot[0].g + CR11 * knot[1].g + CR12 * knot[2].g + CR13 * knot[3].g;
	c1 = CR20 * knot[0].g + CR21 * knot[1].g + CR22 * knot[2].g + CR23 * knot[3].g;
	c0 = CR30 * knot[0].g + CR31 * knot[1].g + CR32 * knot[2].g + CR33 * knot[3].g;
	result->g = ((c3 * x + c2) * x + c1) * x + c0;

	c3 = CR00 * knot[0].b + CR01 * knot[1].b + CR02 * knot[2].b + CR03 * knot[3].b;
	c2 = CR10 * knot[0].b + CR11 * knot[1].b + CR12 * knot[2].b + CR13 * knot[3].b;
	c1 = CR20 * knot[0].b + CR21 * knot[1].b + CR22 * knot[2].b + CR23 * knot[3].b;
	c0 = CR30 * knot[0].b + CR31 * knot[1].b + CR32 * knot[2].b + CR33 * knot[3].b;
	result->b = ((c3 * x + c2) * x + c1) * x + c0;
}  /* ColorSpline */

void MixRGBColors(ARGB_Real *srcColor1, ARGB_Real *srcColor2, ARGB_Real *dstColor, float mix)
{
	if (mix <= 0.0)
		*dstColor = *srcColor1;
	else if (mix >= 1.0)
		*dstColor = *srcColor2;
	else {
		dstColor->r = MIX(srcColor1->r, srcColor2->r, mix);
		dstColor->g = MIX(srcColor1->g, srcColor2->g, mix);
		dstColor->b = MIX(srcColor1->b, srcColor2->b, mix);
	}
}  /* MixRGBColors */

void MixARGBColors(ARGB_Real *srcColor1, ARGB_Real *srcColor2, ARGB_Real *dstColor, float mix)
{
	if (mix <= 0.0)
		*dstColor = *srcColor1;
	else if (mix >= 1.0)
		*dstColor = *srcColor2;
	else if (srcColor1->a == 0.0) {
		dstColor->a = srcColor2->a * mix;
		dstColor->r = srcColor2->r;
		dstColor->g = srcColor2->g;
		dstColor->b = srcColor2->b;
	} else if (srcColor2->a == 0.0) {
		dstColor->a = srcColor1->a * (1.0 - mix);
		dstColor->r = srcColor1->r;
		dstColor->g = srcColor1->g;
		dstColor->b = srcColor1->b;
	} else {
		dstColor->a = MIX(srcColor1->a, srcColor2->a, mix);
		dstColor->r = MIX(srcColor1->r, srcColor2->r, mix);
		dstColor->g = MIX(srcColor1->g, srcColor2->g, mix);
		dstColor->b = MIX(srcColor1->b, srcColor2->b, mix);
	}
}  /* MixARGBColors */

void BlendRGBColors(ARGB_Real *srcColor1, ARGB_Real *srcColor2, ARGB_Real *dstColor, float mix)
{
	if (mix <= 0.0)
		*dstColor = *srcColor1;
	else if (mix >= 1.0)
		*dstColor = *srcColor2;
	else {
		float v1 = MAX3(srcColor1->r, srcColor1->g, srcColor1->b);
		float v2 = MAX3(srcColor2->r, srcColor2->g, srcColor2->b);
		float v = MIX(v1, v2, mix);
		float r = MIX(srcColor1->r, srcColor2->r, mix);
		float g = MIX(srcColor1->g, srcColor2->g, mix);
		float b = MIX(srcColor1->b, srcColor2->b, mix);
		float max = MAX3(r, g, b);
		
		if (max < v) {
			v /= max;
			dstColor->r = r * v;
			dstColor->g = g * v;
			dstColor->b = b * v;
		} else {
			dstColor->r = r;
			dstColor->g = g;
			dstColor->b = b;
		}
	}
}  /* BlendRGBColors */

void BlendARGBColors(ARGB_Real *srcColor1, ARGB_Real *srcColor2, ARGB_Real *dstColor, float mix)
{
	if (mix <= 0.0)
		*dstColor = *srcColor1;
	else if (mix >= 1.0)
		*dstColor = *srcColor2;
	else if (srcColor1->a == 0.0) {
		dstColor->a = srcColor2->a * mix;
		dstColor->r = srcColor2->r;
		dstColor->g = srcColor2->g;
		dstColor->b = srcColor2->b;
	} else if (srcColor2->a == 0.0) {
		dstColor->a = srcColor1->a * (1.0 - mix);
		dstColor->r = srcColor1->r;
		dstColor->g = srcColor1->g;
		dstColor->b = srcColor1->b;
	} else {
		float v1 = MAX3(srcColor1->r, srcColor1->g, srcColor1->b);
		float v2 = MAX3(srcColor2->r, srcColor2->g, srcColor2->b);
		float v = MIX(v1, v2, mix);
		float r = MIX(srcColor1->r, srcColor2->r, mix);
		float g = MIX(srcColor1->g, srcColor2->g, mix);
		float b = MIX(srcColor1->b, srcColor2->b, mix);
		float max = MAX3(r, g, b);
		
		if (max < v) {
			v /= max;
			dstColor->r = r * v;
			dstColor->g = g * v;
			dstColor->b = b * v;
		} else {
			dstColor->r = r;
			dstColor->g = g;
			dstColor->b = b;
		}
		dstColor->a = MIX(srcColor1->a, srcColor2->a, mix);
	}
}  /* BlendARGBColors */

void ApplyARGBColor(RGB_Real *srcColor1, ARGB_Real *srcColor2, RGB_Real *dstColor)
{
	float mix = srcColor2->a;
	
	if (mix <= 0.0)
		*dstColor = *srcColor1;
	else if (mix >= 1.0) {
		dstColor->r = srcColor2->r;
		dstColor->g = srcColor2->g;
		dstColor->b = srcColor2->b;
	} else {
		dstColor->r = MIX(srcColor1->r, srcColor2->r, mix);
		dstColor->g = MIX(srcColor1->g, srcColor2->g, mix);
		dstColor->b = MIX(srcColor1->b, srcColor2->b, mix);
	}
}  /* ApplyARGBColor */
