/*
 *		Copyright (c)1991-1997 Electric Image, Inc.
 *		This code is confidential and is a trade secret of Electric Image, Inc.
 */

#include "ColorSpline.h"
#include "EIShader.h"
#include "FPU.h"
#include "Matrix.h"
#include "Noise.h"
#include "PString.h"
#include <stddef.h>
#include "System.h"

#include "EI_Dialog.h"
#include "EI_Control.h"
#include "PluginUtils.h"
#include "EI_FileRef.h"
#include "EI_Stream.h"

#define MIX(a, b, f) ((a) + (f) * ((b) - (a)))

#define shaderType				'BNOI'
#define shaderVersion			0x00000001
#define shaderRevVersion		0x01000000

#define defShaderDensity		20.0
#define defShaderBumpStrength	1.0
#define defShaderColorThreshold	0.5
#define defShaderColorRange		1.0
#define defShaderBumpColorA		255
#define defShaderBumpColorR		0
#define defShaderBumpColorG		0
#define defShaderBumpColorB		0
#define defShaderSurfaceColorA	255
#define defShaderSurfaceColorR	255
#define defShaderSurfaceColorG	255
#define defShaderSurfaceColorB	255

typedef struct {
	unsigned long shaderOwnerType;
	unsigned long shaderVersionNum;
	float shaderDensity;
	float shaderBumpStrength;
	float shaderColorThreshold;
	float shaderColorRange;
	ARGB shaderBumpColor;
	ARGB shaderSurfaceColor;
} ShaderInterfaceRec, *ShaderInterfacePtr;

typedef struct {
	float shaderDensity;
	float shaderBumpStrength;
	float shaderColorThreshold;
	float shaderColorRange;
	ARGB_Real shaderBumpColor;
	ARGB_Real shaderSurfaceColor;
} ShaderDataRec, *ShaderDataPtr;

/* Static functions */

static void DefaultShaderInterface(ShaderInterfacePtr theShaderInterface)
{
	theShaderInterface->shaderOwnerType = shaderType;
	theShaderInterface->shaderVersionNum = shaderVersion;
	theShaderInterface->shaderDensity = defShaderDensity;
	theShaderInterface->shaderBumpStrength = defShaderBumpStrength;
	theShaderInterface->shaderColorThreshold = defShaderColorThreshold;
	theShaderInterface->shaderColorRange = defShaderColorRange;
	theShaderInterface->shaderBumpColor.argb.a = defShaderBumpColorA;
	theShaderInterface->shaderBumpColor.argb.r = defShaderBumpColorR;
	theShaderInterface->shaderBumpColor.argb.g = defShaderBumpColorG;
	theShaderInterface->shaderBumpColor.argb.b = defShaderBumpColorB;
	theShaderInterface->shaderSurfaceColor.argb.a = defShaderSurfaceColorA;
	theShaderInterface->shaderSurfaceColor.argb.r = defShaderSurfaceColorR;
	theShaderInterface->shaderSurfaceColor.argb.g = defShaderSurfaceColorG;
	theShaderInterface->shaderSurfaceColor.argb.b = defShaderSurfaceColorB;
}  /* DefaultShaderInterface */

static void FixShaderInterface(ShaderInterfacePtr theShaderInterface)
{
	theShaderInterface;
	
	FIXFIELD(theShaderInterface->shaderOwnerType);
	FIXFIELD(theShaderInterface->shaderVersionNum);
	FIXFIELD(theShaderInterface->shaderDensity);
	FIXFIELD(theShaderInterface->shaderBumpStrength);
	FIXFIELD(theShaderInterface->shaderColorThreshold);
	FIXFIELD(theShaderInterface->shaderColorRange);
	
}  /* FixShaderInterface */

static void GetShaderInterface(void* theInterface, long theSize, ShaderInterfacePtr theShaderInterface)
{
	/* Copy the shader interface data */
	
	if (theSize == sizeof(ShaderInterfaceRec))
		*theShaderInterface = *(ShaderInterfacePtr)theInterface;
	else {
		theShaderInterface->shaderOwnerType = 0;
		theShaderInterface->shaderVersionNum = 0;
	}
	
	/* Validate the shader interface data */
	FixShaderInterface(theShaderInterface);
	if (theShaderInterface->shaderOwnerType != shaderType ||
		theShaderInterface->shaderVersionNum != shaderVersion)
		DefaultShaderInterface(theShaderInterface);
}  /* GetShaderInterface */

/* Export functions */

long EIShaderInformation(EIShaderInformationRec *theInformationPtr)
{
	ShaderInterfaceRec theShaderInterface;
		
	/* Get the shader interface parameters */
		
	GetShaderInterface(theInformationPtr->shaderInterface, theInformationPtr->shaderInterfaceSize, &theShaderInterface);
	
	/* Setup the shader feature flags */
	
	BSET(theInformationPtr->shaderFeatureFlags, shaderAntialiasFlag);
	BSET(theInformationPtr->shaderFeatureFlags, shaderColorFlag);
	SETFLAG(theInformationPtr->shaderFeatureFlags, shaderBumpFlag, (theShaderInterface.shaderBumpStrength != 0.0));
	return shaderNoError;
}  /* EIShaderInformation */

long EIShaderInitialize(EIShaderInitializeRec *theInitializePtr)
{
	ShaderDataPtr theShaderData;
	
	/* Initialize the noise library */
	
	if (!InitNoise((NoiseMallocFunction)theInitializePtr->HostMalloc, NULL))
		return shaderMemoryError;
	
	/* Allocate the shader's private data record */
	
	theShaderData = (ShaderDataPtr)theInitializePtr->HostMalloc(sizeof(ShaderDataRec));
	if (theShaderData) {
		ShaderInterfaceRec theShaderInterface;
		
		/* Get the shader interface parameters */
		
		GetShaderInterface(theInitializePtr->shaderInterface, theInitializePtr->shaderInterfaceSize, &theShaderInterface);
				
		/* Setup the shader's private data */
		
		theShaderData->shaderDensity = theShaderInterface.shaderDensity;
		theShaderData->shaderBumpStrength = theShaderInterface.shaderBumpStrength;
		theShaderData->shaderColorThreshold = theShaderInterface.shaderColorThreshold;
		theShaderData->shaderColorRange = theShaderInterface.shaderColorRange == 0.0 ? 0.0 : 1.0 / theShaderInterface.shaderColorRange;
		theShaderData->shaderBumpColor.a = theShaderInterface.shaderBumpColor.argb.a / 255.0;
		theShaderData->shaderBumpColor.r = theShaderInterface.shaderBumpColor.argb.r / 255.0;
		theShaderData->shaderBumpColor.g = theShaderInterface.shaderBumpColor.argb.g / 255.0;
		theShaderData->shaderBumpColor.b = theShaderInterface.shaderBumpColor.argb.b / 255.0;
		theShaderData->shaderSurfaceColor.a = theShaderInterface.shaderSurfaceColor.argb.a / 255.0;
		theShaderData->shaderSurfaceColor.r = theShaderInterface.shaderSurfaceColor.argb.r / 255.0;
		theShaderData->shaderSurfaceColor.g = theShaderInterface.shaderSurfaceColor.argb.g / 255.0;
		theShaderData->shaderSurfaceColor.b = theShaderInterface.shaderSurfaceColor.argb.b / 255.0;
		
		/* Return the pointer to the shader's private data to the host */
		
		theInitializePtr->shaderData = theShaderData;
		return shaderNoError;
	} else
		return shaderMemoryError;
}  /* EIShaderInitialize */

long EIShaderFinish(EIShaderFinishRec *theFinishPtr)
{
	/* Get rid of the shader's private data */
	
	theFinishPtr->HostFree(theFinishPtr->shaderData);
	
	/* Finish the noise library */
	
	FinishNoise((NoiseFreeFunction)theFinishPtr->HostFree);
	return shaderNoError;
}  /* EIShaderFinish */

long EIShaderShade(EIShaderShadeRec *theShadePtr)
{
	EIShaderMapPtr theMap = theShadePtr->shaderMap;
	ShaderDataPtr theShaderData = (ShaderDataPtr)theShadePtr->shaderData;
	float factor = theShaderData->shaderDensity;
	
	if (BTST(theShadePtr->shaderFeatureFlags, shaderBumpFlag)) {
		factor = DNoise2(theMap->map_X0 * factor, theMap->map_Y0 * factor, theMap->map_Z0 * factor, &theShadePtr->shaderBumpA, &theShadePtr->shaderBumpB, &theShadePtr->shaderBumpC);
		theShadePtr->shaderBumpA *= theShaderData->shaderBumpStrength;
		theShadePtr->shaderBumpB *= theShaderData->shaderBumpStrength;
		theShadePtr->shaderBumpC *= theShaderData->shaderBumpStrength;
	} else 
		factor = Noise(theMap->map_X0 * factor, theMap->map_Y0 * factor, theMap->map_Z0 * factor);
	
	/* Calculate the shader color */
	
	if (BTST(theShadePtr->shaderFeatureFlags, shaderColorFlag)) {
		if (theShaderData->shaderColorRange == 0.0)
			factor = factor >= theShaderData->shaderColorThreshold ? 1.0 : 0.0;
		else {
			factor = 0.5 + (factor - theShaderData->shaderColorThreshold) * theShaderData->shaderColorRange;
			if (factor < 0.0)
				factor = 0.0;
			else if (factor > 1.0)
				factor = 1.0;
		}
		MixARGBColors(&theShaderData->shaderSurfaceColor, &theShaderData->shaderBumpColor, &theShadePtr->shaderColor, factor);
	}
	return shaderNoError;
}  /* EIShaderShade */


/*--------------- Custom shader's interface support -------------------- */

#define ERROR_DLG_ID		21001
#define QUEST_DLG_ID		21002

#define ID_DENSITY_ET		4
#define ID_STRENGTH_ET		5
#define ID_THRESHOLD_ET		6
#define ID_RANGE_ET			7
#define ID_COLOR1_RGB		8
#define ID_COLOR2_RGB		9
#define ID_LOAD_BT		    10
#define ID_SAVE_BT		    11
#define ID_DEFAULT_BT	    12
#define ID_DENSITY_SL		13
#define ID_STRENGTH_SL		14
#define ID_THRESHOLD_SL		15
#define ID_RANGE_SL			16

#define SLIDER_MAXVAL		100

#define DENSITY_MIN			0.0f
#define DENSITY_MAX			100.0f
#define STRENGTH_MIN		-2.0f
#define STRENGTH_MAX		2.0f
#define THRESHOLD_MIN		0.0f
#define THRESHOLD_MAX		1.0f
#define RANGE_MIN			0.0f
#define RANGE_MAX			1.0f


typedef struct {
	EI_Dialog * theDialog;
	EI_ResourceFile * theResource;
	EI_ResourceFile theResourceFile;
	EIHostInterfaceSetupProc SetupProc;
} InterfaceSetupRec, *InterfaceSetupPtr;

static InterfaceSetupRec theSetupRec;

/* Get control by ID */

static EI_Control * ID2Ctl( long ID )
{
	return EI_FindControlByID(theSetupRec.theDialog, ID);
}

/* Tell host to update item */

static void UpdateItem( long ID )
{
	theSetupRec.SetupProc(theSetupRec.theDialog, hostSetShaderItem, ID, 0);
}

/* Show error message */

static void ShowError( unsigned char * msg )
{
	StandardDialog(theSetupRec.theResource, ERROR_DLG_ID, msg, 0, 0, 0);
}

/* Show yes/no dialog */

static long ShowYesNo( unsigned char * msg )
{
	return StandardDialog(theSetupRec.theResource, QUEST_DLG_ID, msg, 0, 0, 0);
}

/* Set float value and update item */

static void SetFloatVal( long ID, float val )
{
	SetTheIReal(theSetupRec.theDialog, ID, val, 2);
	theSetupRec.SetupProc(theSetupRec.theDialog, hostSetShaderItem, ID, 0);
}

/* Set color value and update item */

static void SetARGBVal( long ID, short a, short r, short g, short b )
{
	EI_SetColorButtonColor(ID2Ctl(ID), r, g, b, a);
	theSetupRec.SetupProc(theSetupRec.theDialog, hostSetShaderItem, ID, 0);
}

/* Synchronize slider if edit box has been changed */

static void Edit2Slider( long editID, long sliderID, float minVal, float maxVal )
{
	double val = GetTheIReal(theSetupRec.theDialog, editID);
	if (val > maxVal) maxVal = val; 
	if (val < minVal) minVal = val; 
	EI_SetControlValue(ID2Ctl(sliderID), (val - minVal) / (maxVal - minVal) * SLIDER_MAXVAL);
} 

/* Set edit box value if user drags slider */

static void Slider2Edit( long sliderID, long editID, float minVal, float maxVal )
{
	double val = EI_GetControlValue(ID2Ctl(sliderID));
	val = val / SLIDER_MAXVAL * (maxVal - minVal) + minVal;
	
	SetTheIReal(theSetupRec.theDialog, editID, val, 2);
	UpdateItem(editID);
} 

/* Setup edit/slider combo */

static void SetupEditSlider( long editID, long sliderID, float minVal, float maxVal )
{
	EI_SetControlMax(ID2Ctl(sliderID), SLIDER_MAXVAL);
	Edit2Slider(editID, sliderID, minVal, maxVal);
	EI_SetControlDrawImmediate(ID2Ctl(editID), 1);
}

/* Synchronize all sliders */

static void	SetupSliders( void )
{
	SetupEditSlider(ID_DENSITY_ET, ID_DENSITY_SL, DENSITY_MIN, DENSITY_MAX);
	SetupEditSlider(ID_STRENGTH_ET, ID_STRENGTH_SL, STRENGTH_MIN, STRENGTH_MAX);
	SetupEditSlider(ID_THRESHOLD_ET, ID_THRESHOLD_SL, THRESHOLD_MIN, THRESHOLD_MAX);
	SetupEditSlider(ID_RANGE_ET, ID_RANGE_SL, RANGE_MIN, RANGE_MAX);
}	

/* Load shader's settings from file */

static void LoadFile( void )
{
	ARGB color[2];
	float fVal[4];
	unsigned long lVal[2];

	int error = 0;
	EI_Stream * strm = 0;
	EI_FileRef * ref = EI_AskUserForExistingFile("Load BumpyNoise settings");
	if (!ref) return;

	strm = EI_OpenFileStream(ref, cEI_ReadOnly);
	EI_DestroyFileRef(ref);
	if (!strm) return;

	if (!error) error = EI_ReadUInt32(strm, lVal + 0);
	if (!error) error = EI_ReadUInt32(strm, lVal + 1);
	if (!error && (lVal[0] != shaderType || lVal[1] != shaderVersion)) {
	 ShowError("\pIt's not a BumpyNoise file");
	 EI_CloseStream(strm);
	 return;
	}
	
	if (!error) error = EI_ReadFloat32(strm, fVal + 0);
	if (!error) error = EI_ReadFloat32(strm, fVal + 1);
	if (!error) error = EI_ReadFloat32(strm, fVal + 2);
	if (!error) error = EI_ReadFloat32(strm, fVal + 3);
	
	if (!error) error = EI_ReadUInt8(strm, &color[0].argb.a);
	if (!error) error = EI_ReadUInt8(strm, &color[0].argb.r);
	if (!error) error = EI_ReadUInt8(strm, &color[0].argb.g);
	if (!error) error = EI_ReadUInt8(strm, &color[0].argb.b);

	if (!error) error = EI_ReadUInt8(strm, &color[1].argb.a);
	if (!error) error = EI_ReadUInt8(strm, &color[1].argb.r);
	if (!error) error = EI_ReadUInt8(strm, &color[1].argb.g);
	if (!error) error = EI_ReadUInt8(strm, &color[1].argb.b);

	EI_CloseStream(strm);

	if (error) ShowError("\pFile read error");
	else {
	 SetFloatVal(ID_DENSITY_ET, fVal[0]);
	 SetFloatVal(ID_STRENGTH_ET, fVal[1]);
	 SetFloatVal(ID_THRESHOLD_ET, fVal[2]);
	 SetFloatVal(ID_RANGE_ET, fVal[3]);
	 SetARGBVal(ID_COLOR1_RGB, color[0].argb.a, color[0].argb.r, color[0].argb.g, color[0].argb.b);
	 SetARGBVal(ID_COLOR2_RGB, color[1].argb.a, color[1].argb.r, color[1].argb.g, color[1].argb.b);
	 SetupSliders();
	}
}

/* Save shader's settings to file */

static void SaveFile( void )
{
	ARGB color;
	int error = 0;
	EI_Stream * strm = 0;
	EI_FileRef * ref = EI_AskUserForNewFile("Save BumpyNoise Settings");
	if (!ref) return;

	strm = EI_OpenFileStream(ref, cEI_WriteOnly);
	EI_DestroyFileRef(ref);
	if (!strm) return;

	if (!error) error = EI_WriteUInt32(strm, shaderType);
	if (!error) error = EI_WriteUInt32(strm, shaderVersion);
	if (!error) error = EI_WriteFloat32(strm, GetTheIReal(theSetupRec.theDialog, ID_DENSITY_ET));
	if (!error) error = EI_WriteFloat32(strm, GetTheIReal(theSetupRec.theDialog, ID_STRENGTH_ET));
	if (!error) error = EI_WriteFloat32(strm, GetTheIReal(theSetupRec.theDialog, ID_THRESHOLD_ET));
	if (!error) error = EI_WriteFloat32(strm, GetTheIReal(theSetupRec.theDialog, ID_RANGE_ET));
	
	GetColorButtonIDARGB(theSetupRec.theDialog, ID_COLOR1_RGB, &color);
	if (!error) error = EI_WriteUInt8(strm, color.argb.a);
	if (!error) error = EI_WriteUInt8(strm, color.argb.r);
	if (!error) error = EI_WriteUInt8(strm, color.argb.g);
	if (!error) error = EI_WriteUInt8(strm, color.argb.b);

	GetColorButtonIDARGB(theSetupRec.theDialog, ID_COLOR2_RGB, &color);
	if (!error) error = EI_WriteUInt8(strm, color.argb.a);
	if (!error) error = EI_WriteUInt8(strm, color.argb.r);
	if (!error) error = EI_WriteUInt8(strm, color.argb.g);
	if (!error) error = EI_WriteUInt8(strm, color.argb.b);
	
	EI_CloseStream(strm);
	if (error) ShowError("\pFile write error");
}

/* Set default shader's settings */

static void SetDefault( void )
{
	if (!ShowYesNo("\pLoad default values ?")) return;
	SetFloatVal(ID_DENSITY_ET, defShaderDensity);
	SetFloatVal(ID_STRENGTH_ET, defShaderBumpStrength);
	SetFloatVal(ID_THRESHOLD_ET, defShaderColorThreshold);
	SetFloatVal(ID_RANGE_ET, defShaderColorRange);
	SetARGBVal(ID_COLOR1_RGB, defShaderBumpColorA, defShaderBumpColorR, defShaderBumpColorG, defShaderBumpColorB);
	SetARGBVal(ID_COLOR2_RGB, defShaderSurfaceColorA, defShaderSurfaceColorR, defShaderSurfaceColorG, defShaderSurfaceColorB);
	SetupSliders();
}

static void ControlHitFunc( EI_Control *iControl, EI_DrawContext * iContext )
{
	(void) iContext;
	switch (EI_GetControlID(iControl)) {
	 case ID_DENSITY_ET:
	  Edit2Slider(ID_DENSITY_ET, ID_DENSITY_SL, 0.0f, DENSITY_MAX);
	  break;
	  
	 case ID_STRENGTH_ET:
	  Edit2Slider(ID_STRENGTH_ET, ID_STRENGTH_SL, 0.0f, STRENGTH_MAX);
	  break;

	 case ID_THRESHOLD_ET:
	  Edit2Slider(ID_THRESHOLD_ET, ID_THRESHOLD_SL, 0.0f, THRESHOLD_MAX);
	  break;

	 case ID_RANGE_ET:
	  Edit2Slider(ID_RANGE_ET, ID_RANGE_SL, 0.0f, RANGE_MAX);
	  break;

	 case ID_LOAD_BT:
	  LoadFile();
	  break;

	 case ID_SAVE_BT:
	  SaveFile();
	  break;
	  
	 case ID_DEFAULT_BT:
	  SetDefault();
	  break;
	  
	 case ID_DENSITY_SL:
	  Slider2Edit(ID_DENSITY_SL, ID_DENSITY_ET, DENSITY_MIN, DENSITY_MAX);
	  break;
	  
	 case ID_STRENGTH_SL:
	  Slider2Edit(ID_STRENGTH_SL, ID_STRENGTH_ET, STRENGTH_MIN, STRENGTH_MAX);
	  break;

	 case ID_THRESHOLD_SL:
	  Slider2Edit(ID_THRESHOLD_SL, ID_THRESHOLD_ET, THRESHOLD_MIN, THRESHOLD_MAX);
	  break;

	 case ID_RANGE_SL:
	  Slider2Edit(ID_RANGE_SL, ID_RANGE_ET, RANGE_MIN, RANGE_MAX);
	  break;
	}
}

static void SetControlHitFunc( long itemNo )
{
	EI_SetControlHitFunction(ID2Ctl(itemNo), ControlHitFunc);
}

long EIShaderInterfaceSetup( EIShaderInterfaceSetupRec * theInterfaceSetup )
{
	/* Save theInterfaceSetup data */
	
	theSetupRec.theDialog = (EI_Dialog *) theInterfaceSetup->hostDialogReference;
	theSetupRec.theResource = &theSetupRec.theResourceFile;
	theSetupRec.theResourceFile.data = theInterfaceSetup->hostResourceFile;
	theSetupRec.SetupProc = theInterfaceSetup->HostInterfaceSetup;
	
	/* Set dialog title */	
	
	EI_SetDialogTitle(theSetupRec.theDialog, "BumpyNoise");
	
	/* Install control hit function */	
	
	SetControlHitFunc(ID_DENSITY_ET);
	SetControlHitFunc(ID_STRENGTH_ET);
	SetControlHitFunc(ID_THRESHOLD_ET);
	SetControlHitFunc(ID_RANGE_ET);
	
	SetControlHitFunc(ID_LOAD_BT);
	SetControlHitFunc(ID_SAVE_BT);
	SetControlHitFunc(ID_DEFAULT_BT);

	SetControlHitFunc(ID_DENSITY_SL);
	SetControlHitFunc(ID_STRENGTH_SL);
	SetControlHitFunc(ID_THRESHOLD_SL);
	SetControlHitFunc(ID_RANGE_SL);
	
	/* Setup sliders */

	SetupSliders();	
	
	return shaderNoError;
}