/*
 *		Copyright (c)1991-1997 Electric Image, Inc.
 *		This code is confidential and is a trade secret of Electric Image, Inc.
 */

#include "EIShader.h"
#include "FPU.h"
#include "Noise.h"
#include "PString.h"
#include "System.h"

#include <stddef.h>

#define shaderType				'CNOI'
#define shaderVersion			0x00000001
#define shaderRevVersion		0x01000000

#define defShaderDensity		1.0
#define defColorScaleR			1.0
#define defColorScaleG			1.0
#define defColorScaleB			1.0

typedef struct {
	unsigned long shaderOwnerType;
	unsigned long shaderVersionNum;
	float shaderDensity;
	float colorScaleR;
	float colorScaleG;
	float colorScaleB;
} ShaderInterfaceRec, *ShaderInterfacePtr;

typedef struct {
	float shaderDensity;
	float colorScaleR;
	float colorScaleG;
	float colorScaleB;
} ShaderDataRec, *ShaderDataPtr;

/* Static functions */

static void CalcColorNoise(EIShaderShadeRec *theShadePtr)
{
	ShaderDataPtr theShaderData = (ShaderDataPtr)theShadePtr->shaderData;
	EIShaderMapPtr theMap = theShadePtr->shaderMap;
	ARGB_Real *theColor = &theShadePtr->shaderColor;

	float x = theMap->map_X0 * theShaderData->shaderDensity;
	float y = theMap->map_Y0 * theShaderData->shaderDensity;
	float z = theMap->map_Z0 * theShaderData->shaderDensity;
	float CR =  theShaderData->colorScaleR;
	float CG =  theShaderData->colorScaleG;
	float CB =  theShaderData->colorScaleB;
	float density = theShaderData->shaderDensity * MAX3(theMap->map_DX, theMap->map_DY, theMap->map_DZ);
	float twice = density + density;
	float minValue, maxValue;
	float rx = x;
	float ry = y;
	float rz = z;
	float gx = x;
	float gy = y;
	float gz = z;
	float bx = x;
	float by = y;
	float bz = z;
	float rt = 0.0;
	float gt = 0.0;
	float bt = 0.0;
	float r, g, b;
	float scale = 1.0;
	long i = 6;
	
    /* Turbulence */

	while (i && (scale > twice)) {
		r = Noise(rx, ry, rz) - 0.5;
		g = Noise(gx, gy, gz) - 0.5;
		b = Noise(bx, by, bz) - 0.5;
		r = FABS(r);
		g = FABS(g);
		b = FABS(b);
		rt += scale * r;
		gt += scale * g;
		bt += scale * b;
		
		rx += rx + 123.45;
		ry += ry + 321.67;
		rz += rz + 231.33;
		gx += gx + 231.33;
		gy += gy + 123.45;
		gz += gz + 321.67;
		bx += bx + 321.67;
		by += by + 231.33;
		bz += bz + 123.45;
		
		scale *= .5;
		i--;
	}
	
    /* Fade in the last level of turbulence */

	if (i && (scale > density)) {
		float weight = (scale / density) - 1.0;

		r = Noise(rx, ry, rz) - 0.5;
		g = Noise(gx, gy, gz) - 0.5;
		b = Noise(bx, by, bz) - 0.5;
		r = FABS(r);
		g = FABS(g);
		b = FABS(b);
		rt += weight * scale * r;
		gt += weight * scale * g;
		bt += weight * scale * b;
	}
	
	/* Stretch the color range */
	
	minValue = MIN3(rt, gt, bt);
	maxValue = MAX3(rt, gt, bt);
	scale = 1.0 / (maxValue - minValue);
	theColor->a = 1.0;
	theColor->r = (rt - minValue) * scale / CR;
	theColor->g = (gt - minValue) * scale / CG;
	theColor->b = (bt - minValue) * scale / CB;
	
}  /* CalcColorNoise */

static void DefaultShaderInterface(ShaderInterfacePtr theShaderInterface)
{
	theShaderInterface->shaderOwnerType = shaderType;
	theShaderInterface->shaderVersionNum = shaderVersion;
	theShaderInterface->shaderDensity = defShaderDensity;
	theShaderInterface->colorScaleR = defColorScaleR;
	theShaderInterface->colorScaleG = defColorScaleG;
	theShaderInterface->colorScaleB = defColorScaleB;
}  /* DefaultShaderInterface */

static void FixShaderInterface(ShaderInterfacePtr theShaderInterface)
{
	FixField(theShaderInterface->shaderVersionNum);
	FixField(theShaderInterface->shaderDensity);
	FixField(theShaderInterface->colorScaleR);
	FixField(theShaderInterface->colorScaleG);
	FixField(theShaderInterface->colorScaleB);
}  /* FixShaderInterface */

static void GetShaderInterface(void* theInterface, long theSize, ShaderInterfacePtr theShaderInterface)
{
	/* Copy the shader interface data */
	
	if (theSize == sizeof(ShaderInterfaceRec))
		*theShaderInterface = *(ShaderInterfacePtr)theInterface;
	else {
		theShaderInterface->shaderOwnerType = 0;
		theShaderInterface->shaderVersionNum = 0;
	}
	
	/* Validate the shader interface data */
	
	if (theShaderInterface->shaderOwnerType != shaderType)
		DefaultShaderInterface(theShaderInterface);
	else if (theShaderInterface->shaderVersionNum == shaderRevVersion)
		FixShaderInterface(theShaderInterface);
	else if (theShaderInterface->shaderVersionNum != shaderVersion)
		DefaultShaderInterface(theShaderInterface);
}  /* GetShaderInterface */

/* Export functions */

long EIShaderInformation(EIShaderInformationRec *theInformationPtr)
{
	/* Setup the shader feature flags */
	
	BSET(theInformationPtr->shaderFeatureFlags, shaderColorFlag);
	BSET(theInformationPtr->shaderFeatureFlags, shaderAntialiasFlag);
	return shaderNoError;
}  /* EIShaderInformation */

long EIShaderInitialize(EIShaderInitializeRec *theInitializePtr)
{
	ShaderDataPtr theShaderData;
	
	/* Initialize the noise library */
	
	if (!InitNoise((NoiseMallocFunction)theInitializePtr->HostMalloc, NULL))
		return shaderMemoryError;
	
	/* Allocate the shader's private data record */
	
	theShaderData = (ShaderDataPtr)theInitializePtr->HostMalloc(sizeof(ShaderDataRec));
	if (theShaderData) {
		ShaderInterfaceRec theShaderInterface;
		
		/* Get the shader interface parameters */
		
		GetShaderInterface(theInitializePtr->shaderInterface, theInitializePtr->shaderInterfaceSize, &theShaderInterface);
				
		/* Setup the shader's private data */
		
		theShaderData->shaderDensity 	= theShaderInterface.shaderDensity;
		theShaderData->colorScaleR		= theShaderInterface.colorScaleR;
		theShaderData->colorScaleG		= theShaderInterface.colorScaleG;
		theShaderData->colorScaleB		= theShaderInterface.colorScaleB;
		
		
		/* Return the pointer to the shader's private data to the host */
		
		theInitializePtr->shaderData = theShaderData;
		return shaderNoError;
	} else
		return shaderMemoryError;
}  /* EIShaderInitialize */

long EIShaderFinish(EIShaderFinishRec *theFinishPtr)
{
	/* Get rid of the shader's private data */
	
	theFinishPtr->HostFree(theFinishPtr->shaderData);
	
	/* Finish the noise library */
	
	FinishNoise((NoiseFreeFunction)theFinishPtr->HostFree);
	return shaderNoError;
}  /* EIShaderFinish */

long EIShaderShade(EIShaderShadeRec *theShadePtr)
{
	if (BTST(theShadePtr->shaderFeatureFlags, shaderColorFlag))
		CalcColorNoise(theShadePtr);
	return shaderNoError;
}  /* EIShaderShade */
