#include <string.h>
#include "PluginClient.h"
#include "TrackCache.hpp"

#define CHUNK_SIZE	8192
#define COMPARE_EPS	1.0e-5

extern PluginInfoRec * theInfo;

struct CCacheValue {

// data members
	long mFrame;
	char * mGrpID;
	long mChannelID;
	CCacheValue * mNext;
	char mData[4];
};

struct CChunk {
	
// data members
	CChunk * mNext;
	long mOffset;
	char mData[4];
};

//------------- alloc/free ------------------

namespace {

void * MAlloc( long size )
{
	HostMemMallocRec hmr;
	hmr.theReference = theInfo->theHostReference;
	hmr.theSize = size;
	hmr.thePtr = 0;			
	if (theInfo->theHostProc(hostMemMalloc, sizeof(hmr), &hmr) != pluginNoErr) return 0;
	return hmr.thePtr;
}

void MFree( void * thePtr )
{
	HostMemFreeRec hfr;
	if (thePtr) {
	 hfr.theReference = theInfo->theHostReference;
	 hfr.thePtr = thePtr;
	 theInfo->theHostProc(hostMemFree, sizeof(hfr), &hfr);
	} 
} 


} // namespace

long Type2Size( long dataType )
{
	switch (dataType) {
	 case channelLabel: 
	 case channelEnd:	
	  return 0;
	
	 case channelReal:
	 case channelUnsignedReal:
	  return sizeof(double);
	  
	 case channelCRGB:	
	 case channelCHSV:	
	  return sizeof(RGB_Color);
	  
	 case channelARGB:
	 case channelAHSV:
	  return sizeof(ARGB_Color);
	  
	 case channelInteger:
	 case channelUnsigned:
	  return sizeof(long);
	  
	 case channelBoolean:
	  return sizeof(short);
	  
	 case channelCoordinate:
	  return sizeof(DCoordinate);
	 
	 case channelPascalString:
	  return sizeof(Str255);
	}
	return 0;
}

inline Boolean Compare( double d1, double d2 )
{
	return FABS(d1 - d2) < COMPARE_EPS;
}

Boolean CompareChannelValues( void * data1, void * data2, long dataType )
{
	if (!data1 || !data2) return false;
	double * d1 = (double *) data1;
	double * d2 = (double *) data2;
	
	switch (dataType) {
	 case channelLabel: 
	 case channelEnd:	
	  return false;
	
	 case channelReal:
	 case channelUnsignedReal:
	  return Compare(d1[0], d2[0]);
	  
	 case channelCoordinate:
	  return Compare(d1[0], d2[0]) && 
	  		 Compare(d1[1], d2[1]) &&
	  		 Compare(d1[2], d2[2]);
	 
	 case channelPascalString:
	 {
	  uchar * s1 = (uchar *) data1;
	  uchar * s2 = (uchar *) data1;
	  return memcmp(s1, s2, s1[0] + 1);
	 } 
	  
	 default:
	  return memcmp(data1, data2, Type2Size(dataType)) == 0; 
	}
	return 0;
	
}

//------------- CTrackCache ------------------

CCacheValue * CTrackCache::AllocVal( long dataType )
{
	long size = sizeof(CCacheValue) + Type2Size(dataType);
	if (mChunk) {
	 long offs = mChunk->mOffset;
	 if (offs + size <= CHUNK_SIZE) {
	  mChunk->mOffset += size;
	  return (CCacheValue *)(mChunk->mData + offs);
	 } 
	}
	CChunk * chunk = (CChunk *) MAlloc(sizeof(CChunk) + CHUNK_SIZE);
	if (!chunk) return 0;
	chunk->mOffset = size;
	chunk->mNext = mChunk;
	mChunk = chunk;
	return (CCacheValue *)  chunk->mData;
}
   	
void CTrackCache::Free( void )
{
	while (mChunk) {
	 CChunk * chunk = mChunk;
	 mChunk = mChunk->mNext;
	 MFree(chunk);
	}
	mValue = 0;
}
	
CCacheValue * CTrackCache::GetVal( char * grpID, long channelID, long frame )
{
	if (!mActive) return 0;
	CCacheValue * val = mValue;
	while (val) {
	 if (val->mGrpID == grpID && val->mChannelID == channelID && val->mFrame == frame)
	  return val;
	 val = val->mNext; 
	}
	return 0;
}

Boolean CTrackCache::Look( char * grpID, long channelID )
{
	CCacheValue * val = mValue;
	while (val) {
	 if (val->mGrpID == grpID && val->mChannelID == channelID) return true;
	 val = val->mNext; 
	}
	return false;
}

void * CTrackCache::Get( char * grpID, long channelID, long frame )
{
	CCacheValue * val = GetVal(grpID, channelID, frame);
	return val ? val->mData : 0;
}

Boolean CTrackCache::Put( char * grpID, long channelID, long frame, void * src, long dataType )
{
	if (!mActive) return false;
	CCacheValue * val = GetVal(grpID, channelID, frame);
	if (!val) {
	 val = AllocVal(dataType);
	 if (!val) return false;
	 val->mFrame = frame;
	 val->mGrpID = grpID;
	 val->mChannelID = channelID;
	 val->mNext = mValue;
	 mValue = val;
	}
	memmove(val->mData, src, Type2Size(dataType));	
	return true;
}
	
