/*
Copyright 1995-1998 Electric Image, Inc. This source code may be included in
commercial, freeware, or shareware products.  This copyright notice must be
included in its original form in all such products. This source code may not be
offered for sale or distributed by any entity without the written permission of
Electric Image, Inc.
*/

#include <math.h>
#include "EI_ImageBuffer.h"
#include "FACTStuff.h"
#include "PluginServer.h"
#include "PluginFPU.h"
#include "PluginUtils.h"
#include "PString.h"


/********************************************************************************************************/
/*			Noise Genrator																				*/
/********************************************************************************************************/
#define RANDOMSEED(seed) (seed = (long)seed * 1103515245 + 12345)			/* pick a new random seed */
#define RANDOMBITS(seed, bits) (RANDOMSEED(seed) >> (32 - (bits)))			/* create a random number with a specific number of bits */

/********************************************************************************************************/
/*			Datatypes																					*/
/********************************************************************************************************/
 typedef struct PluginDataRec {
    DCoordinate theLightPosition;
    L_Rect theImageRect;
    double theLightRadius;
    ARGB theLightColor;
  } PluginDataRec, *PluginDataPtr;

/********************************************************************************************************/
/*			Prototypes																					*/
/********************************************************************************************************/

int main (ulong theCommand, long theCommandRecSize, void* theCommandRec);

/********************************************************************************************************/
/*									AllocateMemory														*/
/********************************************************************************************************/
static long AllocateMemory(PluginInfoRec *theInfo)
{
 	HostAllocateRec theHostAllocate;
	long theError = 0;

 	theHostAllocate.theReference = theInfo->theHostReference;
	theHostAllocate.theControlData = theInfo->theControlData;
	theHostAllocate.theControlDataSize = 0;
	theHostAllocate.thePluginData = theInfo->thePluginData;
	theHostAllocate.thePluginDataSize = sizeof(PluginDataRec);

	theError = theInfo->theHostProc(PLUGCOMMAND(hostAllocate), PARAMETERSIZE(sizeof(HostAllocateRec)), &theHostAllocate);

	if (theError == 0) {
 		theInfo->theControlData = theHostAllocate.theControlData;
		theInfo->theControlDataSize = theHostAllocate.theControlDataSize;
		theInfo->thePluginData = theHostAllocate.thePluginData;
		theInfo->thePluginDataSize = theHostAllocate.thePluginDataSize;
	}
	return theError;
}  /* AllocateMemory */

/********************************************************************************************************/
/*									Status																*/
/********************************************************************************************************/
static void StatusCopyString(unsigned char *dst, unsigned char *src)
{
	long i;
	
	dst[0] = src[0];
	for (i = 1; i <= src[0]; i++)
		dst[i] = src[i];
}  /* StatusCopyString */

#define StatusTime(theStatus) (((theStatus)->theStatusTick == 0) || ((theStatus)->theStatusTick <= TICKCOUNT()))

static void StatusUpdate(PluginInfoPtr theInfo, HostStatusRec *theStatus, float thePos, long theValue, long *theResult)
{
	if (*theResult == 0) {
		theStatus->theReference = theInfo->theHostReference;
		theStatus->theStatusValue = theValue;
		theStatus->theStatusPos = thePos;
		*theResult = theInfo->theHostProc(PLUGCOMMAND(hostStatus), PARAMETERSIZE(sizeof(HostStatusRec)), theStatus);
	}
}  /* EIPluginStatusUpdate */

static void StatusSetup(PluginInfoPtr theInfo, HostStatusRec *theStatus, PStrPtr theUnit, PStrPtr theHeading, long min, long max, float thePos, long theValue, long *theResult)
{
	theStatus->theReference = theInfo->theHostReference;
	theStatus->theStatusPos = thePos;
	StatusCopyString(theStatus->theStatusUnit, theUnit);
	StatusCopyString(theStatus->theStatusInfo, theHeading);
	theStatus->theStatusMin = min;
	theStatus->theStatusMax = max;
	theStatus->theStatusValue = theValue;
	theStatus->theStatusTick = 0;
	StatusUpdate(theInfo, theStatus, thePos, theValue, theResult);
}  /* EIPluginStatusSetup */

static void StatusFinish(PluginInfoPtr theInfo, HostStatusRec *theStatus, long *theResult)
{
	if ((*theResult == 0) && theStatus->theStatusTick)
		StatusUpdate(theInfo, theStatus, 1.0, theStatus->theStatusMax, theResult);

}  /* EIPluginStatusFinish */

/********************************************************************************************************/
/*									FlareSetup															*/
/********************************************************************************************************/
static long FlareSetup(long theCommandRecSize, void* theCommandRec)
{
	PluginFlareSetupPtr theFlareSetupPtr = (PluginFlareSetupPtr)theCommandRec;
	PluginInfoRec theInfo;
	PluginDataPtr thePluginDataPtr;
  	long theError = 0;

	if (PLUGPARAMETERFAIL(theCommandRecSize, sizeof(PluginFlareSetupRec)))  
		theError = pluginParameterErr;
	else {
		theInfo = theFlareSetupPtr->theInfo;
		thePluginDataPtr = (PluginDataPtr)theInfo.thePluginData;
 		thePluginDataPtr->theLightPosition.x = theFlareSetupPtr->theLightFramePosition.x;
		thePluginDataPtr->theLightPosition.y = theFlareSetupPtr->theLightFramePosition.y;
		thePluginDataPtr->theImageRect = theFlareSetupPtr->theFrameRect;
		thePluginDataPtr->theLightRadius = theFlareSetupPtr->theLightFrameRadius;
		thePluginDataPtr->theLightColor = theFlareSetupPtr->theLightColor;
 	}
	return theError;
}   /* FlareSetup */

/********************************************************************************************************/
/*									FlareGenerate														*/
/********************************************************************************************************/
/*	This function draws into the appropriate buffer and flare elements that may appear in the rectangle */
/*	requested by the Host.																				*/
/********************************************************************************************************/
static long FlareGenerate (long theCommandRecSize, void* theCommandRec)
{
	PluginFlareGeneratePtr theFlareGenerate = (PluginFlareGeneratePtr)theCommandRec;
	PluginInfoRec theInfo;
	HostStatusRec theStatus;
	FlareBuffer theFlareBuffer;
	PluginDataPtr thePluginData;
	char *theScan;
	double xDist, yDist, theDistance, theRadius;
	L_Rect theFlareRect, theVisRect;
	long x, y, lightX, lightY;
	long r, g, b;
	long theError = 0;

	if (PLUGPARAMETERFAIL(theCommandRecSize, sizeof(PluginFlareGenerateRec)))  
		theError = pluginParameterErr;
	else {
		theInfo = theFlareGenerate->theInfo;
		thePluginData = (PluginDataPtr)theInfo.thePluginData;
		theFlareBuffer = theFlareGenerate->theFlareBuffer;

		/* Find the flare's center, radius and extent rectangle */
		
		lightX = thePluginData->theLightPosition.x;
		lightY = thePluginData->theLightPosition.y;
		theRadius = thePluginData->theLightRadius;
		if (theRadius <= 0.0)
			theRadius = 100.0;
		SetLRect(&theFlareRect, lightX - theRadius * 1.5, lightY - theRadius * 1.5, lightX + theRadius * 1.5, lightY + theRadius * 1.5);
		if (SectLRect(&theFlareBuffer.bufferRect, &theFlareRect, &theVisRect))

			/* If the bufferDepth is 48 bits, draw a color flare (up to version 2.7.5) */
			
			if (theFlareBuffer.bufferDepth == 48) {
				theRadius = -4.0 / (theRadius * theRadius);
	
				/* Calculate the pixel address of the top left corner of the visible rectangle */
				
				theScan = theFlareBuffer.bufferBase;
				theScan += (theVisRect.top - theFlareBuffer.bufferRect.top) * theFlareBuffer.bufferRowBytes;
				theScan += (theVisRect.left - theFlareBuffer.bufferRect.left) * sizeof(FlareColor);
	
				/* Find the flare's color */
				
				r = (thePluginData->theLightColor.argb.r * 16320L) / 255;
				g = (thePluginData->theLightColor.argb.g * 16320L) / 255;
				b = (thePluginData->theLightColor.argb.b * 16320L) / 255;
				
				/* Setup the status */
				
				StatusSetup(&theInfo, &theStatus, PS("Scanline"), PS("Generating Glow Flare"), theVisRect.top, theVisRect.bottom, 0.0, theVisRect.top, &theError);
				
				/* Draw the flare into the buffer */
				
				for (y = theVisRect.top; (y < theVisRect.bottom) && (theError == 0); y++) {
					FlareColorPtr thePixel = (FlareColorPtr)theScan;
					for (x = theVisRect.left; x < theVisRect.right; x++) {
						xDist = FABS(lightX - x);
						yDist = FABS(lightY - y);
						theDistance = xDist * xDist + yDist * yDist;
						theDistance = EXP(theDistance * theRadius);
						thePixel->red += MIN(thePixel->red + r * theDistance, 16320);
						thePixel->green += MIN(thePixel->green + g * theDistance, 16320);
						thePixel->blue += MIN(thePixel->blue + b * theDistance, 16320);
						thePixel++;
					}
					theScan += theFlareBuffer.bufferRowBytes;
					
					/* Update the status */
					
					if (StatusTime(&theStatus))
						StatusUpdate(&theInfo, &theStatus, (y - theVisRect.top) / (double)(theVisRect.height), y, &theError);
				}
				
				/* Finish the status */
				
				StatusFinish(&theInfo, &theStatus, &theError);

			/* If the bufferDepth is 32 bits, draw a color flare (version 2.8 and beyond) */
			
			} else if (theFlareBuffer.bufferDepth == 32) {
				theRadius = -4.0 / (theRadius * theRadius);
	
				/* Calculate the pixel address of the top left corner of the visible rectangle */
				
				theScan = theFlareBuffer.bufferBase;
				theScan += (theVisRect.top - theFlareBuffer.bufferRect.top) * theFlareBuffer.bufferRowBytes;
				theScan += (theVisRect.left - theFlareBuffer.bufferRect.left) * sizeof(ARGB);
	
				/* Find the flare's color */
				
				r = thePluginData->theLightColor.argb.r;
				g = thePluginData->theLightColor.argb.g;
				b = thePluginData->theLightColor.argb.b;
				
				/* Setup the status */
				
				StatusSetup(&theInfo, &theStatus, PS("Scanline"), PS("Generating Glow Flare"), theVisRect.top, theVisRect.bottom, 0.0, theVisRect.top, &theError);

				/* Draw the flare into the buffer */

				for (y = theVisRect.top; (y < theVisRect.bottom) && (theError == 0); y++) {
					EI_32BitPixel *thePixel = (EI_32BitPixel *)theScan;
					ulong seed = y;
					
					for (x = theVisRect.left; x < theVisRect.right; x++) {
						long rSum, gSum, bSum;
						long factor;
						long noise = RANDOMBITS(seed, 14);
						
						xDist = FABS(lightX - x);
						yDist = FABS(lightY - y);
						theDistance = xDist * xDist + yDist * yDist;
						factor = 16384.0 * EXP(theDistance * theRadius);

						rSum = EI_GET_PIXEL_RED(*thePixel) + (r * factor + noise) >> 14;
						gSum = EI_GET_PIXEL_GREEN(*thePixel) + (g * factor + noise) >> 14;
						bSum = EI_GET_PIXEL_BLUE(*thePixel) + (b * factor + noise) >> 14;
						
  						EI_SET_PIXEL_RED(*thePixel, MIN(rSum, 255));
 						EI_SET_PIXEL_GREEN(*thePixel, MIN(gSum, 255));
						EI_SET_PIXEL_BLUE(*thePixel, MIN(bSum, 255));
						thePixel++;
					}
					theScan += theFlareBuffer.bufferRowBytes;
					
					/* Update the status */
					
					if (StatusTime(&theStatus))
						StatusUpdate(&theInfo, &theStatus, (y - theVisRect.top) / (double)(theVisRect.height), y, &theError);
				}
				
				/* Finish the status */
				
				StatusFinish(&theInfo, &theStatus, &theError);
			} else if (theFlareBuffer.bufferDepth == 8) {
				
				/* The bufferDepth is 8 bits, so draw the flare's outline */
	
				/* Calculate the pixel address of the top left corner of the visible rectangle */
				
				theScan = theFlareBuffer.bufferBase;
				theScan += (theVisRect.top - theFlareBuffer.bufferRect.top) * theFlareBuffer.bufferRowBytes;
				theScan += theVisRect.left - theFlareBuffer.bufferRect.left;
				
				/* Draw the top edge */
				
				if (theFlareRect.top == theVisRect.top) {
					char *thePixel = theScan;
					for (x = theVisRect.left; x < theVisRect.right; x++) {
						*thePixel = theFlareBuffer.bufferIndex;
						thePixel++;
					}
				}
				
				/* Draw the bottom edge */
				
				if (theFlareRect.bottom == theVisRect.bottom) {
					char *thePixel = theScan + (theVisRect.height - 1) * theFlareBuffer.bufferRowBytes;
					for (x = theVisRect.left; x < theVisRect.right; x++) {
						*thePixel = theFlareBuffer.bufferIndex;
						thePixel++;
					}
				}
				
				/* Draw the left edge */
				
				if (theFlareRect.left == theVisRect.left) {
					char *thePixel = theScan;
					for (y = theVisRect.top; y < theVisRect.bottom; y++) {
						*thePixel = theFlareBuffer.bufferIndex;
						thePixel += theFlareBuffer.bufferRowBytes;
					}
				}
				
				/* Draw the right edge */
				
				if (theFlareRect.right == theVisRect.right) {
					char *thePixel = theScan + (theVisRect.width - 1);
					for (y = theVisRect.top; y < theVisRect.bottom; y++) {
						*thePixel = theFlareBuffer.bufferIndex;
						thePixel += theFlareBuffer.bufferRowBytes;
					}
				}
		}
	}
	return theError;
}   /* FlareGenerate */

/********************************************************************************************************/
/*									Initialize															*/
/********************************************************************************************************/
static long Initialize (long theCommandRecSize, void* theCommandRec) 
{
	PluginInitializePtr thePluginInitializePtr = (PluginInitializePtr)theCommandRec;
	long theError = 0;

	if (PLUGPARAMETERFAIL(theCommandRecSize, sizeof(PluginInitializeRec))) 
		theError = pluginParameterErr;
	else
 		theError = AllocateMemory(&thePluginInitializePtr->theInfo);
	return theError;
}   /* Initialize */

/********************************************************************************************************/
/*									main																*/
/********************************************************************************************************/
EXPDLL long Main (ulong theCommand, long theCommandRecSize, void* theCommandRec)
{	
	long theError = 0;

	switch (HOSTCOMMAND(theCommand))
	{
		case pluginInitialize:		theError = Initialize(theCommandRecSize, theCommandRec);		break;
		case pluginFlareSetup:		theError = FlareSetup(theCommandRecSize, theCommandRec);		break;
		case pluginFlareGenerate:	theError = FlareGenerate(theCommandRecSize, theCommandRec);		break;
	}
	return theError;

}	/* Main */
