/*
Copyright 1995-1998 Electric Image, Inc. This source code is intended for teaching
purposes only. It may not be included in any products (freeware, shareware, or
commercial) without the written permission of Electric Image, Inc. This source
code may be copied and modified but this copyright notice must be included in
its original form. This source code may not be offered for sale or distributed
by any entity without the written permission of Electric Image, Inc. Copies of
the original or modified code may be posted to the ElectricImage Forum on
America-on-line or eWorld to foster a better understanding of the Plug-In
Specification.
*/

/* This code presents an example of a particle system model plugin. */

#include "EI_Dialog.h"
#include "EI_FileRef.h"
#include "EI_ResourceFile.h"
#include "MemoryUtils.h"
#include "PluginClient.h"
#include "PluginFPU.h"
#include "PluginUtils.h"
#include "PluginMatrix.h"
#include "PString.h"

#define pluginType			QUADBYTE('P', 'A', 'R', 'T')		/* The owner plugin's unique type */
#define pluginVersion		0x00000003				/* The owner plugin's version number 0:0:version:revision */
#define pluginName			PS("ElectricImage Particle")
#define pluginCopyright		PS("Copyright  1990-1998 ElectricImage, Inc.")
#define pluginAuthor		PS("Mark Granger|Electric Image, Inc.|117 East Colorado Boulevard, Suite 300|Pasadena, Ca 91105|(818) 577-1627")
#define pluginDescription	PS("This model plugin can generate a variety of particle systems including fountains and explosions.")
#define pluginPermission	BIT(pluginSensitivityFlag) + BIT(pluginTimeSensitivityFlag)

#define defFrameTimeDelta  0.03333333   /* 30fps */
/* Particle Modes: */
/* 1:  Vertical Fountain */
/* 2:  Shower of Sparks */
/* 3:  Bomb Burst */
#define particleMode       1
#if (particleMode == 1)
/* This set of constants will create a vertical fountain with bouncing particles */
#define defMaxParticles		1000		/* The maximum number of particles in the simulation */
#define defEmitFrequency	25.0		/* The number of particles which are created per unit of time */
#define defEmitErr			25.0		/* The error in the number of particles created */
#define defEmitStart		0.0  		/* The time the first particle is created */
#define defEmitEnd			100.0  		/* The time that the last particle is created */

#define defLifeTime			10.0  		/* The lifetime of a particle */
#define defLifeTimeErr		0.0  		/* The error in the lifetime of a particle */

#define defInitXPos			0.0  		/* The starting location of a particle */
#define defInitYPos			0.0
#define defInitZPos			0.0
#define defInitXPosErr		0.0			/* The error in the starting location of a particle */
#define defInitYPosErr		0.0
#define defInitZPosErr		0.0

#define defInitXDir			0.0			/* The inital motion direction of a particle */
#define defInitYDir			1.0
#define defInitZDir			0.0
#define defInitAngle		0.0			/* The angle of deviation from the direction vector */
#define defInitAngleErr		2.0			/* The error in the angle of deviation from the direction vector */

#define defInitVelocity		100.0  		/* The initial particle velocity */
#define defInitVelocityErr	10.0   		/* The error in the initial velocity */
#define defAirResistance	0.0			/* The air resistance factor (0.0...1.0; 0.0=no wind resistance) */

#define defGravityXDir		0.0			/* The gravitational force which pulls the particle per unit of time */
#define defGravityYDir		(-50.0)
#define defGravityZDir		0.0

#define defGroundLevel		0.0  		/* The altitude at which the particles bounce */
#define defBounceFactor		0.5			/* The amount which the particle will bounce after it has hit the ground plane */
#define defBounceFriction	0.9			/* The amount which the particle will bounce in the x and z directions */

#define defA0				255			/* Minimum energy color at the end of the particle's life */
#define defR0				200
#define defG0				0
#define defB0				0

#define defA1				255			/* Medium energy color at the middle of the particle's life */
#define defR1				255
#define defG1				128
#define defB1				0

#define defA2				255			/* Maximum energy color and the beginning of the particle's life */
#define defR2				255
#define defG2				255
#define defB2				255
#endif

#if (particleMode == 2)
/* This set of constants will create a shower of sparks */
#define defMaxParticles		1000		/* The maximum number of particles in the simulation */
#define defEmitFrequency	50.0		/* The number of particles which are created per unit of time */
#define defEmitErr			50.0		/* The error in the number of particles created */
#define defEmitStart		0.0  		/* The time the first particle is created */
#define defEmitEnd			100.0  		/* The time that the last particle is created */

#define defLifeTime			10.0  		/* The lifetime of a particle */
#define defLifeTimeErr		0.0  		/* The error in the lifetime of a particle */

#define defInitXPos			0.0  		/* The starting location of a particle */
#define defInitYPos			100.0
#define defInitZPos			0.0
#define defInitXPosErr		0.0			/* The error in the starting location of a particle */
#define defInitYPosErr		0.0
#define defInitZPosErr		0.0

#define defInitXDir			0.0			/* The inital motion direction of a particle */
#define defInitYDir			0.0
#define defInitZDir			0.0
#define defInitAngle		0.0			/* The angle of deviation from the direction vector */
#define defInitAngleErr		180.0		/* The error in the angle of deviation from the direction vector */

#define defInitVelocity		1.0  		/* The initial particle velocity */
#define defInitVelocityErr	0.1   		/* The error in the initial velocity */
#define defAirResistance	0.1			/* The air resistance factor (0.0...1.0; 0.0=no wind resistance) */

#define defGravityXDir		0.0			/* The gravitational force which pulls the particle per unit of time */
#define defGravityYDir		(-0.5)
#define defGravityZDir		0.0

#define defGroundLevel		0.0  		/* The altitude at which the particles bounce */
#define defBounceFactor		0.5			/* The amount which the particle will bounce after it has hit the ground plane */
#define defBounceFriction	0.9			/* The amount which the particle will bounce in the x and z directions */

#define defA0				255			/* Minimum energy color at the end of the particle's life */
#define defR0				200
#define defG0				0
#define defB0				0

#define defA1				255			/* Medium energy color at the middle of the particle's life */
#define defR1				255
#define defG1				128
#define defB1				0

#define defA2				255			/* Maximum energy color and the beginning of the particle's life */
#define defR2				255
#define defG2				255
#define defB2				255
#endif

#if (particleMode == 3)
/* This set of constants will create a bomb burst effect */
#define defMaxParticles		400			/* The maximum number of particles in the simulation */
#define defEmitFrequency	400.0		/* The number of particles which are created per unit of time */
#define defEmitErr			0.0			/* The error in the number of particles created */
#define defEmitStart		0.0  		/* The time the first particle is created */
#define defEmitEnd			0.0  		/* The time that the last particle is created */

#define defLifeTime			20.0  		/* The lifetime of a particle */
#define defLifeTimeErr		0.0  		/* The error in the lifetime of a particle */

#define defInitXPos			0.0  		/* The starting location of a particle */
#define defInitYPos			20.0
#define defInitZPos			0.0
#define defInitXPosErr		0.0			/* The error in the starting location of a particle */
#define defInitYPosErr		0.0
#define defInitZPosErr		0.0

#define defInitXDir			0.0			/* The inital motion direction of a particle */
#define defInitYDir			0.0
#define defInitZDir			1.0
#define defInitAngle		0.0			/* The angle of deviation from the direction vector */
#define defInitAngleErr		90.0			/* The error in the angle of deviation from the direction vector */

#define defInitVelocity		10.0  		/* The initial particle velocity */
#define defInitVelocityErr	2.0   		/* The error in the initial velocity */
#define defAirResistance	0.0			/* The air resistance factor (0.0...1.0; 0.0=no wind resistance) */

#define defGravityXDir		0.0			/* The gravitational force which pulls the particle per unit of time */
#define defGravityYDir		(-2.0)
#define defGravityZDir		0.0

#define defGroundLevel		0.0  		/* The altitude at which the particles bounce */
#define defBounceFactor		0.5			/* The amount which the particle will bounce after it has hit the ground plane */
#define defBounceFriction	0.9			/* The amount which the particle will bounce in the x and z directions */

#define defA0				255			/* Minimum energy color at the end of the particle's life */
#define defR0				0
#define defG0				0
#define defB0				255

#define defA1				255			/* Medium energy color at the middle of the particle's life */
#define defR1				128
#define defG1				192
#define defB1				255

#define defA2				255			/* Maximum energy color and the beginning of the particle's life */
#define defR2				255
#define defG2				255
#define defB2				255
#endif

#define oldPluginType			QUADBYTE('P', 'A', 'R', 'T')
#define oldPluginVersion		0x00000002
#define oldControlSize			276

#define oldOwnerType			0
#define oldOwnerVersion			4
#define oldLastFrameIndex		8
#define oldLastFrameTime		12
#define oldLastFrameTimeDelta	24
#define oldParticleCount		36
#define oldParticleFlags		40
#define oldFirstActive			44
#define oldFirstInactive		48
#define oldMaxParticles			52
#define oldEmitFrequency		56
#define oldEmitErr				64
#define oldEmitStart			72
#define oldEmitEnd				80
#define oldLifeTime				88
#define oldLifeTimeErr			96
#define oldInitXPos				104
#define oldInitYPos				112
#define oldInitZPos				120
#define oldInitXPosErr			128
#define oldInitYPosErr			136
#define oldInitZPosErr			144
#define oldInitXDir				152
#define oldInitYDir				160
#define oldInitZDir				168
#define oldInitAngle			176
#define oldInitAngleErr			184
#define oldInitVelocity			192
#define oldInitVelocityErr		200
#define oldAirResistance		208
#define oldGravityXDir			216
#define oldGravityYDir			224
#define oldGravityZDir			232
#define	oldGroundLevel			240
#define oldBounceFactor			248
#define oldBounceFriction		256
#define oldEnergyColor0			264
#define oldEnergyColor1			268
#define oldEnergyColor2			272

#define interfaceDialogID  16001

enum {
	cMaxParticlesET = 5,
	cNewParticlesET = 6,
	cParticleLifeTimeET = 7,
	cCreateFromTimeET = 8,
	cXPositionET = 9,
	cYPositionET = 10,
	cZPositionET = 11,
	cXPlusMinusET = 12,
	cYPlusMinusET = 13,
	cZPlusMinusET = 14,
	cNewParticlesPlusMinusET = 15,
	cLifeTimePlusMinusET = 16,
	cCreateToTimeET = 17,
	cDeviationAngleET = 18,
	cVelocityET = 19,
	cXDirectionET = 20,
	cYDirectionET = 21,
	cGravityXET = 22,
	cZDirectionET = 23,
	cGravityZET = 24,
	cYGroundLevelET = 25,
	cBounceFactorET = 26,
	cAirResistanceET = 27,
	cBounceFrictionET = 28,
	cDeviationPlusMinusET = 29,
	cVelocityPlusMinusET = 30,
	cGravityYET = 31,
	cEnergyOneColor = 35,
	cEnergyHalfColor = 36,
	cEnergyZeroColor = 37
};

typedef ARGB ParticleColorArray[3];

typedef struct PluginDataRec {
	double frameTime;
	double frameTimeDelta;
	Matrix4 theMatrix;
	Matrix4 theBlurMatrix;
	Matrix3 directionMatrix;
	long frameIndex;
	ulong vertexFlags;
} PluginDataRec, *PluginDataPtr;

typedef struct ControlDataRec {
	unsigned long ownerType;			/* The owner plugin's unique type */
	long ownerVersion;		/* The owner plugin's version number */

	double lastFrameTime;		/* The time of the last frame which was generated */
	double lastFrameTimeDelta;	/* The time delta (1/fps) of the last frame which was generated */

	double emitFrequency;		/* The number of particles which are created per unit of time */
	double emitErr;				/* The error in the number of particles created */
	double emitStart;			/* The time the first particle is created */
	double emitEnd;				/* The time that the last particle is created */

	double lifeTime;			/* The lifetime of a particle */
	double lifeTimeErr;			/* The error in the lifetime of a particl e*/

	double initXPos;			/* The starting location of a particle */
	double initYPos;
	double initZPos;
	double initXPosErr;			/* The error in the starting location of a particle */
	double initYPosErr;
	double initZPosErr;

	double initXDir;			/* The inital motion direction of a particle */
	double initYDir;
	double initZDir;
	double initAngle;			/* The angle of deviation from the direction vector */
	double initAngleErr;		/* The error in the angle of deviation from the direction vector */

	double initVelocity;  		/* The initial particle velocity */
	double initVelocityErr;		/* The error in the initial velocity */
	double airResistance;		/* The air resistance factor (0.0...1.0; 0.0=no wind resistance) */

	double gravityXDir;			/* The gravitational force which pulls the particle per unit of time */
	double gravityYDir;
	double gravityZDir;

	double groundLevel;			/* The altitude at which the particles bounce */
	double bounceFactor;		/* The amount which the particle will bounce after it has hit the ground plane */
	double bounceFriction;		/* The amount which the particle will bounce in the x and z directions */

	long lastFrameIndex;		/* The index of the last frame which was generated */
	long particleCount;		/* The current number of particles in the system */
	long particleFlags;		/* The option flags for the particle system */
	long firstActive;		/* The index of the first active particle */
	long firstInactive;		/* The index of the first inactive particle */

	long maxParticles;		/* The maximum number of particles in the system */

	ParticleColorArray energyColors;	/* The color of each particle as a function of its energy */
	long particleSeed;				/* Random seed index */
} ControlDataRec, *ControlDataPtr;

typedef struct ParticleRec {
	float xPos, yPos, zPos;		/* Current position */
	float xOld, yOld, zOld;		/* Previous position */
	float xDir, yDir, zDir;		/* Motion vectory */
	float energy;				/* Particle engergy */
	float decay;					/* Speed at which energy decays */
	ARGB color;					/* Particle color */
	long link;				/* Link to the next particle */
} ParticleRec, *ParticlePtr, **ParticleHandle;

static void InitGlobals(boolean initFlag);
static void InitParticleList(ControlDataRec *theControl, PluginDataRec *theData);
static void FixControl(ControlDataPtr theControlRec);
static void UpdateControl(unsigned char* oldControl, ControlDataPtr newControl);
static void DefaultControl(ControlDataPtr theControlRec);
static void DefaultData(PluginDataPtr theData);
static long CheckData(PluginInfoPtr theInfo, boolean setControlData, boolean setPluginData);
static void SendModelReset(PluginInfoPtr theInfo);
static boolean StatusTime(void);
static void UpdateStatus(float thePos, long theValue);
static void AddGroup(char* groupReference, long groupVertexFlags);
static long AddBlurVertex(double xp, double yp, double zp, double xb, double yb, double zb);
static long AddVertex(double xp, double yp, double zp);
static void AddFacet(ARGB color, short numVertices, long v0, long v1, long v2, long v3);
static ulong RandomCache(void);
static void EnergyColor(double theEnergy, ARGB *theColor);
static ParticlePtr FindParticle(long n);
static void AddParticle(ParticlePtr theParticle);
static void DeleteParticle(long theParticleIndex, long lastParticleIndex);
static boolean CalcParticle(ParticlePtr theParticle, long frameCount, double st, double dt);
static void CalcParticles(long numFrames, double dt);
static void CreateParticles(long frame, long numFrames, double time, double dt);
static void InitParticles(void);
static void AgeParticles(void);
static void OutputParticles(void);

static long DoInitialize(long theCommandRecSize, void* theCommandRec);
static long DoFinish(long theCommandRecSize, void* theCommandRec);
static long DoInformation(long theCommandRecSize, void* theCommandRec);

static long DoCheck(long theCommandRecSize, void* theCommandRec);
static long DoDescribe(long theCommandRecSize, void* theCommandRec);

static long DoInterface(long theCommandRecSize, void* theCommandRec);

static long DoSetup(long theCommandRecSize, void* theCommandRec);
static long DoGINF(long theCommandRecSize, void* theCommandRec);
static long DoGATR(long theCommandRecSize, void* theCommandRec);
static long DoGBLR(long theCommandRecSize, void* theCommandRec);
static long DoTMAP(long theCommandRecSize, void* parameterse);
static long DoBMAP(long theCommandRecSize, void* theCommandRec);
static long DoRMAP(long theCommandRecSize, void* theCommandRec);
static long DoGenerate(long theCommandRecSize, void* theCommandRec);

static unsigned long globalValidation1, globalValidation2;
static long result;
static PluginDataPtr theData;
static ControlDataPtr theControl;
static ParticlePtr particleData;
static EIPluginRoutine hostProc;
static char* hostReference;
static HostStatusRec theStatusRec;
static ParticleColorArray theColors;

/* Random number generator */

#define randomCacheSize 97
#define RANDOMIZE(seed) ((long)(seed) * 1103515245 + 12345)
#define RANDOMSEED(seed) (seed = RANDOMIZE(seed))
#define RandomFactor() (((short)(RandomCache() >> 16)) * 3.0517578E-5)	/* Return a random number between -1 and 1 */
#define RandomNum() ((RandomCache() >> 16) * 1.5259022E-5)					/* Return a random number between 0 and 1 */

static ulong randomSeed, randomIndex, randomCache[randomCacheSize];

static void SeedRandomCache(ulong theSeed)
{
	long i;
	
	randomSeed = theSeed;
	for (i = 0; i < randomCacheSize; i++)
		randomCache[i] = RANDOMSEED(randomSeed);
	randomIndex = RANDOMSEED(randomSeed);
}  /* SeedRandomCache */

static ulong RandomCache(void)
{
	long i = (randomCacheSize * (randomIndex >> 16)) >> 16;

	randomIndex = randomCache[i];
	randomCache[i] = RANDOMSEED(randomSeed);
	return randomIndex;
}  /* RandomCache */

static void InitGlobals(boolean initFlag)
{
	if (initFlag || (globalValidation1 != pluginType) || (globalValidation2 != pluginModelType)) {
		globalValidation1 = pluginType;
		globalValidation2 = pluginModelType;
		InitUtils();
	}
}  /* InitGlobals */

static void InitParticleList(ControlDataPtr theControl, PluginDataPtr theData)
{
	long i;
	ParticlePtr theParticle;

	theControl->lastFrameIndex = 0;
	theControl->lastFrameTime = 0.0;
	theControl->lastFrameTimeDelta = defFrameTimeDelta;
	theControl->particleCount = 0;
	theControl->firstActive = -1;
	theControl->firstInactive = 0;
	theParticle = (ParticlePtr)((char*)theData + sizeof(PluginDataRec));
	for (i = 1; i < theControl->maxParticles; i++)
	{
		theParticle->link = i;
		theParticle++;
	}
	theParticle->link = -1;
}  /* InitParticleList */

static void FixControl(ControlDataPtr theControlRec)
{
	theControlRec->ownerType = pluginType;
	theControlRec->ownerVersion = pluginVersion;

	theControlRec->lastFrameIndex = 0x80000000;	/* Force an initialization of the particle list */
	theControlRec->lastFrameTime = 0.0;
	theControlRec->lastFrameTimeDelta = 0.0;
	theControlRec->particleCount = 0;
	theControlRec->firstActive = -1;
	theControlRec->firstInactive = 0;

	FixField(theControlRec->particleFlags);
	
	FixField(theControlRec->maxParticles);
	FixField(theControlRec->emitFrequency);
	FixField(theControlRec->emitErr);
	FixField(theControlRec->emitStart);
	FixField(theControlRec->emitEnd);

	FixField(theControlRec->lifeTime);
	FixField(theControlRec->lifeTimeErr);

	FixField(theControlRec->initXPos);
	FixField(theControlRec->initYPos);
	FixField(theControlRec->initZPos);
	FixField(theControlRec->initXPosErr);
	FixField(theControlRec->initYPosErr);
	FixField(theControlRec->initZPosErr);

	FixField(theControlRec->initXDir);
	FixField(theControlRec->initYDir);
	FixField(theControlRec->initZDir);
	FixField(theControlRec->initAngle);
	FixField(theControlRec->initAngleErr);

	FixField(theControlRec->initVelocity);
	FixField(theControlRec->initVelocityErr);
	FixField(theControlRec->airResistance);

	FixField(theControlRec->gravityXDir);
	FixField(theControlRec->gravityYDir);
	FixField(theControlRec->gravityZDir);

	FixField(theControlRec->groundLevel);
	FixField(theControlRec->bounceFactor);
	FixField(theControlRec->bounceFriction);

	FixField(theControlRec->particleSeed);
}  /* FixControl */

static void UpdateControl(unsigned char* oldControl, ControlDataPtr newControl)
{
	CopyField(oldControl + oldOwnerType, newControl->ownerType);
	newControl->ownerVersion = pluginVersion;

	newControl->lastFrameIndex = 0x80000000;	/* Force an initialization of the particle list */
	newControl->lastFrameTime = 0.0;
	newControl->lastFrameTimeDelta = 0.0;
	newControl->particleCount = 0;
	CopyField(oldControl + oldParticleFlags, newControl->particleFlags);
	newControl->firstActive = -1;
	newControl->firstInactive = 0;

	CopyField(oldControl + oldMaxParticles, newControl->maxParticles);

	CopyField(oldControl + oldEmitFrequency, newControl->emitFrequency);
	CopyField(oldControl + oldEmitErr, newControl->emitErr);
	CopyField(oldControl + oldEmitStart, newControl->emitStart);
	CopyField(oldControl + oldEmitEnd, newControl->emitEnd);

	CopyField(oldControl + oldLifeTime, newControl->lifeTime);
	CopyField(oldControl + oldLifeTimeErr, newControl->lifeTimeErr);

	CopyField(oldControl + oldInitXPos, newControl->initXPos);
	CopyField(oldControl + oldInitYPos, newControl->initYPos);
	CopyField(oldControl + oldInitZPos, newControl->initZPos);
	CopyField(oldControl + oldInitXPosErr, newControl->initXPosErr);
	CopyField(oldControl + oldInitYPosErr, newControl->initYPosErr);
	CopyField(oldControl + oldInitZPosErr, newControl->initZPosErr);

	CopyField(oldControl + oldInitXDir, newControl->initXDir);
	CopyField(oldControl + oldInitYDir, newControl->initYDir);
	CopyField(oldControl + oldInitZDir, newControl->initZDir);
	CopyField(oldControl + oldInitAngle, newControl->initAngle);
	CopyField(oldControl + oldInitAngleErr, newControl->initAngleErr);

	CopyField(oldControl + oldInitVelocity, newControl->initVelocity);
	CopyField(oldControl + oldInitVelocityErr, newControl->initVelocityErr);
	CopyField(oldControl + oldAirResistance, newControl->airResistance);

	CopyField(oldControl + oldGravityXDir, newControl->gravityXDir);
	CopyField(oldControl + oldGravityYDir, newControl->gravityYDir);
	CopyField(oldControl + oldGravityZDir, newControl->gravityZDir);

	CopyField(oldControl + oldGroundLevel, newControl->groundLevel);
	CopyField(oldControl + oldBounceFactor, newControl->bounceFactor);
	CopyField(oldControl + oldBounceFriction, newControl->bounceFriction);
	
	CopyField(oldControl + oldEnergyColor0, newControl->energyColors[0]);
	CopyField(oldControl + oldEnergyColor1, newControl->energyColors[1]);
	CopyField(oldControl + oldEnergyColor2, newControl->energyColors[2]);

	newControl->particleSeed = RANDOMIZE(TICKCOUNT());
}  /* UpdateControl */

static void DefaultControl(ControlDataPtr theControlRec)
{
	theControlRec->ownerType = pluginType;
	theControlRec->ownerVersion = pluginVersion;

	theControlRec->lastFrameIndex = 0x80000000;	/* Force an initialization of the particle list */
	theControlRec->lastFrameTime = 0.0;
	theControlRec->lastFrameTimeDelta = 0.0;
	theControlRec->particleCount = 0;
	theControlRec->firstActive = -1;
	theControlRec->firstInactive = 0;

	theControlRec->particleFlags = 0;

	theControlRec->maxParticles = defMaxParticles;
	theControlRec->emitFrequency = defEmitFrequency;
	theControlRec->emitErr = defEmitErr;
	theControlRec->emitStart = defEmitStart;
	theControlRec->emitEnd = defEmitEnd;

	theControlRec->lifeTime = defLifeTime;
	theControlRec->lifeTimeErr = defLifeTimeErr;

	theControlRec->initXPos = defInitXPos;
	theControlRec->initYPos = defInitYPos;
	theControlRec->initZPos = defInitZPos;
	theControlRec->initXPosErr = defInitXPosErr;
	theControlRec->initYPosErr = defInitYPosErr;
	theControlRec->initZPosErr = defInitZPosErr;

	theControlRec->initXDir = defInitXDir;
	theControlRec->initYDir = defInitYDir;
	theControlRec->initZDir = defInitZDir;
	theControlRec->initAngle = defInitAngle;
	theControlRec->initAngleErr = defInitAngleErr;

	theControlRec->initVelocity = defInitVelocity;
	theControlRec->initVelocityErr = defInitVelocityErr;
	theControlRec->airResistance = defAirResistance;

	theControlRec->gravityXDir = defGravityXDir;
	theControlRec->gravityYDir = defGravityYDir;
	theControlRec->gravityZDir = defGravityZDir;

	theControlRec->groundLevel = defGroundLevel;
	theControlRec->bounceFactor = defBounceFactor;
	theControlRec->bounceFriction = defBounceFriction;

	theControlRec->energyColors[0].argb.a = defA0;
	theControlRec->energyColors[0].argb.r = defR0;
	theControlRec->energyColors[0].argb.g = defG0;
	theControlRec->energyColors[0].argb.b = defB0;
	theControlRec->energyColors[1].argb.a = defA1;
	theControlRec->energyColors[1].argb.r = defR1;
	theControlRec->energyColors[1].argb.g = defG1;
	theControlRec->energyColors[1].argb.b = defB1;
	theControlRec->energyColors[2].argb.a = defA2;
	theControlRec->energyColors[2].argb.r = defR2;
	theControlRec->energyColors[2].argb.g = defG2;
	theControlRec->energyColors[2].argb.b = defB2;

	theControlRec->particleSeed = RANDOMIZE(TICKCOUNT());
}  /* DefaultControl */

static void DefaultData(PluginDataPtr theData)
{
	theData->frameIndex = 0;
	theData->frameTime = 0.0;
	theData->frameTimeDelta = defFrameTimeDelta;
	theData->vertexFlags = 0;
	InitMatrix4(&theData->theMatrix);
	InitMatrix4(&theData->theBlurMatrix);
	InitMatrix3(&theData->directionMatrix);
}  /* DefaultData */

static long CheckData(PluginInfoPtr theInfo, boolean setControlData, boolean setPluginData)
{
	HostAllocateRec theAllocate;
	long newPluginDataSize, newControlDataSize;
	boolean newPluginData, newControlData;
	ControlDataRec theControlRec;
	PluginDataRec theDataRec;
	ControlDataPtr theControl;
	long theResult;

	theResult = 0;

	/* Change the size of the plugin data block or control data block if the size is too small */

	theControl = (ControlDataPtr)theInfo->theControlData;
	if ((!setControlData) && (theInfo->theControlDataSize == oldControlSize) && (theControl->ownerType == oldPluginType) && (theControl->ownerVersion == oldPluginVersion)) {
		UpdateControl((unsigned char*)theControl, &theControlRec);
		setControlData = 1;
	} else if ((!setControlData) && (theInfo->theControlDataSize == sizeof(ControlDataRec))
		&& ((theControl->ownerType == pluginType) || (theControl->ownerType == REVERSELONG(pluginType)))
		&& ((theControl->ownerVersion == pluginVersion) || (theControl->ownerVersion == REVERSELONG(pluginVersion)))) {
		if (theControl->ownerVersion == REVERSELONG(pluginVersion))
			FixControl(theControl);
		theControlRec = *theControl;
	} else {
		setControlData = 1;
		DefaultControl(&theControlRec);
	}

	if (setPluginData || (theInfo->thePluginDataSize < sizeof(PluginDataRec))) {
		setPluginData = 1;
		DefaultData(&theDataRec);
	} else
		theDataRec = *(PluginDataPtr)theInfo->thePluginData;

	newControlDataSize = sizeof(ControlDataRec);
	newControlData = (theInfo->theControlDataSize < newControlDataSize || theInfo->thePluginDataSize > FixAlignment(newControlDataSize));
	newPluginDataSize = sizeof(PluginDataRec) + theControlRec.maxParticles * sizeof(ParticleRec);
	newPluginData = (theInfo->thePluginDataSize < newPluginDataSize || theInfo->thePluginDataSize > FixAlignment(newPluginDataSize));

	if (newPluginData || newControlData) {
		theAllocate.theReference = theInfo->theHostReference;
		theAllocate.theControlDataSize = newControlDataSize;
		theAllocate.theControlData = theInfo->theControlData;
		theAllocate.thePluginDataSize = newPluginDataSize;
		theAllocate.thePluginData = theInfo->thePluginData;
		theResult = theInfo->theHostProc(hostAllocate, sizeof(HostAllocateRec), &theAllocate);
		theInfo->theControlData = theAllocate.theControlData;
		theInfo->theControlDataSize = theAllocate.theControlDataSize;
		theInfo->thePluginData = theAllocate.thePluginData;
		theInfo->thePluginDataSize = theAllocate.thePluginDataSize;
	}

	if (theResult == 0) {
	
		/* Setup the control data to the default values if an incorrect sized block was passed to the plugin */
	
		if (newControlData || setControlData) {
			*(ControlDataPtr)theInfo->theControlData = theControlRec;
			InitParticleList((ControlDataPtr)theInfo->theControlData, (PluginDataPtr)theInfo->thePluginData);
		}
	
		/* Setup the plugin private data */
	
		if (newPluginData || setPluginData)
			*(PluginDataPtr)theInfo->thePluginData = theDataRec;
	}
	return theResult;
}  /* CheckData */

/* Reset the host's model data */

static void SendModelReset(PluginInfoPtr theInfo)
{
	HostModelResetRec theModelReset;

	theModelReset.theReference = theInfo->theHostReference;
	result = theInfo->theHostProc(hostModelReset, sizeof(HostModelResetRec), &theModelReset);
}  /* SendModelReset */

static boolean StatusTime(void)
{
	return ((theStatusRec.theStatusTick == 0) || (TICKCOUNT() >= theStatusRec.theStatusTick));
}  /* StatusTime */

static void UpdateStatus(float thePos, long theValue)
{
	if (result == 0) {
		theStatusRec.theStatusValue = theValue;
		theStatusRec.theStatusPos = thePos;
		result = hostProc(hostStatus, sizeof(HostStatusRec), &theStatusRec);
	}
}  /* UpdateStatus */

static void AddGroup(char* groupReference, long groupVertexFlags)
{
	HostModelAddGroupRec theAddGroup;

	if (result == 0) {
		theAddGroup.theReference = hostReference;
		theAddGroup.theGroup = NULL;
		theAddGroup.theGroupTemplate = groupReference;
		theAddGroup.theGroupVertexFlags = groupVertexFlags;
		result = hostProc(hostModelAddGroup, sizeof(HostModelAddGroupRec), &theAddGroup);
	}
}  /* AddGroup */

static long AddBlurVertex(double xp, double yp, double zp, double xb, double yb, double zb)
{
	HostModelAddVertexRec theAddVertex;

	if (result == 0)
	{
		TransformPoint(&xp, &yp, &zp, &theData->theMatrix);
		TransformPoint(&xb, &yb, &zb, &theData->theBlurMatrix);

		theAddVertex.theReference = hostReference;
		theAddVertex.theColor.colorValue = fieldARGB;
		theAddVertex.theNormal.x = 0.0;
		theAddVertex.theNormal.y = 0.0;
		theAddVertex.theNormal.z = 0.0;
		theAddVertex.thePosition.x = xp;
		theAddVertex.thePosition.y = yp;
		theAddVertex.thePosition.z = zp;
		theAddVertex.theBlurPosition.x = xb;
		theAddVertex.theBlurPosition.y = yb;
		theAddVertex.theBlurPosition.z = zb;
		theAddVertex.theTexturePosition = theAddVertex.thePosition;
		result = hostProc(hostModelAddVertex, sizeof(HostModelAddVertexRec), &theAddVertex);
	}
	if (result == 0)
		return theAddVertex.theIndex;
	else
		return 0;
}  /* AddBlurVertex */

static long AddVertex(double xp, double yp, double zp)
{
	return AddBlurVertex(xp, yp, zp, xp, yp, zp);
}  /* AddVertex */

static void AddFacet(ARGB color, short numVertices, long v0, long v1, long v2, long v3)
{
	HostModelAddFacetRec theAddFacet;

	if (result == 0) {
		theAddFacet.theReference = hostReference;
		theAddFacet.theColor = color;
		theAddFacet.theNumVertices = numVertices;
		theAddFacet.theVertices[0] = v0;
		theAddFacet.theVertices[1] = v1;
		theAddFacet.theVertices[2] = v2;
		theAddFacet.theVertices[3] = v3;
		result = hostProc(hostModelAddFacet, sizeof(HostModelAddFacetRec), &theAddFacet);
	}
}  /* AddFacet */

static void EnergyColor(double theEnergy, ARGB *theColor)
{
	double f1, f2;

	if (theEnergy <= 0.0)
		*theColor = theControl->energyColors[0];
	if (theEnergy >= 1.0)
		*theColor = theControl->energyColors[2];
	if (theEnergy <= 0.5) {
		f2 = theEnergy * 2.0;
		f1 = 1.0 - f2;
		theColor->argb.a = f1 * theControl->energyColors[0].argb.a + f2 * theControl->energyColors[1].argb.a;
		theColor->argb.r = f1 * theControl->energyColors[0].argb.r + f2 * theControl->energyColors[1].argb.r;
		theColor->argb.g = f1 * theControl->energyColors[0].argb.g + f2 * theControl->energyColors[1].argb.g;
		theColor->argb.b = f1 * theControl->energyColors[0].argb.b + f2 * theControl->energyColors[1].argb.b;
		return;
	} else {
		f2 = (theEnergy - 0.5) * 2.0;
		f1 = 1.0 - f2;
		theColor->argb.a = f1 * theControl->energyColors[1].argb.a + f2 * theControl->energyColors[2].argb.a;
		theColor->argb.r = f1 * theControl->energyColors[1].argb.r + f2 * theControl->energyColors[2].argb.r;
		theColor->argb.g = f1 * theControl->energyColors[1].argb.g + f2 * theControl->energyColors[2].argb.g;
		theColor->argb.b = f1 * theControl->energyColors[1].argb.b + f2 * theControl->energyColors[2].argb.b;
	}
}  /* EnergyColor */

static ParticlePtr FindParticle(long n)
{
	return particleData + n;
}  /* FindParticle */

static void AddParticle(ParticlePtr theParticle)
{
	ParticlePtr newParticle;

	if (theControl->firstInactive >= 0) {
		newParticle = FindParticle(theControl->firstInactive);
		theParticle->link = theControl->firstActive;
		theControl->firstActive = theControl->firstInactive;
		theControl->firstInactive = newParticle->link;
		*newParticle = *theParticle;
		theControl->particleCount++;
	}
}  /* AddParticle */

static void DeleteParticle(long theParticleIndex, long lastParticleIndex)
{
	ParticlePtr theParticle;

	theParticle = FindParticle(theParticleIndex);
	if (lastParticleIndex >= 0)
		FindParticle(lastParticleIndex)->link = theParticle->link;
	else
		theControl->firstActive = theParticle->link;
	theParticle->link = theControl->firstInactive;
	theControl->firstInactive = theParticleIndex;
	theControl->particleCount--;
}  /* DeleteParticle */

/* Calculate a new particle postion after frameCount frames with a duration of st time per frame.  The first */
/* frame has a duration of dt to allow for particles which are not created at an exact frame time. */

static boolean CalcParticle(ParticlePtr theParticle, long frameCount, double st, double dt)
{
	double bounceTime;
	double speed, newEnergy;
	boolean resultFlag = 0;

	/* A particle may have an zero delta time on the first frame */
	
	if (dt <= 0.0) {
		dt = st;
		frameCount--;
	}
	while (frameCount > 0)

		/* If the particle's energy has run out, kill it */

		if (theParticle->energy <= 0.0) {
			resultFlag = 0;
			frameCount = 0;
		} else {
	
			/* Compute the particle's current energy */
	
			newEnergy = theParticle->energy - dt * theParticle->decay;
			if (newEnergy < 0.0) {
				dt = dt * theParticle->energy / (theParticle->energy - newEnergy);
				theParticle->energy = 0.0;
			} else
				theParticle->energy = newEnergy;
	
			/* Compute the particle's color based upon its energy */
	
			EnergyColor(theParticle->energy, &theParticle->color);
			theParticle->xOld = theParticle->xPos;
			theParticle->yOld = theParticle->yPos;
			theParticle->zOld = theParticle->zPos;
	
			/* Slow down the particle's speed by the air resistance */
	
			speed = 1.0 - theControl->airResistance * dt;
			theParticle->xDir *= speed;
			theParticle->yDir *= speed;
			theParticle->zDir *= speed;
	
			/* Adjust the particle's motion vector for gravity */
	
			bounceTime = dt * theControl->gravityYDir;
			if ((ABS(theParticle->yDir) > bounceTime) || (theParticle->yPos - theControl->groundLevel > bounceTime))
				theParticle->yDir += bounceTime;
			theParticle->xDir += dt * theControl->gravityXDir;
			theParticle->zDir += dt * theControl->gravityZDir;
	
			/* Move the particle by adding the product of its motion vector and the time duration */
	
			theParticle->xPos += dt * theParticle->xDir;
			theParticle->yPos += dt * theParticle->yDir;
			theParticle->zPos += dt * theParticle->zDir;
	
			/* Compute the bounce if the particle's new y position is below the ground plane */
	
			if (theParticle->yPos < theControl->groundLevel) {
				double xBounce, zBounce;
				
				/* Slow down the particle by the ground's bounce friction */
	
				theParticle->xDir *= theControl->bounceFriction;
				theParticle->yDir *= -theControl->bounceFactor;
				theParticle->zDir *= theControl->bounceFriction;
	
				/* Compute the moment in time of the bounce and the impact point */
	
				bounceTime = (theParticle->yOld - theControl->groundLevel) / (theParticle->yOld - theParticle->yPos);
				xBounce = theParticle->xOld + (theParticle->xPos - theParticle->xOld) * bounceTime;
				zBounce = theParticle->zOld + (theParticle->zPos - theParticle->zOld) * bounceTime;
				dt *= 1.0 - bounceTime;
	
				/* The particle is moved to its new location after the bounce */
	
				theParticle->xPos = xBounce + theParticle->xDir * dt;
				theParticle->yPos = theControl->groundLevel + theParticle->yDir * dt;
				theParticle->zPos = zBounce + theParticle->zDir * dt;
			}
			dt = st;
			frameCount--;
			resultFlag = 1;
		}

	if (StatusTime())
		UpdateStatus(theStatusRec.theStatusPos, theStatusRec.theStatusValue);
	return resultFlag;
}  /* CalcParticle */

static void CalcParticles(long numFrames, double dt)
{
	long lastParticle, nextParticle, particleIndex;
	ParticlePtr theParticle;

	if (numFrames > 0) {
		lastParticle = -1;
		particleIndex = theControl->firstActive;
		while ((particleIndex >= 0) && (result == 0))
		{
			theParticle = FindParticle(particleIndex);
			if (CalcParticle(theParticle, numFrames, dt, dt))
			{
				lastParticle = particleIndex;
				particleIndex = theParticle->link;
			} else {
				nextParticle = theParticle->link;
				DeleteParticle(particleIndex, lastParticle);
				particleIndex = nextParticle;
			}
		}
	}
}  /* CalcParticles */

static void CreateParticles(long frame, long numFrames, double time, double dt)
{
	long count;
	double factor, theta;
	ParticleRec theParticle;

	if (theControl->firstInactive >= 0) {
	
		/* Initialize the random number seed based on the current frame number */
	
		SeedRandomCache(frame + theControl->particleSeed);
		
		if (theControl->emitStart == theControl->emitEnd)
	
			/* If the particle creation duration is zero, emit a steady stream of particles */
	
			count = (theControl->emitFrequency + theControl->emitErr * RandomFactor()) * dt + RandomNum();
		else if ((theControl->emitStart < time + dt) && (theControl->emitEnd >= time)) {
	
			/* If the current time duration overlaps the particle creation duration, create a random number of particles */
			/* based upon the overlapping time duration */
	
			if (theControl->emitStart > time)
				dt -= (theControl->emitStart - time);
			if (theControl->emitEnd - theControl->emitStart < dt)
				dt = theControl->emitEnd - theControl->emitStart;
			count = (theControl->emitFrequency + theControl->emitErr * RandomFactor()) * dt + RandomNum();
		} else

			/* if the current time duration does not overlap the particle creation duration, do not create any particles */

			count = 0;
	
		/* Create the particles */
	
		while ((count > 0) && (theControl->firstInactive >= 0) && (result == 0)) {

			/* Compute the azimuth and angle for the particle direction vector */
	
			factor = FABS(theControl->initAngle + theControl->initAngleErr * RandomFactor());
			while (factor > 180.0)
				factor = FABS(factor - 360.0);
			theParticle.zPos = 1.0 - factor * 0.01111111111;
			factor = SQRT(1.0 - SQR(theParticle.zPos));
			theta = RandomFactor() * PI;
			theParticle.yPos = COS(theta) * factor;
			theParticle.xPos = SIN(theta) * factor;
	
			/* Rotate the particle direction vector with the particle direction matrix */
	
			theParticle.xDir = theParticle.xPos * theData->directionMatrix.m.m00 + theParticle.yPos * theData->directionMatrix.m.m10 + theParticle.zPos * theData->directionMatrix.m.m20;
			theParticle.yDir = theParticle.xPos * theData->directionMatrix.m.m01 + theParticle.yPos * theData->directionMatrix.m.m11 + theParticle.zPos * theData->directionMatrix.m.m21;
			theParticle.zDir = theParticle.xPos * theData->directionMatrix.m.m02 + theParticle.yPos * theData->directionMatrix.m.m12 + theParticle.zPos * theData->directionMatrix.m.m22;

			/* Scale the particle direction vector with the particle velocity */
	
			factor = theControl->initVelocity + theControl->initVelocityErr * RandomFactor();
			theParticle.xDir *= factor;
			theParticle.yDir *= factor;
			theParticle.zDir *= factor;
	
			/* Compute the particle's starting location */
	
			theParticle.xPos = theControl->initXPos + theControl->initXPosErr * RandomFactor();
			theParticle.yPos = theControl->initYPos + theControl->initYPosErr * RandomFactor();
			theParticle.zPos = theControl->initZPos + theControl->initZPosErr * RandomFactor();
			theParticle.xOld = theParticle.xPos;
			theParticle.yOld = theParticle.yPos;
			theParticle.zOld = theParticle.zPos;
	
			/* Compute the particle's lifetime energy & decay rate */
	
			theParticle.energy = 1.0;
			theParticle.decay = 1.0 / (theControl->lifeTime + theControl->lifeTimeErr * RandomFactor());
	
			/* If the particle is above the ground level and does not die right away, add it to the active list */
	
			if ((theParticle.yPos >= theControl->groundLevel) && CalcParticle(&theParticle, numFrames, dt, dt * RandomNum()))
				AddParticle(&theParticle);
			count--;
		}
	}
}  /* CreateParticles */

static void InitParticles(void)
{
	double xDir, yDir, zDir, factor;
	
	/* Precompute the particle direction matrix */

	xDir = theControl->initXDir;
	yDir = theControl->initYDir;
	zDir = theControl->initZDir;
	factor = SQR(xDir) + SQR(yDir) + SQR(zDir);
	if (factor > 0.0) {
		factor = 1.0 / SQRT(factor);
		xDir *= factor;
		yDir *= factor;
		zDir *= factor;
	
		factor = 1.0 - SQR(zDir);
		if (factor > 0.0) {
			factor = SQRT(factor);
		
			theData->directionMatrix.m.m12 = -factor;
			factor = 1.0 / factor;
		
			theData->directionMatrix.m.m00 = yDir * factor;
			theData->directionMatrix.m.m01 = -xDir * factor;
			theData->directionMatrix.m.m02 = 0.0;
		
			factor *= zDir;
			theData->directionMatrix.m.m10 = xDir * factor;
			theData->directionMatrix.m.m11 = yDir * factor;
		
			theData->directionMatrix.m.m20 = xDir;
			theData->directionMatrix.m.m21 = yDir;
		} else
			InitMatrix3(&theData->directionMatrix);
		theData->directionMatrix.m.m22 = zDir;
	} else
		InitMatrix3(&theData->directionMatrix);
}  /* InitParticles */

static void AgeParticles(void)
{
	long frame, start, count;
	double time;

	/* Compute the frame and time of the oldest living particle */

	count = floor((theControl->lifeTime + theControl->lifeTimeErr) / theData->frameTimeDelta) + 1;
	frame = theData->frameIndex - count;
	time = theData->frameTime - count * theData->frameTimeDelta;
	start = frame;

	/* Setup the status box */

	theStatusRec.theReference = hostReference;
	theStatusRec.theStatusPos = 0.0;
	PStrCpy(theStatusRec.theStatusUnit, PS("Frame"));
	PStrCpy(theStatusRec.theStatusInfo, PS("Aging Particles"));
	theStatusRec.theStatusMin = 0;
	theStatusRec.theStatusMax = count;
	theStatusRec.theStatusValue = 0;
	theStatusRec.theStatusTick = 0;
	UpdateStatus(0.0, 0);

	/* If there are no valid active particles, initialize the list, otherwise update the list to the present time */

	if ((theControl->lastFrameIndex < frame) || (theControl->lastFrameIndex > theData->frameIndex) || (theControl->lastFrameTimeDelta != theData->frameTimeDelta)) {
		InitParticleList(theControl, theData);
		theControl->lastFrameTimeDelta = theData->frameTimeDelta;
	} else {
		CalcParticles(theData->frameIndex - theControl->lastFrameIndex, theData->frameTimeDelta);
		frame = theControl->lastFrameIndex;
		time = theControl->lastFrameTime;
	}
	theControl->lastFrameIndex = theData->frameIndex;
	theControl->lastFrameTime = theData->frameTime;

	/* Create particles for every frame up to the present time */

	while ((frame < theData->frameIndex) && (result == 0)) {
		theStatusRec.theStatusPos = (frame - start + 1.0) / count;
		theStatusRec.theStatusValue = frame - start + 1;
		CreateParticles(frame, theData->frameIndex - frame, time, theData->frameTimeDelta);
		time += theData->frameTimeDelta;
		frame++;
	}

	/* Finish updating the status box */

	if (result == 0)
		UpdateStatus(1.0, count);
	else
		InitParticleList(theControl, theData);
}  /* AgeParticles */

static void OutputParticles(void)
{

	if ((theControl->particleCount > 0) && (result == 0)) {
		long count, particleIndex;
		ulong vertexFlags;
		
		theStatusRec.theReference = hostReference;
		theStatusRec.theStatusPos = 0.0;
		PStrCpy(theStatusRec.theStatusUnit, PS("Particle"));
		PStrCpy(theStatusRec.theStatusInfo, PS("Creating Particles"));
		theStatusRec.theStatusMin = 0;
		theStatusRec.theStatusMax = theControl->particleCount;
		theStatusRec.theStatusValue = 0;
		theStatusRec.theStatusTick = 0;
		UpdateStatus(0.0, 0);
				
		/* Particles do not create vertex colors or texture positions currently */
		
		vertexFlags = theData->vertexFlags;
		BCLR(vertexFlags, pluginVertexColorFlag);
		BCLR(vertexFlags, pluginVertexTextureFlag);

		AddGroup(NULL, vertexFlags);
		count = 1;
		particleIndex = theControl->firstActive;
		while (particleIndex >= 0 && result == 0) {
			ParticlePtr theParticle;

			if (StatusTime())
				UpdateStatus((double)count / theControl->particleCount, count);
			theParticle = FindParticle(particleIndex);
			AddFacet(theParticle->color, 1, AddBlurVertex(theParticle->xPos, theParticle->yPos, theParticle->zPos, theParticle->xOld, theParticle->yOld, theParticle->zOld), 0, 0, 0);
			particleIndex = theParticle->link;
			count++;
		}
		UpdateStatus(1.0, theControl->particleCount);
	}
}  /* OutputParticles */

/* Initialize the plugin's data structures and perform memory allocation */

static long DoInitialize(long theCommandRecSize, void* theCommandRec)
{
	PluginInitializePtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	InitGlobals(1);
	theParameters = (PluginInitializePtr)theCommandRec;
	if ((theCommandRecSize < sizeof(PluginInitializeRec)) || (theParameters->theInfo.theHostProc == NULL))
		theResult = pluginParameterErr;
	else
		theResult = CheckData(&theParameters->theInfo, 0, 1);
	return theResult;
}  /* DoInitialize */

/* Finish up anything which needs it.  This is the last time that the plugin will be called */

static long DoFinish(long theCommandRecSize, void* theCommandRec)
{
	PluginFinishPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginFinishPtr)theCommandRec;
	if ((theCommandRecSize < sizeof(PluginFinishRec)) || (theParameters->theInfo.theHostProc == NULL))
		theResult = pluginParameterErr;
	FinishUtils();
	return theResult;
}  /* DoFinish */

/* Get information about the plugin */

static long DoInformation(long theCommandRecSize, void* theCommandRec)
{
	PluginInformationPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginInformationPtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginInformationRec))
		theResult = pluginParameterErr;
	else {
		theParameters->thePluginType = pluginType;
		theParameters->thePluginVersion = pluginVersion;
		PStrCpy(theParameters->thePluginName, pluginName);
		PStrCpy(theParameters->thePluginCopyright, pluginCopyright);
		PStrCpy(theParameters->thePluginAuthor, pluginAuthor);
		PStrCpy(theParameters->thePluginDescription, pluginDescription);
		theParameters->thePluginPermission = pluginPermission;
	}
	return theResult;
}  /* DoInformation */

/* Check to see if a particular command is supported */

static long DoCheck(long theCommandRecSize, void* theCommandRec)
{
	PluginCheckPtr theParameters;
	
	theParameters = (PluginCheckPtr)theCommandRec;
	if (theCommandRecSize >= sizeof(PluginCheckRec))
		switch (theParameters->theCommand) {
			case pluginInitialize:
			case pluginFinish:
			case pluginInformation:
			case pluginCheck:
			case pluginDescribe:
			case pluginInterface:
			case pluginModelSetup:
			case pluginModelGINF:
			case pluginModelGATR:
			case pluginModelGBLR:
			case pluginModelTMAP:
			case pluginModelBMAP:
			case pluginModelRMAP:
			case pluginModelGenerate:
				return pluginNoErr;
		}
	return pluginParameterErr;
}  /* DoCheck */

/* Describe an error number */

static long DoDescribe(long theCommandRecSize, void* theCommandRec)
{
	PluginDescribePtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginDescribePtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginDescribeRec))
		theResult = pluginParameterErr;
	else
		DescribePluginError(theParameters->theResult, theParameters->theDescription);
	return theResult;
}  /* DoDescribe */

/* Display the plugin's user interface and edit theCommandRec */

static long DoInterface(long theCommandRecSize, void* theCommandRec)
{
	EI_ResourceFile *resourceFile;
	EI_Dialog *theDialog;
	boolean changeFlag;
	ControlDataPtr theControl;
	PluginInterfacePtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginInterfacePtr)theCommandRec;
	if ((theCommandRecSize < sizeof(PluginInterfaceRec)) || (theParameters->theInfo.theControlData == NULL) || (theParameters->theInfo.theControlDataSize < sizeof(ControlDataRec)))
		theResult = pluginParameterErr;
	else
		theResult = CheckData(&theParameters->theInfo, 0, 0);
	if (theResult == 0)
		resourceFile = OpenResourceFile("Particle.rsc");
	if (resourceFile == NULL)
		theResult = pluginResourceErr;
	if (theResult == 0) {
		theControl = (ControlDataPtr)theParameters->theInfo.theControlData;
		theDialog = EI_MakeDialog(resourceFile, interfaceDialogID);
		if (theDialog != NULL) {
			SetTheINum(theDialog, cMaxParticlesET, theControl->maxParticles);
			SetTheIReal(theDialog, cNewParticlesET, theControl->emitFrequency, 8);
			SetTheIReal(theDialog, cNewParticlesPlusMinusET, theControl->emitErr, 8);
			SetTheIReal(theDialog, cCreateFromTimeET, theControl->emitStart, 8);
			SetTheIReal(theDialog, cCreateToTimeET, theControl->emitEnd, 8);
			SetTheIReal(theDialog, cParticleLifeTimeET, theControl->lifeTime, 8);
			SetTheIReal(theDialog, cLifeTimePlusMinusET, theControl->lifeTimeErr, 8);
			SetTheIReal(theDialog, cXPositionET, theControl->initXPos, 8);
			SetTheIReal(theDialog, cYPositionET, theControl->initYPos, 8);
			SetTheIReal(theDialog, cZPositionET, theControl->initZPos, 8);
			SetTheIReal(theDialog, cXPlusMinusET, theControl->initXPosErr, 8);
			SetTheIReal(theDialog, cYPlusMinusET, theControl->initYPosErr, 8);
			SetTheIReal(theDialog, cZPlusMinusET, theControl->initZPosErr, 8);
			SetTheIReal(theDialog, cXDirectionET, theControl->initXDir, 8);
			SetTheIReal(theDialog, cYDirectionET, theControl->initYDir, 8);
			SetTheIReal(theDialog, cZDirectionET, theControl->initZDir, 8);
			SetTheIReal(theDialog, cDeviationAngleET, theControl->initAngle, 8);
			SetTheIReal(theDialog, cDeviationPlusMinusET, theControl->initAngleErr, 8);
			SetTheIReal(theDialog, cVelocityET, theControl->initVelocity, 8);
			SetTheIReal(theDialog, cVelocityPlusMinusET, theControl->initVelocityErr, 8);
			SetTheIReal(theDialog, cAirResistanceET, theControl->airResistance, 8);
			SetTheIReal(theDialog, cGravityXET, theControl->gravityXDir, 8);
			SetTheIReal(theDialog, cGravityYET, theControl->gravityYDir, 8);
			SetTheIReal(theDialog, cGravityZET, theControl->gravityZDir, 8);
			SetTheIReal(theDialog, cYGroundLevelET, theControl->groundLevel, 8);
			SetTheIReal(theDialog, cBounceFactorET, theControl->bounceFactor, 8);
			SetTheIReal(theDialog, cBounceFrictionET, theControl->bounceFriction, 8);
			SetColorButtonIDARGB(theDialog, cEnergyOneColor, &theControl->energyColors[0]);
			SetColorButtonIDARGB(theDialog, cEnergyHalfColor, &theControl->energyColors[1]);
			SetColorButtonIDARGB(theDialog, cEnergyZeroColor, &theControl->energyColors[2]);
			if (EI_ExecuteDialog(theDialog)) {
				changeFlag = 0;
				SetLongInt(&theControl->maxParticles, GetTheINum(theDialog, cMaxParticlesET), &changeFlag);
				SetDouble(&theControl->emitFrequency, GetTheIReal(theDialog, cNewParticlesET), &changeFlag);
				SetDouble(&theControl->emitErr, GetTheIReal(theDialog, cNewParticlesPlusMinusET), &changeFlag);
				SetDouble(&theControl->emitStart, GetTheIReal(theDialog, cCreateFromTimeET), &changeFlag);
				SetDouble(&theControl->emitEnd, GetTheIReal(theDialog, cCreateToTimeET), &changeFlag);
				SetDouble(&theControl->lifeTime, GetTheIReal(theDialog, cParticleLifeTimeET), &changeFlag);
				SetDouble(&theControl->lifeTimeErr, GetTheIReal(theDialog, cLifeTimePlusMinusET), &changeFlag);
				SetDouble(&theControl->initXPos, GetTheIReal(theDialog, cXPositionET), &changeFlag);
				SetDouble(&theControl->initYPos, GetTheIReal(theDialog, cYPositionET), &changeFlag);
				SetDouble(&theControl->initZPos, GetTheIReal(theDialog, cZPositionET), &changeFlag);
				SetDouble(&theControl->initXPosErr, GetTheIReal(theDialog, cXPlusMinusET), &changeFlag);
				SetDouble(&theControl->initYPosErr, GetTheIReal(theDialog, cYPlusMinusET), &changeFlag);
				SetDouble(&theControl->initZPosErr, GetTheIReal(theDialog, cZPlusMinusET), &changeFlag);
				SetDouble(&theControl->initXDir, GetTheIReal(theDialog, cXDirectionET), &changeFlag);
				SetDouble(&theControl->initYDir, GetTheIReal(theDialog, cYDirectionET), &changeFlag);
				SetDouble(&theControl->initZDir, GetTheIReal(theDialog, cZDirectionET), &changeFlag);
				SetDouble(&theControl->initAngle, GetTheIReal(theDialog, cDeviationAngleET), &changeFlag);
				SetDouble(&theControl->initAngleErr, GetTheIReal(theDialog, cDeviationPlusMinusET), &changeFlag);
				SetDouble(&theControl->initVelocity, GetTheIReal(theDialog, cVelocityET), &changeFlag);
				SetDouble(&theControl->initVelocityErr, GetTheIReal(theDialog, cVelocityPlusMinusET), &changeFlag);
				SetDouble(&theControl->airResistance, GetTheIReal(theDialog, cAirResistanceET), &changeFlag);
				SetDouble(&theControl->gravityXDir, GetTheIReal(theDialog, cGravityXET), &changeFlag);
				SetDouble(&theControl->gravityYDir, GetTheIReal(theDialog, cGravityYET), &changeFlag);
				SetDouble(&theControl->gravityZDir, GetTheIReal(theDialog, cGravityZET), &changeFlag);
				SetDouble(&theControl->groundLevel, GetTheIReal(theDialog, cYGroundLevelET), &changeFlag);
				SetDouble(&theControl->bounceFactor, GetTheIReal(theDialog, cBounceFactorET), &changeFlag);
				SetDouble(&theControl->bounceFriction, GetTheIReal(theDialog, cBounceFrictionET), &changeFlag);
				GetColorButtonIDARGB(theDialog, cEnergyOneColor, &theControl->energyColors[0]);
				GetColorButtonIDARGB(theDialog, cEnergyHalfColor, &theControl->energyColors[1]);
				GetColorButtonIDARGB(theDialog, cEnergyZeroColor, &theControl->energyColors[2]);
				if (changeFlag) {
					theControl->lastFrameIndex = 0x80000000;	/* Force an initialization of the particle list */
					SendModelReset(&theParameters->theInfo);
				}
			} else
				theResult = pluginCancelErr;
			EI_DestroyDialog(theDialog);
		}
		EI_CloseResourceFile(resourceFile);
	}
	return theResult;
}  /* DoInterface */

/* Setup the model generation variables */

static long DoSetup(long theCommandRecSize, void* theCommandRec)
{
	PluginModelSetupPtr theParameters;
	ControlDataPtr theControl;
	PluginDataPtr theData;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelSetupPtr)theCommandRec;
	if ((theCommandRecSize < sizeof(PluginModelSetupRec)) || (theParameters->theInfo.theHostProc == NULL))
		theResult = pluginParameterErr;
	else
		theResult = CheckData(&theParameters->theInfo, 0, 0);
	if (theResult == 0) {
		theControl = (ControlDataPtr)theParameters->theInfo.theControlData;
		theData = (PluginDataPtr)theParameters->theInfo.thePluginData;
		
		theData->theMatrix = theParameters->theGroupMatrix;
		theData->theBlurMatrix = theParameters->theGroupBlurMatrix;
	
		if (theData->frameTimeDelta != theParameters->theFrameTimeDelta) {
			theData->frameTimeDelta = theParameters->theFrameTimeDelta;
			theControl->lastFrameIndex = 0x80000000;	/* Force an initialization of the particle list */
		}
		theData->frameIndex = theParameters->theFrameIndex;
		theData->frameTime = theParameters->theFrameTime;
		SendModelReset(&theParameters->theInfo);
	}
	return theResult;
}  /* DoSetup */

/* Setup the group GINF record */

static long DoGINF(long theCommandRecSize, void* theCommandRec)
{
	PluginModelGINFPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelGINFPtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginModelGINFRec))
		theResult = pluginParameterErr;
	else
		theParameters->theChangeFlag = 0;
	return theResult;
}  /* DoGINF */

/* Setup the group GATR record */

static long DoGATR(long theCommandRecSize, void* theCommandRec)
{
	PluginModelGATRPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelGATRPtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginModelGATRRec))
		theResult = pluginParameterErr;
	else
		theParameters->theChangeFlag = 0;
	return theResult;
}  /* DoGATR */

/* Setup the group GBLR record */

static long DoGBLR(long theCommandRecSize, void* theCommandRec)
{
	PluginModelGBLRPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelGBLRPtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginModelGBLRRec))
		theResult = pluginParameterErr;
	else
		theParameters->theChangeFlag = 0;
	return theResult;
}  /* DoGBLR */

/* Setup the group TMAP record */

static long DoTMAP(long theCommandRecSize, void* theCommandRec)
{
	PluginModelTMAPPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelTMAPPtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginModelTMAPRec))
		theResult = pluginParameterErr;
	else
		theParameters->theChangeFlag = 0;
	return theResult;
}  /* DoTMAP */

/* Setup the group BMAP record */

static long DoBMAP(long theCommandRecSize, void* theCommandRec)
{
	PluginModelBMAPPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelBMAPPtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginModelBMAPRec))
		theResult = pluginParameterErr;
	else
		theParameters->theChangeFlag = 0;
	return theResult;
}  /* DoBMAP */

/* Setup the group RMAP record */

static long DoRMAP(long theCommandRecSize, void* theCommandRec)
{
	PluginModelRMAPPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelRMAPPtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginModelRMAPRec))
		theResult = pluginParameterErr;
	else
		theParameters->theChangeFlag = 0;
	return theResult;
}  /* DoRMAP */

/* Generate the coordinates and polygons for the group */

static long DoGenerate(long theCommandRecSize, void* theCommandRec)
{
	PluginModelGeneratePtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelGeneratePtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginModelGenerateRec))
		theResult = pluginParameterErr;
	else 
		theResult = CheckData(&theParameters->theInfo, 0, 0);
	if (theResult == 0) {
		theData = (PluginDataPtr)theParameters->theInfo.thePluginData;
		theControl = (ControlDataPtr)theParameters->theInfo.theControlData;
		particleData = (ParticlePtr)((char*)theData + sizeof(PluginDataRec));
		hostProc = theParameters->theInfo.theHostProc;
		hostReference = theParameters->theInfo.theHostReference;
		theData->vertexFlags = theParameters->theVertexFlags;
			
		/* Initialize the particle list */
	
		InitParticles();
	
		/* Create the particles in the list and age them to the present time */
	
		AgeParticles();
	
		/* Write out the particles */
	
		OutputParticles();
	}
	return theResult;
}  /* DoGenerate */

/* The main function is called by the host application for all plugin commands */

EXPDLL long Main(unsigned long theCommand, long theCommandRecSize, void* theCommandRec)
{
	long theResult;
	
	/* Initialize the plugin's global variables if they have not been initialize previously */

	InitGlobals(0);

	/* Perform the requested command */

	switch (theCommand) {
		case pluginInitialize:		theResult = DoInitialize(theCommandRecSize, theCommandRec);		break;
		case pluginFinish:			theResult = DoFinish(theCommandRecSize, theCommandRec);			break;
		case pluginInformation:		theResult = DoInformation(theCommandRecSize, theCommandRec);	break;
		case pluginCheck:			theResult = DoCheck(theCommandRecSize, theCommandRec);			break;
		case pluginDescribe:		theResult = DoDescribe(theCommandRecSize, theCommandRec);		break;
		case pluginInterface:		theResult = DoInterface(theCommandRecSize, theCommandRec);		break;
		case pluginModelSetup:		theResult = DoSetup(theCommandRecSize, theCommandRec);			break;
		case pluginModelGINF:		theResult = DoGINF(theCommandRecSize, theCommandRec);			break;
		case pluginModelGATR:		theResult = DoGATR(theCommandRecSize, theCommandRec);			break;
		case pluginModelGBLR:		theResult = DoGBLR(theCommandRecSize, theCommandRec);			break;
		case pluginModelTMAP:		theResult = DoTMAP(theCommandRecSize, theCommandRec);			break;
		case pluginModelBMAP:		theResult = DoBMAP(theCommandRecSize, theCommandRec);			break;
		case pluginModelRMAP:		theResult = DoRMAP(theCommandRecSize, theCommandRec);			break;
		case pluginModelGenerate:	theResult = DoGenerate(theCommandRecSize, theCommandRec);		break;
		default:					theResult = pluginCommandErr;									break;
	}
	return theResult;
}  /* Main */
