/*
 *		Copyright (c)1991-1997 Electric Image, Inc.
 *		This code is confidential and is a trade secret of Electric Image, Inc.
 */

#include "ColorSpline.h"
#include "EIShader.h"
#include "FPU.h"
#include "Matrix.h"
#include "Noise.h"
#include "PString.h"
#include "RoundUtils.h"
#include <stddef.h>
#include "System.h"

#define MIX(a, b, f) ((a) + (f) * ((b) - (a)))

#define shaderType				'WOOD'
#define shaderVersion			0x00000001
#define shaderRevVersion		0x01000000

#define defShaderDensity		10.0
#define defShaderFrequencies	1
#define defShaderTurbulence		0.3
#define defShaderNoiseFactor	0.5
#define defShaderVeinSpacing	1.0
#define defShaderVeinMin		0.8
#define defShaderVeinMax		0.83
#define defShaderColors			2
#define defShaderColor0A		255
#define defShaderColor0R		128
#define defShaderColor0G		51
#define defShaderColor0B		13
#define defShaderColor1A		255
#define defShaderColor1R		38
#define defShaderColor1G		13
#define defShaderColor1B		3
#define defShaderColorA			255
#define defShaderColorR			0
#define defShaderColorG			0
#define defShaderColorB			0

typedef struct {
	unsigned long shaderOwnerType;
	unsigned long shaderVersionNum;
	float shaderDensity;
	unsigned long shaderFrequencies;
	float shaderTurbulence;
	float shaderNoiseFactor;
	float shaderVeinSpacing;
	float shaderVeinMin;
	float shaderVeinMax;
	unsigned long shaderColors;
	ARGB shaderColor[10];
} ShaderInterfaceRec, *ShaderInterfacePtr;

typedef struct {
	float shaderDensity;
	unsigned long shaderFrequencies;
	float shaderTurbulence;
	float shaderNoiseFactor;
	float shaderVeinSpacing;
	float shaderVeinMin;
	float shaderVeinMax;
	unsigned long shaderColors;
	ARGB_Real shaderColor[10];
} ShaderDataRec, *ShaderDataPtr;

/* Shader function */

static void CalcWood(EIShaderShadeRec *theShadePtr)
{
	EIShaderMapPtr theMap = theShadePtr->shaderMap;
	ShaderDataPtr theShaderData = (ShaderDataPtr)theShadePtr->shaderData;
	float density = theShaderData->shaderDensity;
	float x = density * theMap->map_X0;
	float y = density * theMap->map_Y0;
	float z = density * theMap->map_Z0;
	float pixelSize = density * MIN3(theMap->map_DX, theMap->map_DY, theMap->map_DZ);
	float noiseFactor = theShaderData->shaderNoiseFactor;
	float offset = SQRT(SQR(y) + SQR(z));
	float turbulence = theShaderData->shaderTurbulence;
	float veinScale = theShaderData->shaderVeinSpacing;
	unsigned long detail = theShaderData->shaderFrequencies;
	
	float noiseScale = 6.0 * turbulence;
	float twice = pixelSize + pixelSize;
	float scale = 1.0;

	/* Prevent the texture from shifting while animating the turblence */
	
	offset -= noiseScale;

	/* Calculate cubic interpolated noise for the higher octaves */
	
	while ((detail--) && (scale > twice)) {
		offset += noiseScale * Noise(x, y, z);
		x += x;
		y += y;
		z += z;
		noiseScale *= noiseFactor;
		scale *= 0.5;
	}

    /* Gradual fade out of highest-frequency component near limit */

	if ((detail > 0) && (scale > pixelSize))
        offset += ((scale / pixelSize) - 1.0) * noiseScale * Noise(x, y, z);
	
	/* Calculate the wood color function */
	
	offset *= veinScale;
	offset -= FLOOR(offset);
	offset = SmoothStep(0.0, theShaderData->shaderVeinMin, offset) - SmoothStep(theShaderData->shaderVeinMax, 1.0, offset);
	ColorSpline(offset, theShaderData->shaderColors, theShaderData->shaderColor, &theShadePtr->shaderColor);
}  /* CalcWood */

/* Static functions */

static void DefaultShaderInterface(ShaderInterfacePtr theShaderInterface)
{
	long i;
	
	theShaderInterface->shaderOwnerType = shaderType;
	theShaderInterface->shaderVersionNum = shaderVersion;
	theShaderInterface->shaderDensity = defShaderDensity;
	theShaderInterface->shaderFrequencies = defShaderFrequencies;
	theShaderInterface->shaderTurbulence = defShaderTurbulence;
	theShaderInterface->shaderNoiseFactor = defShaderNoiseFactor;
	theShaderInterface->shaderVeinSpacing = defShaderVeinSpacing;
	theShaderInterface->shaderVeinMin = defShaderVeinMin;
	theShaderInterface->shaderVeinMax = defShaderVeinMax;
	theShaderInterface->shaderColors = defShaderColors;
	theShaderInterface->shaderColor[0].argb.a = defShaderColor0A;
	theShaderInterface->shaderColor[0].argb.r = defShaderColor0R;
	theShaderInterface->shaderColor[0].argb.g = defShaderColor0G;
	theShaderInterface->shaderColor[0].argb.b = defShaderColor0B;
	theShaderInterface->shaderColor[1].argb.a = defShaderColor1A;
	theShaderInterface->shaderColor[1].argb.r = defShaderColor1R;
	theShaderInterface->shaderColor[1].argb.g = defShaderColor1G;
	theShaderInterface->shaderColor[1].argb.b = defShaderColor1B;
	for (i = 2; i < 10; i++) {
		theShaderInterface->shaderColor[i].argb.a = defShaderColorA;
		theShaderInterface->shaderColor[i].argb.r = defShaderColorR;
		theShaderInterface->shaderColor[i].argb.g = defShaderColorG;
		theShaderInterface->shaderColor[i].argb.b = defShaderColorB;
	}
}  /* DefaultShaderInterface */

static void FixShaderInterface(ShaderInterfacePtr theShaderInterface)
{
	theShaderInterface;
	
	FIXFIELD(theShaderInterface->shaderOwnerType);
	FIXFIELD(theShaderInterface->shaderVersionNum);
	FIXFIELD(theShaderInterface->shaderDensity);
	FIXFIELD(theShaderInterface->shaderFrequencies);
	FIXFIELD(theShaderInterface->shaderTurbulence);
	FIXFIELD(theShaderInterface->shaderNoiseFactor);
	FIXFIELD(theShaderInterface->shaderVeinSpacing);
	FIXFIELD(theShaderInterface->shaderVeinMin);
	FIXFIELD(theShaderInterface->shaderVeinMax);
	FIXFIELD(theShaderInterface->shaderColors);
	
}  /* FixShaderInterface */

static void GetShaderInterface(void* theInterface, long theSize, ShaderInterfacePtr theShaderInterface)
{
	/* Copy the shader interface data */
	
	if (theSize == sizeof(ShaderInterfaceRec))
		*theShaderInterface = *(ShaderInterfacePtr)theInterface;
	else {
		theShaderInterface->shaderOwnerType = 0;
		theShaderInterface->shaderVersionNum = 0;
	}
	
	/* Validate the shader interface data */
	
	FixShaderInterface(theShaderInterface);
	if (theShaderInterface->shaderOwnerType != shaderType ||
		theShaderInterface->shaderVersionNum != shaderVersion)
		DefaultShaderInterface(theShaderInterface);
}  /* GetShaderInterface */

/* Export functions */

long EIShaderInformation(EIShaderInformationRec *theInformationPtr)
{
	/* Setup the shader feature flags */
	
	BSET(theInformationPtr->shaderFeatureFlags, shaderAntialiasFlag);
	BSET(theInformationPtr->shaderFeatureFlags, shaderColorFlag);
	return shaderNoError;
}  /* EIShaderInformation */

long EIShaderInitialize(EIShaderInitializeRec *theInitializePtr)
{
	ShaderDataPtr theShaderData;
	
	/* Initialize the noise library */
	
	if (!InitNoise((NoiseMallocFunction)theInitializePtr->HostMalloc, NULL))
		return shaderMemoryError;
	
	/* Allocate the shader's private data record */
	
	theShaderData = (ShaderDataPtr)theInitializePtr->HostMalloc(sizeof(ShaderDataRec));
	if (theShaderData) {
		ShaderInterfaceRec theShaderInterface;
		long i;
		
		/* Get the shader interface parameters */
		
		GetShaderInterface(theInitializePtr->shaderInterface, theInitializePtr->shaderInterfaceSize, &theShaderInterface);

		/* Setup the shader's private data */
		
		theShaderData->shaderDensity = theShaderInterface.shaderDensity;
		theShaderData->shaderFrequencies = theShaderInterface.shaderFrequencies;
		theShaderData->shaderTurbulence = theShaderInterface.shaderTurbulence;
		theShaderData->shaderNoiseFactor = theShaderInterface.shaderNoiseFactor;
		theShaderData->shaderVeinSpacing = 1.0 / theShaderInterface.shaderVeinSpacing;
		theShaderData->shaderVeinMin = MIN(theShaderInterface.shaderVeinMin, theShaderInterface.shaderVeinMax);
		theShaderData->shaderVeinMax = MAX(theShaderInterface.shaderVeinMin, theShaderInterface.shaderVeinMax);
		theShaderData->shaderColors = theShaderInterface.shaderColors;
		for (i = 0; i < 10; i++) {
			theShaderData->shaderColor[i].a = theShaderInterface.shaderColor[i].argb.a / 255.0;
			theShaderData->shaderColor[i].r = theShaderInterface.shaderColor[i].argb.r / 255.0;
			theShaderData->shaderColor[i].g = theShaderInterface.shaderColor[i].argb.g / 255.0;
			theShaderData->shaderColor[i].b = theShaderInterface.shaderColor[i].argb.b / 255.0;
		}
		
		/* Return the pointer to the shader's private data to the host */
		
		theInitializePtr->shaderData = theShaderData;
		return shaderNoError;
	} else
		return shaderMemoryError;
}  /* EIShaderInitialize */

long EIShaderFinish(EIShaderFinishRec *theFinishPtr)
{
	/* Get rid of the shader's private data */
	
	theFinishPtr->HostFree(theFinishPtr->shaderData);
	
	/* Finish the noise library */
	
	FinishNoise((NoiseFreeFunction)theFinishPtr->HostFree);
	return shaderNoError;
}  /* EIShaderFinish */

long EIShaderShade(EIShaderShadeRec *theShadePtr)
{
	if (BTST(theShadePtr->shaderFeatureFlags, shaderColorFlag))
		CalcWood(theShadePtr);
	return shaderNoError;
}  /* EIShaderShade */
