/*
 *		Copyright (c) 1987-1999 Play, Inc.
 *		This code is confidential and is a trade secret of Electric Image, Inc.
 */

#include "Noise.h"
#ifdef __MMX__
#include "MMX_Noise.h"
#endif
#include "FPU.h"
#include "RoundUtils.h"

/* Solid noise generator constants */

#define noiseP1			1847L
#define noiseP2			2801L
#define noiseP3			4561L
#define noiseP4			6311L
#define noisePhi		40503L
#define noisePhi1		(noisePhi * noiseP1)
#define noisePhi2		(noisePhi * noiseP2)
#define noisePhi3		(noisePhi * noiseP3)
#define noisePhi4		(noisePhi * noiseP4)

/* Cellular noise constants */

#define cellularPoints	1024
#define cellularDensity	3.0

typedef struct {
	short p;
	unsigned char m;
} CellularFeature;

/* Noise global variables */

float *noiseData;
static CellularFeature* cellularFeatures;

#define FETCH_NOISE(n, i) (*((n) + ((i) & noiseMask)))

/* Random number functions */

/* MacRandom is a replacement for the Mac's Random function.  It produces identical results */

short MacRandom(unsigned long *randomSeed)
{
	unsigned long seed = *randomSeed;
	unsigned long seed0 = 0x41A7L * (unsigned short)seed;
	unsigned long seed1 = 0x41A7L * (seed >> 16) + (seed0 >> 16);
	short result;

	seed = ((seed1 & 0x00007FFF) << 16) + ((seed1 + seed1) >> 16) + (unsigned short)seed0 - 0x7FFFFFFF;
	if ((long)seed < 0)
		seed += 0x7FFFFFFF;
	result = *randomSeed = seed;
	if (result == 0x8000)
		result = 0;
	return result;
}  /* MacRandom */

/* Perlin noise basis functions */

long InitNoise(NoiseMallocFunction noiseMalloc, unsigned long *noiseSeed)
{
	unsigned long seed = noiseSeed ? *noiseSeed : 1;
	unsigned long theSize = (noisePoints + 1) * sizeof(float);
	long i = 0;

	noiseData = (float*)noiseMalloc(theSize);
	if (!noiseData) 
		return 0;

#ifdef __MMX__
	if (!MMX_InitNoise((MMX_NoiseMallocFunction)noiseMalloc, seed))
		return 0;
#endif
	i = 0;
	while (i <= noisePoints) {
		RANDOMSEED(seed);
		noiseData[i++] = seed * 2.328306436E-10;			/* yields a number between 0 and <1 */
	}
	if (noiseSeed)
		*noiseSeed = seed;
	return 1;
}  /* InitNoise */

void FinishNoise(NoiseFreeFunction noiseFree)
{
#ifdef __MMX__
	MMX_FinishNoise((MMX_NoiseFreeFunction)noiseFree);
#endif
	if (noiseData) {
		noiseFree((void*)noiseData);
		noiseData = 0L;
	}
}  /* FinishNoise */

float Noise1D(float x)
{
	float *theData = noiseData;
	float x2, x1, x0;
	long xa, xb, xc;
	long xi;
	
	xi = FLOOR(x);
	x -= xi;

	x1 = x * x;
	x2 = 0.5 * x1;
	x1 = 0.5 + x - x1;
	x0 = 0.5 - x + x2;

	xa = xi * noisePhi1;
	xb = xa + noisePhi1;
	xc = xb + noisePhi1;

	return x0 * FETCH_NOISE(theData, xa) + x1 * FETCH_NOISE(theData, xb) + x2 * FETCH_NOISE(theData, xc);
}  /* Noise1D */

float Noise1D2(float x)
{
	float *theData = noiseData;
	float n;
	long xa, xb;
	long xi;
	
	xi = FLOOR(x);
	x -= xi;

	xa = xi * noisePhi1;
	xb = xa + noisePhi1;

	n = FETCH_NOISE(theData, xa);
	return n + x * (FETCH_NOISE(theData, xb) - n);
}  /* Noise1D2 */

float Noise2D(float x, float y)
{
	float *theData = noiseData;
	float x2, x1, x0, y2, y1, y0;
	long xa, xb, xc, ya, yb, yc;
	long xi, yi;
	
	xi = FLOOR(x);
	yi = FLOOR(y);
	x -= xi;
	y -= yi;

	x1 = x * x;
	y1 = y * y;

	x2 = 0.5 * x1;
	y2 = 0.5 * y1;

	x1 = 0.5 + x - x1;
	y1 = 0.5 + y - y1;

	x0 = 0.5 - x + x2;
	y0 = 0.5 - y + y2;

	xa = xi * noisePhi1;
	ya = yi * noisePhi2;

	xb = xa + noisePhi1;
	yb = ya + noisePhi2;

	xc = xb + noisePhi1;
	yc = yb + noisePhi2;

	y0 *= (x0 * FETCH_NOISE(theData, xa + ya) + x1 * FETCH_NOISE(theData, xb + ya) + x2 * FETCH_NOISE(theData, xc + ya));
	y1 *= (x0 * FETCH_NOISE(theData, xa + yb) + x1 * FETCH_NOISE(theData, xb + yb) + x2 * FETCH_NOISE(theData, xc + yb));
	y2 *= (x0 * FETCH_NOISE(theData, xa + yc) + x1 * FETCH_NOISE(theData, xb + yc) + x2 * FETCH_NOISE(theData, xc + yc));

	return	y0 + y1 + y2;
}  /* Noise2D */

float Noise2D2(float x, float y)
{
	float *theData = noiseData;
	float n1, n2;
	long xa, xb, ya, yb;
	long xi, yi;
	
	xi = FLOOR(x);
	yi = FLOOR(y);
	x -= xi;
	y -= yi;

	xa = xi * noisePhi1;
	ya = yi * noisePhi2;

	xb = xa + noisePhi1;
	yb = ya + noisePhi2;

	n1 = FETCH_NOISE(theData, xa + ya);
	n1 += x * (FETCH_NOISE(theData, xb + ya) - n1);

	n2 = FETCH_NOISE(theData, xa + yb);
	n2 += x * (FETCH_NOISE(theData, xb + yb) - n2);
	
	return n1 + y * (n2 - n1);
}  /* Noise2D2 */

void DNoise2D(float x, float y, float* aNum, float* bNum)
{
	float *theData = noiseData;
	float x2, x1, x0, y2, y1, y0;
	float xd1, xd0, yd1, yd0;
	float p0, p1, p2;
	long xa, xb, xc, ya, yb, yc;
	long xi, yi;
	
	xi = FLOOR(x);
	yi = FLOOR(y);
	x -= xi;
	y -= yi;

	x1 = x * x;
	y1 = y * y;

	x2 = 0.5 * x1;
	y2 = 0.5 * y1;

	x1 = 0.5 + x - x1;
	y1 = 0.5 + y - y1;

	x0 = 0.5 - x + x2;
	y0 = 0.5 - y + y2;

	xd1 = 1.0 - x - x;
	yd1 = 1.0 - y - y;

	xd0 = x - 1.0;
	yd0 = y - 1.0;

	xa = xi * noisePhi1;
	ya = yi * noisePhi2;

	xb = xa + noisePhi1;
	yb = ya + noisePhi2;

	xc = xb + noisePhi1;
	yc = yb + noisePhi2;
	
	p0 = FETCH_NOISE(theData, xa + ya);
	p1 = FETCH_NOISE(theData, xb + ya);
	p2 = FETCH_NOISE(theData, xc + ya);
	
	y0 *= (xd0 * p0 + xd1 * p1 + x * p2);
	yd0 *= (x0 * p0 + x1 * p1 + x2 * p2);
	
	p0 = FETCH_NOISE(theData, xa + yb);
	p1 = FETCH_NOISE(theData, xb + yb);
	p2 = FETCH_NOISE(theData, xc + yb);

	y1 *= (xd0 * p0 + xd1 * p1 + x * p2);
	yd1 *= (x0 * p0 + x1 * p1 + x2 * p2);
	
	p0 = FETCH_NOISE(theData, xa + yc);
	p1 = FETCH_NOISE(theData, xb + yc);
	p2 = FETCH_NOISE(theData, xc + yc);
	
	y2 *= (xd0 * p0 + xd1 * p1 + x * p2);
	y *= (x0 * p0 + x1 * p1 + x2 * p2);
	
	*aNum = y0 + y1 + y2;
	*bNum = yd0 + yd1 + y;
}  /* DNoise2D */

float DNoise2D2(float x, float y, float* aNum, float* bNum)
{
	float *theData = noiseData;
	float n0, n1, n2;
	float x2, x1, x0, y2, y1, y0;
	float xd1, xd0, yd1, yd0;
	float p0, p1, p2;
	long xa, xb, xc, ya, yb, yc;
	long xi, yi;
	
	xi = FLOOR(x);
	yi = FLOOR(y);
	x -= xi;
	y -= yi;

	x1 = x * x;
	y1 = y * y;

	x2 = 0.5 * x1;
	y2 = 0.5 * y1;

	x1 = 0.5 + x - x1;
	y1 = 0.5 + y - y1;

	x0 = 0.5 - x + x2;
	y0 = 0.5 - y + y2;

	xd1 = 1.0 - x - x;
	yd1 = 1.0 - y - y;

	xd0 = x - 1.0;
	yd0 = y - 1.0;

	xa = xi * noisePhi1;
	ya = yi * noisePhi2;

	xb = xa + noisePhi1;
	yb = ya + noisePhi2;

	xc = xb + noisePhi1;
	yc = yb + noisePhi2;
	
	p0 = FETCH_NOISE(theData, xa + ya);
	p1 = FETCH_NOISE(theData, xb + ya);
	p2 = FETCH_NOISE(theData, xc + ya);

	*aNum = y0 * (xd0 * p0 + xd1 * p1 + x * p2);
	n0 = x0 * p0 + x1 * p1 + x2 * p2;

	p0 = FETCH_NOISE(theData, xa + yb);
	p1 = FETCH_NOISE(theData, xb + yb);
	p2 = FETCH_NOISE(theData, xc + yb);

	*aNum += y1 * (xd0 * p0 + xd1 * p1 + x * p2);
	n1 = x0 * p0 + x1 * p1 + x2 * p2;

	p0 = FETCH_NOISE(theData, xa + yc);
	p1 = FETCH_NOISE(theData, xb + yc);
	p2 = FETCH_NOISE(theData, xc + yc);
	
	*aNum += y2 * (xd0 * p0 + xd1 * p1 + x * p2);
	n2 = x0 * p0 + x1 * p1 + x2 * p2;

	*bNum = yd0 * n0 + yd1 * n1 + y * n2;
	
	return y0 * n0 + y1 * n1 + y2 * n2;
}  /* DNoise2D2 */

float Noise(float x, float y, float z)
{
#ifdef __MMX__
	long noise = MMX_Noise(x * 65536.0, y * 65536.0, z * 65536.0);
	
	MMX_EMMS();
	return noise * (1.0 / 32767);
#else
	float *theData = noiseData;
	float x0, x1, x2, y0, y1, y2, z0, z1, z2;
	long xa, xb, xc, ya, yb, yc, za, zb, zc;
	long aa, ab, ac, ba, bb, bc, ca, cb, cc;
	long xi, yi, zi;
	
	xi = FLOOR(x);
	yi = FLOOR(y);
	zi = FLOOR(z);

	x -= xi;
	y -= yi;
	z -= zi;
	
	x1 = x * x;
	y1 = y * y;
	z1 = z * z;

	x2 = 0.5 * x1;
	y2 = 0.5 * y1;
	z2 = 0.5 * z1;

	x1 = 0.5 + x - x1;
	y1 = 0.5 + y - y1;
	z1 = 0.5 + z - z1;

	x0 = 0.5 - x + x2;
	y0 = 0.5 - y + y2;
	z0 = 0.5 - z + z2;

	xa = xi * noisePhi1;
	ya = yi * noisePhi2;
	za = zi * noisePhi3;

	xb = xa + noisePhi1;
	yb = ya + noisePhi2;
	zb = za + noisePhi3;

	xc = xb + noisePhi1;
	yc = yb + noisePhi2;
	zc = zb + noisePhi3;
	
	aa = xa + ya;
	ba = xb + ya;
	ca = xc + ya;
	ab = xa + yb;
	bb = xb + yb;
	cb = xc + yb;
	ac = xa + yc;
	bc = xb + yc;
	cc = xc + yc;

	z0 *=	(y0 * (x0 * FETCH_NOISE(theData, aa + za) + x1 * FETCH_NOISE(theData, ba + za) + x2 * FETCH_NOISE(theData, ca + za)) +
			 y1 * (x0 * FETCH_NOISE(theData, ab + za) + x1 * FETCH_NOISE(theData, bb + za) + x2 * FETCH_NOISE(theData, cb + za)) +
			 y2 * (x0 * FETCH_NOISE(theData, ac + za) + x1 * FETCH_NOISE(theData, bc + za) + x2 * FETCH_NOISE(theData, cc + za)));
	z1 *=	(y0 * (x0 * FETCH_NOISE(theData, aa + zb) + x1 * FETCH_NOISE(theData, ba + zb) + x2 * FETCH_NOISE(theData, ca + zb)) +
			 y1 * (x0 * FETCH_NOISE(theData, ab + zb) + x1 * FETCH_NOISE(theData, bb + zb) + x2 * FETCH_NOISE(theData, cb + zb)) +
			 y2 * (x0 * FETCH_NOISE(theData, ac + zb) + x1 * FETCH_NOISE(theData, bc + zb) + x2 * FETCH_NOISE(theData, cc + zb)));
	z2 *=	(y0 * (x0 * FETCH_NOISE(theData, aa + zc) + x1 * FETCH_NOISE(theData, ba + zc) + x2 * FETCH_NOISE(theData, ca + zc)) +
			 y1 * (x0 * FETCH_NOISE(theData, ab + zc) + x1 * FETCH_NOISE(theData, bb + zc) + x2 * FETCH_NOISE(theData, cb + zc)) +
			 y2 * (x0 * FETCH_NOISE(theData, ac + zc) + x1 * FETCH_NOISE(theData, bc + zc) + x2 * FETCH_NOISE(theData, cc + zc)));
	
	return z0 + z1 + z2;
#endif
}  /* Noise */

float Noise2(float x, float y, float z)
{
#ifdef __MMX__
	long noise = MMX_Noise2(x * 65536.0, y * 65536.0, z * 65536.0);
	
	MMX_EMMS();
	return noise * (1.0 / 32767);
#else
	float *theData = noiseData;
	float n1, n2, na, nb;
	long xa, xb, ya, yb, za, zb;
	long aa, ab, ba, bb;
	long xi, yi, zi;
	
	xi = FLOOR(x);
	yi = FLOOR(y);
	zi = FLOOR(z);
	x -= xi;
	y -= yi;
	z -= zi;

	xa = xi * noisePhi1;
	ya = yi * noisePhi2;
	za = zi * noisePhi3;

	xb = xa + noisePhi1;
	yb = ya + noisePhi2;
	zb = za + noisePhi3;

	aa = xa + ya;
	ba = xb + ya;
	ab = xa + yb;
	bb = xb + yb;

	n1 = FETCH_NOISE(theData, aa + za);
	n1 += x * (FETCH_NOISE(theData, ba + za) - n1);
	n2 = FETCH_NOISE(theData, ab + za);
	n2 += x * (FETCH_NOISE(theData, bb + za) - n2);
	na = n1 + y * (n2 - n1);
	
	n1 = FETCH_NOISE(theData, aa + zb);
	n1 += x * (FETCH_NOISE(theData, ba + zb) - n1);
	n2 = FETCH_NOISE(theData, ab + zb);
	n2 += x * (FETCH_NOISE(theData, bb + zb) - n2);
	nb = n1 + y * (n2 - n1);
	
	return na + z * (nb - na);
#endif
}  /* Noise2 */

float Noise4D(float x, float y, float z, float t)
{
	float *theData = noiseData;
	float x0, x1, x2, y0, y1, y2, z0, z1, z2, t0, t1, t2;
	long xa, xb, xc, ya, yb, yc, za, zb, zc, ta, tb, tc;
	long aa, ab, ac, ba, bb, bc, ca, cb, cc, tza, tzb,tzc;
	long xi, yi, zi, ti;
	
	xi = FLOOR(x);
	yi = FLOOR(y);
	zi = FLOOR(z);
	ti = FLOOR(t);
	x -= xi;
	y -= yi;
	z -= zi;
	t -= ti;

	x1 = x * x;
	y1 = y * y;
	z1 = z * z;
	t1 = t * t;

	x2 = 0.5 * x1;
	y2 = 0.5 * y1;
	z2 = 0.5 * z1;
	t2 = 0.5 * t1;

	x1 = 0.5 + x - x1;
	y1 = 0.5 + y - y1;
	z1 = 0.5 + z - z1;
	t1 = 0.5 + t - t1;

	x0 = 0.5 - x + x2;
	y0 = 0.5 - y + y2;
	z0 = 0.5 - z + z2;
	t0 = 0.5 - t + t2;

	xa = xi * noisePhi1;
	ya = yi * noisePhi2;
	za = zi * noisePhi3;
	ta = ti * noisePhi4;
	
	xb = xa + noisePhi1;
	yb = ya + noisePhi2;
	zb = za + noisePhi3;
	tb = ta + noisePhi4;

	xc = xb + noisePhi1;
	yc = yb + noisePhi2;
	zc = zb + noisePhi3;
	tc = tb + noisePhi4;
	
	aa = xa + ya;
	ba = xb + ya;
	ca = xc + ya;
	ab = xa + yb;
	bb = xb + yb;
	cb = xc + yb;
	ac = xa + yc;
	bc = xb + yc;
	cc = xc + yc;
	
	tza = za + ta;
	tzb = zb + ta;
	tzc = zc + ta;

	t =  z0 * (y0 * (x0 * FETCH_NOISE(theData, aa + tza) + x1 * FETCH_NOISE(theData, ba + tza) + x2 * FETCH_NOISE(theData, ca + tza)) +
			   y1 * (x0 * FETCH_NOISE(theData, ab + tza) + x1 * FETCH_NOISE(theData, bb + tza) + x2 * FETCH_NOISE(theData, cb + tza)) +
			   y2 * (x0 * FETCH_NOISE(theData, ac + tza) + x1 * FETCH_NOISE(theData, bc + tza) + x2 * FETCH_NOISE(theData, cc + tza)));
	t += z1 * (y0 * (x0 * FETCH_NOISE(theData, aa + tzb) + x1 * FETCH_NOISE(theData, ba + tzb) + x2 * FETCH_NOISE(theData, ca + tzb)) +
			   y1 * (x0 * FETCH_NOISE(theData, ab + tzb) + x1 * FETCH_NOISE(theData, bb + tzb) + x2 * FETCH_NOISE(theData, cb + tzb)) +
			   y2 * (x0 * FETCH_NOISE(theData, ac + tzb) + x1 * FETCH_NOISE(theData, bc + tzb) + x2 * FETCH_NOISE(theData, cc + tzb)));
	t += z2 * (y0 * (x0 * FETCH_NOISE(theData, aa + tzc) + x1 * FETCH_NOISE(theData, ba + tzc) + x2 * FETCH_NOISE(theData, ca + tzc)) +
			   y1 * (x0 * FETCH_NOISE(theData, ab + tzc) + x1 * FETCH_NOISE(theData, bb + tzc) + x2 * FETCH_NOISE(theData, cb + tzc)) +
			   y2 * (x0 * FETCH_NOISE(theData, ac + tzc) + x1 * FETCH_NOISE(theData, bc + tzc) + x2 * FETCH_NOISE(theData, cc + tzc)));
	t0 *= t;
	
	tza = za + tb;
	tzb = zb + tb;
	tzc = zc + tb;

	t =  z0 * (y0 * (x0 * FETCH_NOISE(theData, aa + tza) + x1 * FETCH_NOISE(theData, ba + tza) + x2 * FETCH_NOISE(theData, ca + tza)) +
			   y1 * (x0 * FETCH_NOISE(theData, ab + tza) + x1 * FETCH_NOISE(theData, bb + tza) + x2 * FETCH_NOISE(theData, cb + tza)) +
			   y2 * (x0 * FETCH_NOISE(theData, ac + tza) + x1 * FETCH_NOISE(theData, bc + tza) + x2 * FETCH_NOISE(theData, cc + tza)));
	t += z1 * (y0 * (x0 * FETCH_NOISE(theData, aa + tzb) + x1 * FETCH_NOISE(theData, ba + tzb) + x2 * FETCH_NOISE(theData, ca + tzb)) +
			   y1 * (x0 * FETCH_NOISE(theData, ab + tzb) + x1 * FETCH_NOISE(theData, bb + tzb) + x2 * FETCH_NOISE(theData, cb + tzb)) +
			   y2 * (x0 * FETCH_NOISE(theData, ac + tzb) + x1 * FETCH_NOISE(theData, bc + tzb) + x2 * FETCH_NOISE(theData, cc + tzb)));
	t += z2 * (y0 * (x0 * FETCH_NOISE(theData, aa + tzc) + x1 * FETCH_NOISE(theData, ba + tzc) + x2 * FETCH_NOISE(theData, ca + tzc)) +
			   y1 * (x0 * FETCH_NOISE(theData, ab + tzc) + x1 * FETCH_NOISE(theData, bb + tzc) + x2 * FETCH_NOISE(theData, cb + tzc)) +
			   y2 * (x0 * FETCH_NOISE(theData, ac + tzc) + x1 * FETCH_NOISE(theData, bc + tzc) + x2 * FETCH_NOISE(theData, cc + tzc)));
	t1 *= t;
	
	tza = za + tc;
	tzb = zb + tc;
	tzc = zc + tc;

	t =  z0 * (y0 * (x0 * FETCH_NOISE(theData, aa + tza) + x1 * FETCH_NOISE(theData, ba + tza) + x2 * FETCH_NOISE(theData, ca + tza)) +
			   y1 * (x0 * FETCH_NOISE(theData, ab + tza) + x1 * FETCH_NOISE(theData, bb + tza) + x2 * FETCH_NOISE(theData, cb + tza)) +
			   y2 * (x0 * FETCH_NOISE(theData, ac + tza) + x1 * FETCH_NOISE(theData, bc + tza) + x2 * FETCH_NOISE(theData, cc + tza)));
	t += z1 * (y0 * (x0 * FETCH_NOISE(theData, aa + tzb) + x1 * FETCH_NOISE(theData, ba + tzb) + x2 * FETCH_NOISE(theData, ca + tzb)) +
			   y1 * (x0 * FETCH_NOISE(theData, ab + tzb) + x1 * FETCH_NOISE(theData, bb + tzb) + x2 * FETCH_NOISE(theData, cb + tzb)) +
			   y2 * (x0 * FETCH_NOISE(theData, ac + tzb) + x1 * FETCH_NOISE(theData, bc + tzb) + x2 * FETCH_NOISE(theData, cc + tzb)));
	t += z2 * (y0 * (x0 * FETCH_NOISE(theData, aa + tzc) + x1 * FETCH_NOISE(theData, ba + tzc) + x2 * FETCH_NOISE(theData, ca + tzc)) +
			   y1 * (x0 * FETCH_NOISE(theData, ab + tzc) + x1 * FETCH_NOISE(theData, bb + tzc) + x2 * FETCH_NOISE(theData, cb + tzc)) +
			   y2 * (x0 * FETCH_NOISE(theData, ac + tzc) + x1 * FETCH_NOISE(theData, bc + tzc) + x2 * FETCH_NOISE(theData, cc + tzc)));
	t2 *= t;
	
	return t0 + t1 + t2;
}  /* Noise4D */

float Noise4D2(float x, float y, float z, float t)
{
	float *theData = noiseData;
	float n1, n2, na, nb, np, nq;
	long xa, xb, ya, yb, za, zb, ta, tb;
	long aa, ab, ba, bb, tza, tzb;
	long xi, yi, zi, ti;
	
	xi = FLOOR(x);
	yi = FLOOR(y);
	zi = FLOOR(z);
	ti = FLOOR(t);
	x -= xi;
	y -= yi;
	z -= zi;
	t -= ti;
		
	xa = xi * noisePhi1;
	ya = yi * noisePhi2;
	za = zi * noisePhi3;
	ta = ti * noisePhi4;
	
	xb = xa + noisePhi1;
	yb = ya + noisePhi2;
	zb = za + noisePhi3;
	tb = ta + noisePhi4;

	aa = xa + ya;
	ba = xb + ya;
	ab = xa + yb;
	bb = xb + yb;

	tza = za + ta;
	tzb = zb + ta;
	
	n1 = FETCH_NOISE(theData, aa + tza);
	n1 += x * (FETCH_NOISE(theData, ba + tza) - n1);
	n2 = FETCH_NOISE(theData, ab + tza);
	n2 += x * (FETCH_NOISE(theData, bb + tza) - n2);
	na = n1 + y * (n2 - n1);
	
	n1 = FETCH_NOISE(theData, aa + tzb);
	n1 += x * (FETCH_NOISE(theData, ba + tzb) - n1);
	n2 = FETCH_NOISE(theData, ab + tzb);
	n2 += x * (FETCH_NOISE(theData, bb + tzb) - n2);
	nb = n1 + y * (n2 - n1);

	np = na + z * (nb - na);

	tza = za + tb;
	tzb = zb + tb;

	n1 = FETCH_NOISE(theData, aa + tza);
	n1 += x * (FETCH_NOISE(theData, ba + tza) - n1);
	n2 = FETCH_NOISE(theData, ab + tza);
	n2 += x * (FETCH_NOISE(theData, bb + tza) - n2);
	na = n1 + y * (n2 - n1);
	
	n1 = FETCH_NOISE(theData, aa + tzb);
	n1 += x * (FETCH_NOISE(theData, ba + tzb) - n1);
	n2 = FETCH_NOISE(theData, ab + tzb);
	n2 += x * (FETCH_NOISE(theData, bb + tzb) - n2);
	nb = n1 + y * (n2 - n1);

	nq = na + z * (nb - na);

	return np + t * (nq - np);
}  /* Noise4D2 */

void DNoise(float x, float y, float z, float* aNum, float* bNum, float* cNum)
{
#ifdef __MMX__
	long a, b, c;
	
	MMX_DNoise2(x * 65536.0, y * 65536.0, z * 65536.0, &a, &b, &c);
	MMX_EMMS();
	*aNum = a * (1.0 / 32767.0);
	*bNum = b * (1.0 / 32767.0);
	*cNum = c * (1.0 / 32767.0);
	return noise * (1.0 / 32767.0);
#else
	float *theData = noiseData;
	float x2, x1, x0, y2, y1, y0, z2, z1, z0;
	float xd1, xd0, yd1, yd0, zd1, zd0;
	float n0, n1, n2;
	float p0, p1, p2;
	float sum;
	long xa, xb, xc, ya, yb, yc, za, zb, zc;
	long aa, ab, ac, ba, bb, bc, ca, cb, cc;
	long xi, yi, zi;
	
	xi = FLOOR(x);
	yi = FLOOR(y);
	zi = FLOOR(z);
	x -= xi;
	y -= yi;
	z -= zi;

	x1 = x * x;
	y1 = y * y;
	z1 = z * z;

	x2 = 0.5 * x1;
	y2 = 0.5 * y1;
	z2 = 0.5 * z1;

	x1 = 0.5 + x - x1;
	y1 = 0.5 + y - y1;
	z1 = 0.5 + z - z1;

	x0 = 0.5 - x + x2;
	y0 = 0.5 - y + y2;
	z0 = 0.5 - z + z2;

	xd1 = 1.0 - x - x;
	yd1 = 1.0 - y - y;
	zd1 = 1.0 - z - z;

	xd0 = x - 1.0;
	yd0 = y - 1.0;
	zd0 = z - 1.0;

	xa = xi * noisePhi1;
	ya = yi * noisePhi2;
	za = zi * noisePhi3;

	xb = xa + noisePhi1;
	yb = ya + noisePhi2;
	zb = za + noisePhi3;

	xc = xb + noisePhi1;
	yc = yb + noisePhi2;
	zc = zb + noisePhi3;
	
	aa = xa + ya;
	ba = xb + ya;
	ca = xc + ya;
	ab = xa + yb;
	bb = xb + yb;
	cb = xc + yb;
	ac = xa + yc;
	bc = xb + yc;
	cc = xc + yc;
	
	p0 = FETCH_NOISE(theData, aa + za);
	p1 = FETCH_NOISE(theData, ba + za);
	p2 = FETCH_NOISE(theData, ca + za);

	sum =  y0 * (xd0 * p0 + xd1 * p1 + x * p2);
	n0 = x0 * p0 + x1 * p1 + x2 * p2;

	p0 = FETCH_NOISE(theData, ab + za);
	p1 = FETCH_NOISE(theData, bb + za);
	p2 = FETCH_NOISE(theData, cb + za);

	sum += y1 * (xd0 * p0 + xd1 * p1 + x * p2);
	n1 = x0 * p0 + x1 * p1 + x2 * p2;

	p0 = FETCH_NOISE(theData, ac + za);
	p1 = FETCH_NOISE(theData, bc + za);
	p2 = FETCH_NOISE(theData, cc + za);

	sum += y2 * (xd0 * p0 + xd1 * p1 + x * p2);
	n2 = x0 * p0 + x1 * p1 + x2 * p2;
	
	*aNum = z0 * sum;
	*bNum = z0 * (yd0 * n0 + yd1 * n1 + y * n2);
	*cNum = zd0 * (y0 * n0 + y1 * n1 + y2 * n2);
	
	p0 = FETCH_NOISE(theData, aa + zb);
	p1 = FETCH_NOISE(theData, ba + zb);
	p2 = FETCH_NOISE(theData, ca + zb);

	sum =  y0 * (xd0 * p0 + xd1 * p1 + x * p2);
	n0 = x0 * p0 + x1 * p1 + x2 * p2;

	p0 = FETCH_NOISE(theData, ab + zb);
	p1 = FETCH_NOISE(theData, bb + zb);
	p2 = FETCH_NOISE(theData, cb + zb);

	sum += y1 * (xd0 * p0 + xd1 * p1 + x * p2);
	n1 = x0 * p0 + x1 * p1 + x2 * p2;

	p0 = FETCH_NOISE(theData, ac + zb);
	p1 = FETCH_NOISE(theData, bc + zb);
	p2 = FETCH_NOISE(theData, cc + zb);

	sum += y2 * (xd0 * p0 + xd1 * p1 + x * p2);
	n2 = x0 * p0 + x1 * p1 + x2 * p2;
	
	*aNum += z1 * sum;
	*bNum += z1 * (yd0 * n0 + yd1 * n1 + y * n2);
	*cNum += zd1 * (y0 * n0 + y1 * n1 + y2 * n2);

	p0 = FETCH_NOISE(theData, aa + zc);
	p1 = FETCH_NOISE(theData, ba + zc);
	p2 = FETCH_NOISE(theData, ca + zc);
	
	sum =  y0 * (xd0 * p0 + xd1 * p1 + x * p2);
	n0 = x0 * p0 + x1 * p1 + x2 * p2;

	p0 = FETCH_NOISE(theData, ab + zc);
	p1 = FETCH_NOISE(theData, bb + zc);
	p2 = FETCH_NOISE(theData, cb + zc);

	sum += y1 * (xd0 * p0 + xd1 * p1 + x * p2);
	n1 = x0 * p0 + x1 * p1 + x2 * p2;

	p0 = FETCH_NOISE(theData, ac + zc);
	p1 = FETCH_NOISE(theData, bc + zc);
	p2 = FETCH_NOISE(theData, cc + zc);

	sum += y2 * (xd0 * p0 + xd1 * p1 + x * p2);
	n2 = x0 * p0 + x1 * p1 + x2 * p2;
	
	*aNum += z2 * sum;
	*bNum += z2 * (yd0 * n0 + yd1 * n1 + y * n2);
	*cNum += z * (y0 * n0 + y1 * n1 + y2 * n2);
#endif
}  /* DNoise */

float DNoise2(float x, float y, float z, float* aNum, float* bNum, float* cNum)
{
#ifdef __MMX__
	long a, b, c;
	long noise = MMX_DNoise2(x * 65536.0, y * 65536.0, z * 65536.0, &a, &b, &c);
	
	MMX_EMMS();
	*aNum = a * (1.0 / 32767.0);
	*bNum = b * (1.0 / 32767.0);
	*cNum = c * (1.0 / 32767.0);
	return noise * (1.0 / 32767.0);
#else
	float *theData = noiseData;
	float x2, x1, x0, y2, y1, y0, z2, z1, z0;
	float xd1, xd0, yd1, yd0, zd1, zd0;
	float n0, n1, n2;
	float p0, p1, p2;
	float sum, total;
	long xa, xb, xc, ya, yb, yc, za, zb, zc;
	long aa, ab, ac, ba, bb, bc, ca, cb, cc;
	long xi, yi, zi;
	
	xi = FLOOR(x);
	yi = FLOOR(y);
	zi = FLOOR(z);
	x -= xi;
	y -= yi;
	z -= zi;

	x1 = x * x;
	y1 = y * y;
	z1 = z * z;

	x2 = 0.5 * x1;
	y2 = 0.5 * y1;
	z2 = 0.5 * z1;

	x1 = 0.5 + x - x1;
	y1 = 0.5 + y - y1;
	z1 = 0.5 + z - z1;

	x0 = 0.5 - x + x2;
	y0 = 0.5 - y + y2;
	z0 = 0.5 - z + z2;

	xd1 = 1.0 - x - x;
	yd1 = 1.0 - y - y;
	zd1 = 1.0 - z - z;

	xd0 = x - 1.0;
	yd0 = y - 1.0;
	zd0 = z - 1.0;

	xa = xi * noisePhi1;
	ya = yi * noisePhi2;
	za = zi * noisePhi3;

	xb = xa + noisePhi1;
	yb = ya + noisePhi2;
	zb = za + noisePhi3;

	xc = xb + noisePhi1;
	yc = yb + noisePhi2;
	zc = zb + noisePhi3;
	
	aa = xa + ya;
	ba = xb + ya;
	ca = xc + ya;
	ab = xa + yb;
	bb = xb + yb;
	cb = xc + yb;
	ac = xa + yc;
	bc = xb + yc;
	cc = xc + yc;
	
	p0 = FETCH_NOISE(theData, aa + za);
	p1 = FETCH_NOISE(theData, ba + za);
	p2 = FETCH_NOISE(theData, ca + za);

	sum =  y0 * (xd0 * p0 + xd1 * p1 + x * p2);
	n0 = x0 * p0 + x1 * p1 + x2 * p2;

	p0 = FETCH_NOISE(theData, ab + za);
	p1 = FETCH_NOISE(theData, bb + za);
	p2 = FETCH_NOISE(theData, cb + za);

	sum += y1 * (xd0 * p0 + xd1 * p1 + x * p2);
	n1 = x0 * p0 + x1 * p1 + x2 * p2;

	p0 = FETCH_NOISE(theData, ac + za);
	p1 = FETCH_NOISE(theData, bc + za);
	p2 = FETCH_NOISE(theData, cc + za);

	sum += y2 * (xd0 * p0 + xd1 * p1 + x * p2);
	n2 = x0 * p0 + x1 * p1 + x2 * p2;
	
	*aNum = z0 * sum;
	*bNum = z0 * (yd0 * n0 + yd1 * n1 + y * n2);
	sum = y0 * n0 + y1 * n1 + y2 * n2;
	*cNum = zd0 * sum;
	total = z0 * sum;
	
	p0 = FETCH_NOISE(theData, aa + zb);
	p1 = FETCH_NOISE(theData, ba + zb);
	p2 = FETCH_NOISE(theData, ca + zb);

	sum =  y0 * (xd0 * p0 + xd1 * p1 + x * p2);
	n0 = x0 * p0 + x1 * p1 + x2 * p2;

	p0 = FETCH_NOISE(theData, ab + zb);
	p1 = FETCH_NOISE(theData, bb + zb);
	p2 = FETCH_NOISE(theData, cb + zb);

	sum += y1 * (xd0 * p0 + xd1 * p1 + x * p2);
	n1 = x0 * p0 + x1 * p1 + x2 * p2;

	p0 = FETCH_NOISE(theData, ac + zb);
	p1 = FETCH_NOISE(theData, bc + zb);
	p2 = FETCH_NOISE(theData, cc + zb);

	sum += y2 * (xd0 * p0 + xd1 * p1 + x * p2);
	n2 = x0 * p0 + x1 * p1 + x2 * p2;
	
	*aNum += z1 * sum;
	*bNum += z1 * (yd0 * n0 + yd1 * n1 + y * n2);
	sum = y0 * n0 + y1 * n1 + y2 * n2;
	*cNum += zd1 * sum;
	total += z1 * sum;

	p0 = FETCH_NOISE(theData, aa + zc);
	p1 = FETCH_NOISE(theData, ba + zc);
	p2 = FETCH_NOISE(theData, ca + zc);
	
	sum =  y0 * (xd0 * p0 + xd1 * p1 + x * p2);
	n0 = x0 * p0 + x1 * p1 + x2 * p2;

	p0 = FETCH_NOISE(theData, ab + zc);
	p1 = FETCH_NOISE(theData, bb + zc);
	p2 = FETCH_NOISE(theData, cb + zc);

	sum += y1 * (xd0 * p0 + xd1 * p1 + x * p2);
	n1 = x0 * p0 + x1 * p1 + x2 * p2;

	p0 = FETCH_NOISE(theData, ac + zc);
	p1 = FETCH_NOISE(theData, bc + zc);
	p2 = FETCH_NOISE(theData, cc + zc);

	sum += y2 * (xd0 * p0 + xd1 * p1 + x * p2);
	n2 = x0 * p0 + x1 * p1 + x2 * p2;
	
	*aNum += z2 * sum;
	*bNum += z2 * (yd0 * n0 + yd1 * n1 + y * n2);
	sum = y0 * n0 + y1 * n1 + y2 * n2;
	*cNum += z * sum;
	return total + z2 * sum;
#endif
}  /* DNoise2 */

/* Cellular noise basis functions */

long InitCellularNoise(NoiseMallocFunction noiseMalloc, unsigned long *noiseSeed)
{
	unsigned long theSize = cellularPoints * 2 * sizeof(CellularFeature);
	
	cellularFeatures = (CellularFeature*)noiseMalloc(theSize);
	if (cellularFeatures) {
		double density = cellularDensity;
		double featureProbability[10];
		double expDensity = 1.0 / EXP(density);
		double m = 1.0;
		double mFactoral = 1.0;
		double probability = 0.0;
		double probabilityStep = 1.0 / cellularPoints;
		unsigned long seed = noiseSeed ? *noiseSeed : 1;
		long i, j;
		
		/* Calculate the probability that m features will occur within a single unit cube */
		/* The number of features is artificialy limited to the range of 1 to 9 for */
		/* reasons of code optimization */
		
		featureProbability[0] = expDensity;
		for (i = 1; i <= 9; i++) {
			expDensity *= density;
			featureProbability[i] = featureProbability[i - 1] + expDensity / mFactoral;
			m += 1.0;
			mFactoral *= m;
		}
		
		/* Evenly distribute the feature points over a probability range of 0 to 1 */
		
		for (i = 0; i < cellularPoints; i++) {
			j = 1;
			while ((j < 9) && (probability > featureProbability[j]))
				j++;
			cellularFeatures[i].m = j;
			probability += probabilityStep;
		}
		
		/* Initialize the cellular permutation array */
		
		for (i = 0; i < cellularPoints; i++)
			cellularFeatures[i].p = i;
		for (i = 0; i < cellularPoints; i++) {
			long temp;
			
			j = RANDOMSEED(seed) & (cellularPoints - 1);
			temp = cellularFeatures[j].p;
			cellularFeatures[j].p = cellularFeatures[i].p;
			cellularFeatures[i].p = temp;
		}
		for (i = 0; i < cellularPoints; i++)
			cellularFeatures[i + cellularPoints] = cellularFeatures[i];
		
		if (noiseSeed)
			*noiseSeed = seed;
		return 1;
	} else
		return 0;
}  /* InitCellularNoise */

void FinishCellularNoise(NoiseFreeFunction noiseFree)
{
	if (cellularFeatures) {
		noiseFree((void*)cellularFeatures);
		cellularFeatures = 0L;
	}
}  /* FinishCellularNoise */

void SetupCellularDots(CellularDotInfo *theInfo)
{
	theInfo->cellDotCount = 0;
	theInfo->cellXI = 0x8000000;
	theInfo->cellYI = 0x8000000;
	theInfo->cellZI = 0x8000000;
}  /* SetupCellularDots */

void DoCellularDots(CellularDotData *f, unsigned long *n, long xi, long yi, long zi, float xr, float yr, float zr)
{
	long index;
	long m, i;
	unsigned long seed;

	f += *n;
	
	xi &= cellularPoints - 1;
	yi &= cellularPoints - 1;
	zi &= cellularPoints - 1;
	
	index = cellularFeatures[cellularFeatures[zi].p + yi].p + xi;
	m = cellularFeatures[index].m;
	seed = index;
	index <<= 4;
	
	*n += m;
	for (i = 0; i < m; i++) {
		f->vx = xr + RANDOMNUM(seed);
		f->vy = yr + RANDOMNUM(seed);
		f->vz = zr + RANDOMNUM(seed);
		f->index = RANDOMSEED(seed);
		f++;
	}
}  /* DoCellularDots */

void BuildCellularDots(CellularDotInfo *theInfo, float x, float y, float z)
{
	long xfi = FLOOR(x);
	long yfi = FLOOR(y);
	long zfi = FLOOR(z);
	
	/* Build a list of potential dots around the current cell */
	
	if ((xfi != theInfo->cellXI) || (yfi != theInfo->cellYI) || (zfi != theInfo->cellZI)) {
		float xfr = xfi - 1;
		float yfr = yfi - 1;
		float zfr = zfi - 1;
		float xr, yr, zr;
		long xi, yi, zi;
		
		/* Initialize the result feature array */
		
		theInfo->cellDotCount = 0;
		theInfo->cellXI = xfi;
		theInfo->cellYI = yfi;
		theInfo->cellZI = zfi;
		
		/* Compute the cells */
		
		for (xi = -1, xr = xfr; xi <= 1; xi++, xr += 1.0)
			for (yi = -1, yr = yfr; yi <= 1; yi++, yr += 1.0)
				for (zi = -1, zr = zfr; zi <= 1; zi++, zr += 1.0)
					DoCellularDots(theInfo->cellDotList, &theInfo->cellDotCount, xfi + xi, yfi + yi, zfi + zi, xr, yr, zr);
	}
}  /* BuildCellularDots */

static void DoCell(CellularData *f, long n, long xi, long yi, long zi, float x, float y, float z)
{
	long index;
	long m, i;
	unsigned long seed;

	xi &= cellularPoints - 1;
	yi &= cellularPoints - 1;
	zi &= cellularPoints - 1;
	
	index = cellularFeatures[cellularFeatures[zi].p + yi].p + xi;
	m = cellularFeatures[index].m;
	seed = index;
	index <<= 4;
	
	for (i = 0; i < m; i++) {
		float vx = RANDOMNUM(seed) - x;
		float vy = RANDOMNUM(seed) - y;
		float vz = RANDOMNUM(seed) - z;
		float distance = vx * vx + vy * vy + vz * vz;
		
		if (distance < f[n].distance) {
			long j = n;
		
			while (j && (distance < f[j - 1].distance)) {
				f[j] = f[j - 1];
				j--;
			}
			
			f[j].distance = distance;
			f[j].vx = vx;
			f[j].vy = vy;
			f[j].vz = vz;
			f[j].index = index + i;
		}
	}
}  /* DoCell */

static void DoCell2D(CellularData *f, long n, long xi, long yi, float x, float y)
{
	long index;
	long m, i;
	unsigned long seed;

	xi &= cellularPoints - 1;
	yi &= cellularPoints - 1;
	
	index = cellularFeatures[yi].p + xi;
	m = cellularFeatures[index].m;
	seed = index;
	index <<= 4;
	
	for (i = 0; i < m; i++) {
		float vx = RANDOMNUM(seed) - x;
		float vy = RANDOMNUM(seed) - y;
		float distance = vx * vx + vy * vy;
		
		if (distance < f[n].distance) {
			long j = n;
		
			while (j && (distance < f[j - 1].distance)) {
				f[j] = f[j - 1];
				j--;
			}
			
			f[j].distance = distance;
			f[j].vx = vx;
			f[j].vy = vy;
			f[j].index = index + i;
		}
	}
}  /* DoCell2D */

void CellularNoise(CellularData *f, unsigned long n, float xr, float yr, float zr)
{
	float *distance = &f[n].distance;
	float xp, xn, yp, yn, zp, zn;
	float xrp, xrn, yrp, yrn, zrp, zrn;
	long xip, xin, yip, yin, zip, zin;
	long xi = FLOOR(xr);
	long yi = FLOOR(yr);
	long zi = FLOOR(zr);
	long i;
	
	xr -= xi;
	yr -= yi;
	zr -= zi;

	xn = xr * xr;
	yn = yr * yr;
	zn = zr * zr;
	
	xp = 1.0 - xr;
	yp = 1.0 - yr;
	zp = 1.0 - zr;
	xp *= xp;
	yp *= yp;
	zp *= zp;
	
	xrn = xr + 1.0;
	yrn = yr + 1.0;
	zrn = zr + 1.0;

	xrp = xr - 1.0;
	yrp = yr - 1.0;
	zrp = zr - 1.0;

	xin = xi - 1;
	yin = yi - 1;
	zin = zi - 1;

	xip = xi + 1;
	yip = yi + 1;
	zip = zi + 1;
	
	/* Initialize the result feature array */
	
	for (i = 0; i <= n; i++) {
		f[i].distance = 100.0;
		f[i].index = 0;
	}
	
	/* Compute the central cell */
	
	DoCell(f, n, xi, yi, zi, xr, yr, zr);
	
	/* Center z cells */
	
	if (xn < *distance)
		DoCell(f, n, xin, yi, zi, xrn, yr, zr);
	if (xp < *distance)
		DoCell(f, n, xip, yi, zi, xrp, yr, zr);
	
	if (yn < *distance) {
		DoCell(f, n, xi, yin, zi, xr, yrn, zr);
		if (xn < *distance)
			DoCell(f, n, xin, yin, zi, xrn, yrn, zr);
		if (xp < *distance)
			DoCell(f, n, xip, yin, zi, xrp, yrn, zr);
	}
	if (yp < *distance) {
		DoCell(f, n, xi, yip, zi, xr, yrp, zr);
		if (xn < *distance)
			DoCell(f, n, xin, yip, zi, xrn, yrp, zr);
		if (xp < *distance)
			DoCell(f, n, xip, yip, zi, xrp, yrp, zr);
	}
	
	/* Negative z cells */
	
	if (zn < *distance) {
		DoCell(f, n, xi, yi, zin, xr, yr, zrn);

		if (xn < *distance)
			DoCell(f, n, xin, yi, zin, xrn, yr, zrn);
		if (xp < *distance)
			DoCell(f, n, xip, yi, zin, xrp, yr, zrn);

		if (yn < *distance) {
			DoCell(f, n, xi, yin, zin, xr, yrn, zrn);
			if (xn < *distance)
				DoCell(f, n, xin, yin, zin, xrn, yrn, zrn);
			if (xp < *distance)
				DoCell(f, n, xip, yin, zin, xrp, yrn, zrn);
		}
		if (yp < *distance) {
			DoCell(f, n, xi, yip, zin, xr, yrp, zrn);
			if (xn < *distance)
				DoCell(f, n, xin, yip, zin, xrn, yrp, zrn);
			if (xp < *distance)
				DoCell(f, n, xip, yip, zin, xrp, yrp, zrn);
		}
	}
	
	/* Positive z cells */
	
	if (zp < *distance) {
		DoCell(f, n, xi, yi, zip, xr, yr, zrp);
		
		if (xn < *distance)
			DoCell(f, n, xin, yi, zip, xrn, yr, zrp);
		if (xp < *distance)
			DoCell(f, n, xip, yi, zip, xrp, yr, zrp);
		
		if (yn < *distance) {
			DoCell(f, n, xi, yin, zip, xr, yrn, zrp);
			if (xn < *distance)
				DoCell(f, n, xin, yin, zip, xrn, yrn, zrp);
			if (xp < *distance)
				DoCell(f, n, xip, yin, zip, xrp, yrn, zrp);
		}
		if (yp < *distance) {
			DoCell(f, n, xi, yip, zip, xr, yrp, zrp);
			if (xn < *distance)
				DoCell(f, n, xin, yip, zip, xrn, yrp, zrp);
			if (xp < *distance)
				DoCell(f, n, xip, yip, zip, xrp, yrp, zrp);
		}
	}
}  /* CellularNoise */

void CellularNoise2D(CellularData *f, unsigned long n, float xr, float yr)
{
	float *distance = &f[n].distance;
	float xp, xn, yp, yn;
	float xrp, xrn, yrp, yrn;
	long xip, xin, yip, yin;
	long xi = FLOOR(xr);
	long yi = FLOOR(yr);
	long i;
	
	xr -= xi;
	yr -= yi;

	xn = xr * xr;
	yn = yr * yr;
	
	xp = 1.0 - xr;
	yp = 1.0 - yr;
	xp *= xp;
	yp *= yp;
	
	xrn = xr + 1.0;
	yrn = yr + 1.0;

	xrp = xr - 1.0;
	yrp = yr - 1.0;

	xin = xi - 1;
	yin = yi - 1;

	xip = xi + 1;
	yip = yi + 1;
	
	/* Initialize the result feature array */
	
	for (i = 0; i <= n; i++) {
		f[i].distance = 100.0;
		f[i].index = 0;
	}
	
	/* Compute the central cell */
	
	DoCell2D(f, n, xi, yi, xr, yr);
	
	/* Center z cells */
	
	if (xn < *distance)
		DoCell2D(f, n, xin, yi, xrn, yr);
	if (xp < *distance)
		DoCell2D(f, n, xip, yi, xrp, yr);
	
	if (yn < *distance) {
		DoCell2D(f, n, xi, yin, xr, yrn);
		if (xn < *distance)
			DoCell2D(f, n, xin, yin, xrn, yrn);
		if (xp < *distance)
			DoCell2D(f, n, xip, yin, xrp, yrn);
	}
	if (yp < *distance) {
		DoCell2D(f, n, xi, yip, xr, yrp);
		if (xn < *distance)
			DoCell2D(f, n, xin, yip, xrn, yrp);
		if (xp < *distance)
			DoCell2D(f, n, xip, yip, xrp, yrp);
	}
}  /* CellularNoise2D */

/* Fractal texture routines */

#ifdef __MMX__
float FractalNoise(float x, float y, float z, float noiseFactor, float noiseScale, long detail1, long detail2)
{
	long i = 0;
	long sum = 16384;
	long x2 = x * 65536.0;
	long y2 = y * 65536.0;
	long z2 = z * 65536.0;
	long noiseScale2 = noiseScale * 65536.0;
	long noiseFactor2 = noiseFactor * 65536.0;
	
	/* Calculate cubic interpolated noise for the higher octaves */
	
	if (detail1)
		while (i < detail1) {
			sum += ((MMX_Noise(x2, y2, z2) - 16384) * noiseScale2) / 65536;
			if (++i == detail2)
				break;
			x2 += x2;
			y2 += y2;
			z2 += z2;
			noiseScale2 = (noiseScale2 * noiseFactor2) / 65536;
		}
	
	/* Calculate linear interpolated noise for the lower octaves */
	
	if (i < detail2)
		while (1) {
			sum += ((MMX_Noise2(x2, y2, z2) - 16384) * noiseScale2) / 65536;
			if (++i == detail2)
				break;
			x2 += x2;
			y2 += y2;
			z2 += z2;
			noiseScale2 = (noiseScale2 * noiseFactor2) / 65536;
		}
	MMX_EMMS();
	
	/* Return the fractal noise value */
	
	return sum * (1.0 / 32767.0);
}  /* FractalNoise */
#else
float FractalNoise(float x, float y, float z, float noiseFactor, float noiseScale, long detail1, long detail2)
{
	float sum = 0.5;
	long i = 0;

	/* Calculate cubic interpolated noise for the higher octaves */
	
	if (detail1)
		while (i < detail1) {
			sum += (Noise(x, y, z) - 0.5) * noiseScale;
			if (++i == detail2)
				break;
			x += x;
			y += y;
			z += z;
			noiseScale *= noiseFactor;
		}
	
	/* Calculate linear interpolated noise for the lower octaves */
	
	if (i < detail2)
		while (1) {
			sum += (Noise2(x, y, z) - 0.5) * noiseScale;
			if (++i == detail2)
				break;
			x += x;
			y += y;
			z += z;
			noiseScale *= noiseFactor;
		}
	
	/* Return the fractal noise value */
	
	return sum;
}  /* FractalNoise */
#endif

float FractalNoise2D(float x, float y, float noiseFactor, float noiseScale, long detail1, long detail2)
{
	float sum = 0.5;
	long i = 0;

	/* Calculate cubic interpolated noise for the higher octaves */
	
	if (detail1)
		while (i < detail1) {
			sum += (Noise2D(x, y) - 0.5) * noiseScale;
			if (++i == detail2)
				break;
			x += x;
			y += y;
			noiseScale *= noiseFactor;
		}
	
	/* Calculate linear interpolated noise for the lower octaves */
	
	if (i < detail2)
		while (1) {
			sum += (Noise2D(x, y) - 0.5) * noiseScale;
			if (++i == detail2)
				break;
			x += x;
			y += y;
			noiseScale *= noiseFactor;
		}
	
	/* Return the fractal noise value */
	
	return sum;
}  /* FractalNoise2D */

float CellularFractalNoise(float x, float y, float z, float xOff, float yOff, float zOff, float noiseFactor, float noiseScale, long detail1, long detail2)
{
	float sum;
	long i = 0;
		
	/* Calculate Perlin noise for the lower octaves */
	
	if (detail1) {
		sum = 0.5;
		while (i < detail1) {
			sum += (Noise(x, y, z) - 0.5) * noiseScale;
			if (++i == detail2)
				break;
			x += x;
			y += y;
			z += z;
			noiseScale *= noiseFactor;
		}
	} else
		sum = 0.0;
		
	/* Calculate cellular noise for the higher octaves */
	
	if (i < detail2)
		while (1) {
			CellularData f;
			
			CellularNoise(&f, 0, x + xOff, y + yOff, z + zOff);
			sum += f.distance * noiseScale;
			if (++i == detail2)
				break;
			x += x;
			y += y;
			z += z;
			noiseScale *= noiseFactor;
		}
	
	/* Return the fractal noise value */
	
	return sum;
}  /* CellularFractalNoise */

float CellularFractalNoise2D(float x, float y, float xOff, float yOff, float noiseFactor, float noiseScale, long detail1, long detail2)
{
	float sum;
	long i = 0;

	if (detail1) {
		sum = 0.5;
		while (i < detail1) {
			sum += (Noise2D(x, y) - 0.5) * noiseScale;
			if (++i == detail2)
				break;
			x += x;
			y += y;
			noiseScale *= noiseFactor;
		}
	} else
		sum = 0.0;
		
	/* Calculate cubic interpolated noise for the higher octaves */
	
	if (i < detail2)
		while (1) {
			CellularData f;
			
			CellularNoise2D(&f, 0, x + xOff, y + yOff);
			sum += f.distance * noiseScale;
			if (++i == detail2)
				break;
			x += x;
			y += y;
			noiseScale *= noiseFactor;
		}
	
	/* Return the fractal noise value */
	
	return sum;
}  /* CellularFractalNoise2D */

float MarbleNoise(float x, float y, float z, float noiseFactor, float offset, float turbulence, float scale, long detail1, long detail2)
{
	float noiseScale = 6.0 * turbulence;
	long i = 0;

	/* Prevent the texture from shifting while animating the turblence */
	
	offset -= noiseScale;

	/* Calculate cubic interpolated noise for the higher octaves */
	
	if (detail1)
		while (i < detail1) {
			offset += Noise(x, y, z) * noiseScale;
			x += x;
			y += y;
			z += z;
			noiseScale *= noiseFactor;
			i++;
		}
	
	/* Calculate linear interpolated noise for the lower octaves */
	
	if (i < detail2)
		while (1) {
			offset += Noise2(x, y, z) * noiseScale;
			if (++i == detail2)
				break;
			x += x;
			y += y;
			z += z;
			noiseScale *= noiseFactor;
		}
	
	/* Calculate the marble function */
	
	offset = SIN(offset * scale);
	return (offset < 0) ? -offset : offset;
}  /* MarbleNoise */

float MarbleNoise2D(float x, float y, float noiseFactor, float offset, float turbulence, float scale, long detail1, long detail2)
{
	float noiseScale = 6.0 * turbulence;
	long i = 0;

	/* Prevent the texture from shifting while animating the turblence */
	
	offset -= noiseScale;

	/* Calculate cubic interpolated noise for the higher octaves */
	
	if (detail1)
		while (i < detail1) {
			offset += Noise2D(x, y) * noiseScale;
			x += x;
			y += y;
			noiseScale *= noiseFactor;
			i++;
		}
	
	/* Calculate linear interpolated noise for the lower octaves */
	
	if (i < detail2)
		while (1) {
			offset += Noise2D2(x, y) * noiseScale;
			if (++i == detail2)
				break;
			x += x;
			y += y;
			noiseScale *= noiseFactor;
		}
	
	/* Calculate the marble function */
	
	offset = SIN(offset * scale);
	return (offset < 0) ? -offset : offset;

}  /* MarbleNoise2D */

long Mandelbrot(double xPos, double yPos, long crunch)
{
	double x, y, xx, yy, n;
	long k;
	
	x = xPos;
	y = yPos;
	xx = x * x;
	yy = y * y;
	n = 4.0;
	for (k = crunch; k > 0; k--) {
		if (xx + yy <= n)
			break;
		y = x * y + yPos;
		y += y;
		x = xx - yy + xPos;
		xx = x * x;
		yy = y * y;
	}
	return crunch - k;
}  /* Mandelbrot */
