/*
Copyright 1995-1998 Electric Image, Inc. This source code may only be used by
registered owners of the ElectricImage Plug-In Kit. Owners of the Kit may
include the source code in their products (freeware, shareware, or commercial).
This copyright notice must be included in its original form in all such
products. This source code may not be offered for sale or distributed by any
entity without the written permission of Electric Image, Inc.
*/

#include "PluginMatrix.h"
#include "MemoryUtils.h"
#include "PluginFPU.h"

#define matrix_minScale	1.0e-10

static void PivotGauss4(Matrix4Ptr srcMatrix, long* indexVector);
static void PivotSolve4(Matrix4Ptr srcMatrix, double* coeffVector, double* resultVector, long* indexVector);

/*
 *		Static Functions
 */

static void PivotGauss4(Matrix4Ptr srcMatrix, long* indexVector)
{
	Vector4 scaleVector;
	double sMax, rMax, r, xMult, tmp;
	long indexSwap;
	long i, j, k;

	for (i = 0; i <= 3; i++) {
		indexVector[i] = i;
		sMax = 0.0;
		for (j = 0; j <= 3; j++) {
			tmp = FABS(srcMatrix->matrix[i][j]);
			if (sMax < tmp)
				sMax = tmp;
		}
		scaleVector[i] = sMax;
	}
	for (k = 0; k <= 2; k++) {
		rMax = 0.0;
		for (i = k; i <= 3; i++) {
			r = FABS(srcMatrix->matrix[indexVector[i]][k]) / scaleVector[indexVector[i]];
			if (r > rMax) {
				j = i;
				rMax = r;
			}
		}
		indexSwap = indexVector[j];
		indexVector[j] = indexVector[k];
		indexVector[k] = indexSwap;
		for (i = k + 1; i <= 3; i++) {
			xMult = srcMatrix->matrix[indexVector[i]][k] / srcMatrix->matrix[indexSwap][k];
			for (j = k + 1; j <= 3; j++)
				srcMatrix->matrix[indexVector[i]][j] = srcMatrix->matrix[indexVector[i]][j] - xMult * srcMatrix->matrix[indexSwap][j];
			srcMatrix->matrix[indexVector[i]][k] = xMult;
		}
	}
}  /* PivotGauss4 */

static void PivotSolve4(Matrix4Ptr srcMatrix, double* coeffVector, double* resultVector, long* indexVector)
{
	long i, j;
	double sum;

	for (i = 0; i <= 2; i++)
		for (j = i + 1; j <= 3; j++)
			coeffVector[indexVector[j]] = coeffVector[indexVector[j]] - srcMatrix->matrix[indexVector[j]][i] * coeffVector[indexVector[i]];
	resultVector[3] = coeffVector[indexVector[3]] / srcMatrix->matrix[indexVector[3]][3];
	for (i = 2; i >= 0; i--) {
		sum = coeffVector[indexVector[i]];
		for (j = i + 1; j <= 3; j++)
			sum -= srcMatrix->matrix[indexVector[i]][j] * resultVector[j];
		resultVector[i] = sum / srcMatrix->matrix[indexVector[i]][i];
	}
}  /* PivotSolve4 */

/* Math Routines */

boolean Equal0(float n)
{
	n += 1.0;
	return (double)n == 1.0;
}  /* Equal0 */

boolean Equal1(float n)
{
	return (double)n == 1.0;
}  /* Equal1 */

void Rot(double sinTheta, double cosTheta, double* n1, double* n2)
{
	double n3;

	n3 = *n1 * cosTheta - *n2 * sinTheta;
	*n2 = *n1 * sinTheta + *n2 * cosTheta;
	*n1 = n3;
}  /* Rot */

/* Matrix Routines */

void InvertMatrix4(Matrix4Ptr theMatrix, Matrix4 srcMatrix)
{
	Matrix4 identity;
	Vector4 coeffVector, resultVector;
	long indexVector[4];
	long i, j;
	
	PivotGauss4(&srcMatrix, indexVector);
	InitMatrix4(&identity);
	for (i = 0; i <= 3; i++) {
		CopyMem(identity.mv[i], coeffVector, sizeof(Vector4));
		PivotSolve4(&srcMatrix, coeffVector, resultVector, indexVector);
		for (j = 0; j <= 3; j++)
			theMatrix->matrix[j][i] = resultVector[j];
	}
}  /* InvertMatrix4 */

void TransposeMatrix4(Matrix4Ptr theMatrix, Matrix4 srcMatrix)
{
	theMatrix->m.m00 = srcMatrix.m.m00;
	theMatrix->m.m01 = srcMatrix.m.m10;
	theMatrix->m.m02 = srcMatrix.m.m20;
	theMatrix->m.m03 = srcMatrix.m.m30;
	theMatrix->m.m10 = srcMatrix.m.m01;
	theMatrix->m.m11 = srcMatrix.m.m11;
	theMatrix->m.m12 = srcMatrix.m.m21;
	theMatrix->m.m13 = srcMatrix.m.m31;
	theMatrix->m.m20 = srcMatrix.m.m02;
	theMatrix->m.m21 = srcMatrix.m.m12;
	theMatrix->m.m22 = srcMatrix.m.m22;
	theMatrix->m.m23 = srcMatrix.m.m32;
	theMatrix->m.m30 = srcMatrix.m.m03;
	theMatrix->m.m31 = srcMatrix.m.m13;
	theMatrix->m.m32 = srcMatrix.m.m23;
	theMatrix->m.m33 = srcMatrix.m.m33;
}  /* TransposeMatrix4 */

void AppendMatrix4(Matrix4Ptr theMatrix, Matrix4Ptr b)
{
	Matrix4 a = *theMatrix;
	
	theMatrix->m.m00 = a.m.m00 * b->m.m00 + a.m.m01 * b->m.m10 + a.m.m02 * b->m.m20 + a.m.m03 * b->m.m30;
	theMatrix->m.m01 = a.m.m00 * b->m.m01 + a.m.m01 * b->m.m11 + a.m.m02 * b->m.m21 + a.m.m03 * b->m.m31;
	theMatrix->m.m02 = a.m.m00 * b->m.m02 + a.m.m01 * b->m.m12 + a.m.m02 * b->m.m22 + a.m.m03 * b->m.m32;
	theMatrix->m.m03 = a.m.m00 * b->m.m03 + a.m.m01 * b->m.m13 + a.m.m02 * b->m.m23 + a.m.m03 * b->m.m33;
	
	theMatrix->m.m10 = a.m.m10 * b->m.m00 + a.m.m11 * b->m.m10 + a.m.m12 * b->m.m20 + a.m.m13 * b->m.m30;
	theMatrix->m.m11 = a.m.m10 * b->m.m01 + a.m.m11 * b->m.m11 + a.m.m12 * b->m.m21 + a.m.m13 * b->m.m31;
	theMatrix->m.m12 = a.m.m10 * b->m.m02 + a.m.m11 * b->m.m12 + a.m.m12 * b->m.m22 + a.m.m13 * b->m.m32;
	theMatrix->m.m13 = a.m.m10 * b->m.m03 + a.m.m11 * b->m.m13 + a.m.m12 * b->m.m23 + a.m.m13 * b->m.m33;
	
	theMatrix->m.m20 = a.m.m20 * b->m.m00 + a.m.m21 * b->m.m10 + a.m.m22 * b->m.m20 + a.m.m23 * b->m.m30;
	theMatrix->m.m21 = a.m.m20 * b->m.m01 + a.m.m21 * b->m.m11 + a.m.m22 * b->m.m21 + a.m.m23 * b->m.m31;
	theMatrix->m.m22 = a.m.m20 * b->m.m02 + a.m.m21 * b->m.m12 + a.m.m22 * b->m.m22 + a.m.m23 * b->m.m32;
	theMatrix->m.m23 = a.m.m20 * b->m.m03 + a.m.m21 * b->m.m13 + a.m.m22 * b->m.m23 + a.m.m23 * b->m.m33;
	
	theMatrix->m.m30 = a.m.m30 * b->m.m00 + a.m.m31 * b->m.m10 + a.m.m32 * b->m.m20 + a.m.m33 * b->m.m30;
	theMatrix->m.m31 = a.m.m30 * b->m.m01 + a.m.m31 * b->m.m11 + a.m.m32 * b->m.m21 + a.m.m33 * b->m.m31;
	theMatrix->m.m32 = a.m.m30 * b->m.m02 + a.m.m31 * b->m.m12 + a.m.m32 * b->m.m22 + a.m.m33 * b->m.m32;
	theMatrix->m.m33 = a.m.m30 * b->m.m03 + a.m.m31 * b->m.m13 + a.m.m32 * b->m.m23 + a.m.m33 * b->m.m33;
}  /* AppendMatrix4 */

void InsertMatrix4(Matrix4Ptr theMatrix, Matrix4Ptr a)
{
	Matrix4 b = *theMatrix;

	theMatrix->m.m00 = a->m.m00 * b.m.m00 + a->m.m01 * b.m.m10 + a->m.m02 * b.m.m20 + a->m.m03 * b.m.m30;
	theMatrix->m.m10 = a->m.m10 * b.m.m00 + a->m.m11 * b.m.m10 + a->m.m12 * b.m.m20 + a->m.m13 * b.m.m30;
	theMatrix->m.m20 = a->m.m20 * b.m.m00 + a->m.m21 * b.m.m10 + a->m.m22 * b.m.m20 + a->m.m23 * b.m.m30;
	theMatrix->m.m30 = a->m.m30 * b.m.m00 + a->m.m31 * b.m.m10 + a->m.m32 * b.m.m20 + a->m.m33 * b.m.m30;
	
	theMatrix->m.m01 = a->m.m00 * b.m.m01 + a->m.m01 * b.m.m11 + a->m.m02 * b.m.m21 + a->m.m03 * b.m.m31;
	theMatrix->m.m11 = a->m.m10 * b.m.m01 + a->m.m11 * b.m.m11 + a->m.m12 * b.m.m21 + a->m.m13 * b.m.m31;
	theMatrix->m.m21 = a->m.m20 * b.m.m01 + a->m.m21 * b.m.m11 + a->m.m22 * b.m.m21 + a->m.m23 * b.m.m31;
	theMatrix->m.m31 = a->m.m30 * b.m.m01 + a->m.m31 * b.m.m11 + a->m.m32 * b.m.m21 + a->m.m33 * b.m.m31;
	
	theMatrix->m.m02 = a->m.m00 * b.m.m02 + a->m.m01 * b.m.m12 + a->m.m02 * b.m.m22 + a->m.m03 * b.m.m32;
	theMatrix->m.m12 = a->m.m10 * b.m.m02 + a->m.m11 * b.m.m12 + a->m.m12 * b.m.m22 + a->m.m13 * b.m.m32;
	theMatrix->m.m22 = a->m.m20 * b.m.m02 + a->m.m21 * b.m.m12 + a->m.m22 * b.m.m22 + a->m.m23 * b.m.m32;
	theMatrix->m.m32 = a->m.m30 * b.m.m02 + a->m.m31 * b.m.m12 + a->m.m32 * b.m.m22 + a->m.m33 * b.m.m32;
	
	theMatrix->m.m03 = a->m.m00 * b.m.m03 + a->m.m01 * b.m.m13 + a->m.m02 * b.m.m23 + a->m.m03 * b.m.m33;
	theMatrix->m.m13 = a->m.m10 * b.m.m03 + a->m.m11 * b.m.m13 + a->m.m12 * b.m.m23 + a->m.m13 * b.m.m33;
	theMatrix->m.m23 = a->m.m20 * b.m.m03 + a->m.m21 * b.m.m13 + a->m.m22 * b.m.m23 + a->m.m23 * b.m.m33;
	theMatrix->m.m33 = a->m.m30 * b.m.m03 + a->m.m31 * b.m.m13 + a->m.m32 * b.m.m23 + a->m.m33 * b.m.m33;
}  /* InsertMatrix4 */

void NormalMatrix4(Matrix4Ptr theMatrix)
{
	InvertMatrix4(theMatrix, *theMatrix);
	TransposeMatrix4(theMatrix, *theMatrix);
}  /* NormalMatrix4 */

void TranslateMatrix4(Matrix4Ptr theMatrix, double tx, double ty, double tz)
{
	theMatrix->m.m00 = theMatrix->m.m03 * tx + theMatrix->m.m00;
	theMatrix->m.m01 = theMatrix->m.m03 * ty + theMatrix->m.m01;
	theMatrix->m.m02 = theMatrix->m.m03 * tz + theMatrix->m.m02;
	theMatrix->m.m10 = theMatrix->m.m13 * tx + theMatrix->m.m10;
	theMatrix->m.m11 = theMatrix->m.m13 * ty + theMatrix->m.m11;
	theMatrix->m.m12 = theMatrix->m.m13 * tz + theMatrix->m.m12;
	theMatrix->m.m20 = theMatrix->m.m23 * tx + theMatrix->m.m20;
	theMatrix->m.m21 = theMatrix->m.m23 * ty + theMatrix->m.m21;
	theMatrix->m.m22 = theMatrix->m.m23 * tz + theMatrix->m.m22;
	theMatrix->m.m30 = theMatrix->m.m33 * tx + theMatrix->m.m30;
	theMatrix->m.m31 = theMatrix->m.m33 * ty + theMatrix->m.m31;
	theMatrix->m.m32 = theMatrix->m.m33 * tz + theMatrix->m.m32;
}  /* TranslateMatrix4 */

void XRotateMatrix4(Matrix4Ptr theMatrix, double sinTheta, double cosTheta)
{
	double temp = theMatrix->m.m01;
	theMatrix->m.m01 = temp * cosTheta - theMatrix->m.m02 * sinTheta;
	theMatrix->m.m02 = temp * sinTheta + theMatrix->m.m02 * cosTheta;
	temp = theMatrix->m.m11;
	theMatrix->m.m11 = temp * cosTheta - theMatrix->m.m12 * sinTheta;
	theMatrix->m.m12 = temp * sinTheta + theMatrix->m.m12 * cosTheta;
	temp = theMatrix->m.m21;
	theMatrix->m.m21 = temp * cosTheta - theMatrix->m.m22 * sinTheta;
	theMatrix->m.m22 = temp * sinTheta + theMatrix->m.m22 * cosTheta;
	temp = theMatrix->m.m31;
	theMatrix->m.m31 = temp * cosTheta - theMatrix->m.m32 * sinTheta;
	theMatrix->m.m32 = temp * sinTheta + theMatrix->m.m32 * cosTheta;
}  /* XRotateMatrix4 */

void YRotateMatrix4(Matrix4Ptr theMatrix, double sinTheta, double cosTheta)
{
	double temp = theMatrix->m.m00;
	theMatrix->m.m00 = temp * cosTheta + theMatrix->m.m02 * sinTheta;
	theMatrix->m.m02 = theMatrix->m.m02 * cosTheta - temp * sinTheta;
	temp = theMatrix->m.m10;
	theMatrix->m.m10 = temp * cosTheta + theMatrix->m.m12 * sinTheta;
	theMatrix->m.m12 = theMatrix->m.m12 * cosTheta - temp * sinTheta;
	temp = theMatrix->m.m20;
	theMatrix->m.m20 = temp * cosTheta + theMatrix->m.m22 * sinTheta;
	theMatrix->m.m22 = theMatrix->m.m22 * cosTheta - temp * sinTheta;
	temp = theMatrix->m.m30;
	theMatrix->m.m30 = temp * cosTheta + theMatrix->m.m32 * sinTheta;
	theMatrix->m.m32 = theMatrix->m.m32 * cosTheta - temp * sinTheta;
}  /* YRotateMatrix4 */

void ZRotateMatrix4(Matrix4Ptr theMatrix, double sinTheta, double cosTheta)
{
	double temp = theMatrix->m.m00;
	theMatrix->m.m00 = temp * cosTheta - theMatrix->m.m01 * sinTheta;
	theMatrix->m.m01 = temp * sinTheta + theMatrix->m.m01 * cosTheta;
	temp = theMatrix->m.m10;
	theMatrix->m.m10 = temp * cosTheta - theMatrix->m.m11 * sinTheta;
	theMatrix->m.m11 = temp * sinTheta + theMatrix->m.m11 * cosTheta;
	temp = theMatrix->m.m20;
	theMatrix->m.m20 = temp * cosTheta - theMatrix->m.m21 * sinTheta;
	theMatrix->m.m21 = temp * sinTheta + theMatrix->m.m21 * cosTheta;
	temp = theMatrix->m.m30;
	theMatrix->m.m30 = temp * cosTheta - theMatrix->m.m31 * sinTheta;
	theMatrix->m.m31 = temp * sinTheta + theMatrix->m.m31 * cosTheta;
}  /* ZRotateMatrix4 */

void ScaleMatrix4(Matrix4Ptr theMatrix, double sx, double sy, double sz)
{
	if (FABS(sx) < matrix_minScale)
		sx = (sx >= 0.0 ? matrix_minScale : -matrix_minScale);
	if (FABS(sy) < matrix_minScale)
		sy = (sy >= 0.0 ? matrix_minScale : -matrix_minScale);
	if (FABS(sz) < matrix_minScale)
		sz = (sz >= 0.0 ? matrix_minScale : -matrix_minScale);
	theMatrix->m.m00 *= sx;
	theMatrix->m.m01 *= sy;
	theMatrix->m.m02 *= sz;
	theMatrix->m.m10 *= sx;
	theMatrix->m.m11 *= sy;
	theMatrix->m.m12 *= sz;
	theMatrix->m.m20 *= sx;
	theMatrix->m.m21 *= sy;
	theMatrix->m.m22 *= sz;
	theMatrix->m.m30 *= sx;
	theMatrix->m.m31 *= sy;
	theMatrix->m.m32 *= sz;
}  /* ScaleMatrix4 */

void TransformPoint(double* x, double* y, double* z, Matrix4Ptr theMatrix)
{
	double xp = *x, yp = *y, zp = *z;

	*x = xp * theMatrix->m.m00 + yp * theMatrix->m.m10 + zp * theMatrix->m.m20 + theMatrix->m.m30;
	*y = xp * theMatrix->m.m01 + yp * theMatrix->m.m11 + zp * theMatrix->m.m21 + theMatrix->m.m31;
	*z = xp * theMatrix->m.m02 + yp * theMatrix->m.m12 + zp * theMatrix->m.m22 + theMatrix->m.m32;
}  /* TransformPoint */

void TransformNormal(double* a, double* b, double* c, Matrix4Ptr theMatrix)
{
	double ap = *a, bp = *b, cp = *c;
	double aNew = ap * theMatrix->m.m00 + bp * theMatrix->m.m10 + cp * theMatrix->m.m20;
	double bNew = ap * theMatrix->m.m01 + bp * theMatrix->m.m11 + cp * theMatrix->m.m21;
	double cNew = ap * theMatrix->m.m02 + bp * theMatrix->m.m12 + cp * theMatrix->m.m22;
	double distance = SQRT(SQR(aNew) + SQR(bNew) + SQR(cNew));

	if (distance == 0.0) {
		*a = 0.0;
		*b = 0.0;
		*c = 1.0;
	} else {
		distance = 1.0 / distance;
		*a = aNew * distance;
		*b = bNew * distance;
		*c = cNew * distance;
	}
}  /* TransformNormal */

void TransformDCoordinate(DCoordPtr theCoordinate, Matrix4Ptr theMatrix)
{
	double xp = theCoordinate->x, yp = theCoordinate->y, zp = theCoordinate->z;

	theCoordinate->x = xp * theMatrix->m.m00 + yp * theMatrix->m.m10 + zp * theMatrix->m.m20 + theMatrix->m.m30;
	theCoordinate->y = xp * theMatrix->m.m01 + yp * theMatrix->m.m11 + zp * theMatrix->m.m21 + theMatrix->m.m31;
	theCoordinate->z = xp * theMatrix->m.m02 + yp * theMatrix->m.m12 + zp * theMatrix->m.m22 + theMatrix->m.m32;
}  /* TransformDCoordinate */

