/* Authored by Jules Bloomenthal, Xerox PARC.
 * Copyright (c) Xerox Corporation, 1991.  All rights reserved.
 * Permission is granted to reproduce, use and distribute this code for
 * any and all purposes, provided that this notice appears in all copies. */

/* Modified for use as an ElectricImage model plug-in by Mark Granger, Electric Image Inc. */

#include "PluginClient.h"
#ifdef __PLUGINLIBRARY__
#include "FPU.h"
#else
#include "PluginFPU.h"
#endif
#include "Implicit.h"
#include "PString.h"

#include <stddef.h>

extern BlobPtr firstBlob;
extern long blobCount, result;
extern ulong theVertexFlags;
extern FunctionRoutine theFunctionRoutine;
extern ColorFunctionRoutine theColorFunctionRoutine;

static ulong randomSeed;

#define RAND() 		((randomSeed = (long)randomSeed * 1103515245 + 12345) * 2.328306436E-10)	/* yields a number between 0 and <1 */
#define HASHBIT	    (5)
#define HASHSIZE    (long)(1L << (3 * HASHBIT))		/* hash table size (32768) */
#define HASHSIZE2	(HASHSIZE + HASHSIZE)
#define HASHMASK	((1L << HASHBIT) - 1L)
#define HASH(i,j,k) ((((((i) & HASHMASK) << HASHBIT) | ((j) & HASHMASK)) << HASHBIT) | ((k) & HASHMASK))
#define TESTBIT(i, bit) (((i) >> (bit)) & 1L)
#define FLIP(i,bit) ((i)^1 << (bit))					/* flip the given bit of i */

#define RES	10 /* # converge iterations    */

/* the LBN corner of cube (i, j, k), corresponds with location */
/* (start.x + (i - 0.5) * cubeSize, start.y + (j - 0.5) * cubeSize, start.z + (k - 0.5) * cubeSize) */

#define LLL	0  /* left direction:	-x, -i */
#define RRR	1  /* right direction:	+x, +i */
#define BBB	2  /* bottom direction: -y, -j */
#define TTT	3  /* top direction:	+y, +j */
#define NNN	4  /* near direction:	-z, -k */
#define FFF	5  /* far direction:	+z, +k */

#define LBN	0  /* left bottom near corner  */
#define LBF	1  /* left bottom far corner   */
#define LTN	2  /* left top near corner     */
#define LTF	3  /* left top far corner      */
#define RBN	4  /* right bottom near corner */
#define RBF	5  /* right bottom far corner  */
#define RTN	6  /* right top near corner    */
#define RTF	7  /* right top far corner     */

/**** Cubical Polygonization (optional) ****/

#define LB	0  /* left bottom edge	*/
#define LT	1  /* left top edge	*/
#define LN	2  /* left near edge	*/
#define LF	3  /* left far edge	*/
#define RB	4  /* right bottom edge */
#define RT	5  /* right top edge	*/
#define RN	6  /* right near edge	*/
#define RF	7  /* right far edge	*/
#define BN	8  /* bottom near edge	*/
#define BF	9  /* bottom far edge	*/
#define TN	10 /* top near edge	*/
#define TF	11 /* top far edge	*/

typedef struct {					/* test the function for a signed value */
	Coordinate p;					/* location of test */
	double value;					/* function value at p */
	long ok;						/* if value is of correct sign */
} TestRec, *TestPtr;

typedef struct {					/* corner of a cube */
	long i, j, k;				/* (i, j, k) is index within lattice */
	float x, y, z, value;			/* location and function value */
} CornerRec, *CornerPtr;

typedef struct {					/* partitioning cell (cube) */
	long i, j, k;				/* lattice location of cube */
	CornerPtr corners[8];			/* eight corners */
} CubeRec, *CubePtr;

typedef struct _CubeList {			/* linked list of cubes acting as stack */
	CubeRec cube;					/* a single cube */
	struct _CubeList *next;			/* remaining elements */
} CubeList, *CubeListPtr;

typedef struct _CenterList {		/* list of cube locations */
	long i, j, k;				/* cube location */
	struct _CenterList *next;		/* remaining elements */
} CenterList, *CenterListPtr;

typedef struct _CornerList {		/* list of corners */
	long i, j, k;				/* corner id */
	float value;						/* corner value */
	struct _CornerList *next;		/* remaining elements */
} CornerList, *CornerListPtr;

typedef struct _EdgeList {			/* list of edges */
	long i1, j1, k1, i2, j2, k2;	/* edge corner ids */
	VertexPtr vid;					/* vertex id */
	struct _EdgeList *next;			/* remaining elements */
} EdgeList, *EdgeListPtr;

typedef struct _IntList {			/* list of integers */
	long i;						/* an integer */
	struct _IntList *next;			/* remaining elements */
} IntList, *IntListPtr;

typedef struct _IntLists {			/* list of list of integers */
	IntListPtr list;				/* a list of integers */
	struct _IntLists *next;			/* remaining elements */
} IntLists, *IntListsPtr;


static double cubeSize, normalDelta;	/* cube size, normal delta */
static Coordinate start;			/* start point on surface */
static CubeListPtr cubes;			/* active cubes */
static CubeListPtr freeCubes;		/* inactive cubes */
static CenterListPtr *centers;		/* cube center hash table */
static CornerListPtr *corners;		/* corner value hash table */
static EdgeListPtr *edges;			/* edge and vertex id hash table */
static IntListsPtr *cubetable;		/* the cube table */
static long cubeTotal; 			/* the total number of cubes */
static long cubeCount;			/* the number of cubes which have been processed */

/* edge: LB, LT, LN, LF, RB, RT, RN, RF, BN, BF, TN, TF */

static long corner1[12]	   = {LBN,LTN,LBN,LBF,RBN,RTN,RBN,RBF,LBN,LBF,LTN,LTF};
static long corner2[12]	   = {LBF,LTF,LTN,LTF,RBF,RTF,RTN,RTF,RBN,RBF,RTN,RTF};

/* face on left when going corner1 to corner2 */

static long leftface[12]	   = {BBB,  LLL,  LLL,  FFF,  RRR,  TTT,  NNN,  RRR,  NNN,  BBB,  TTT,  FFF};

/* face on right when going corner1 to corner2 */

static long rightface[12]   = {LLL,  TTT,  NNN,  LLL,  BBB,  RRR,  RRR,  FFF,  BBB,  FFF,  NNN,  TTT};

/**** An Implicit Surface Polygonizer ****/

static boolean ProcessBlob(BlobPtr theBlob);
static void ProcessCubes(boolean tetMode);
static void TestFace(long i, long j, long k, CubePtr old, long face, long c1, long c2, long c3, long c4);
static void AddCube(CubePtr newCube);
static CornerPtr SetCorner(long i, long j, long k);
static void FindTest(TestPtr test, boolean sign, double x, double y, double z);
static void DoTetrahedron(CubePtr cube, long c1, long c2, long c3, long c4);
static void DoCube(CubePtr cube);
static long NextCWedge(long edge, long face);
static long OtherFace(long edge, long face);
static void MakeCubeTable(void);
static boolean SetCenter(long i, long j, long k);
static void SetEdge(long i1, long j1, long k1, long i2, long j2, long k2, VertexPtr vid);
static VertexPtr GetEdge(long i1, long j1, long k1, long i2, long j2, long k2);
static VertexPtr VertID(CornerPtr c1, CornerPtr c2);
static void VNormal(CoordPtr position, CoordPtr normal, CoordPtr blur, CoordPtr texture, ARGBPtr theColor);
static void Converge(Coordinate *p1, Coordinate *p2, double v, Coordinate *p);

/* polygonize: polygonize the implicit surface function
 *   arguments are:
 *	 double cubeSize
 *	     width of the partitioning cube
 *	 double x, y, z
 *	     coordinates of a starting point on or near the surface
 *	     may be defaulted to 0., 0., 0.
 *	 long tetFlag
 *	     true: decompose cube and polygonize six tetrahedra
 *	     false: polygonize cube directly
 *   returns error or NULL
 */

void Polygonize(double theCubeSize, double x, double y, double z, boolean tetMode)
{
	TestRec in, out;
	long n;
	long k = HASHSIZE;
	
	SetupStatus(PS("Cube"), PS("Creating Surface"), 0, cubeTotal, 0.0, 0);
	if (result)
		return;

	cubeSize = theCubeSize;
	normalDelta = cubeSize / (double)(RES * RES);
	
	/* allocate hash tables and build cube polygon table: */

	centers = (CenterListPtr*)AllocateMem(k, sizeof(CenterListPtr));
	if (result)
		return;
	for (n = 0; n < k; n++)
		centers[n] = NULL;

	corners = (CornerListPtr*)AllocateMem(k, sizeof(CornerListPtr));
	if (result)
		return;
	for (n = 0; n < k; n++)
		corners[n] = NULL;

	edges =	(EdgeListPtr*)AllocateMem(2 * k, sizeof(EdgeListPtr));
	if (result)
		return;
	for (n = 0; n < 2 * k; n++)
		edges[n] = NULL;

	if (!tetMode) {
		cubetable = (IntListsPtr*)AllocateMem(256, sizeof(IntListsPtr));
		if (result)
			return;
		for (n = 0; n < 256; n++)
			cubetable[n] = NULL;
		
		MakeCubeTable();
		if (result)
			return;
	} else
		cubetable = NULL;
		
	/* Find a point on the surface, beginning the search at (x, y, z): */

	randomSeed = 1;
	FindTest(&in, 1, x, y, z);
	FindTest(&out, 0, x, y, z);
	if ((!in.ok) || (!out.ok))
		return;
	Converge(&in.p, &out.p, in.value, &start);
	
	/* push initial cube on stack: */

	cubes = (CubeListPtr)AllocateMem(1, sizeof(CubeList)); /* list of 1 */
	if (result)
		return;
	cubes->cube.i = 0;
	cubes->cube.j = 0;
	cubes->cube.k = 0;
	cubes->next = NULL;
	freeCubes = NULL;
	cubeTotal = 1;
	cubeCount = 0;
	
	/* set corners of initial cube: */

	for (n = 0; n < 8; n++) {
		cubes->cube.corners[n] = SetCorner(TESTBIT(n, 2), TESTBIT(n, 1), TESTBIT(n, 0));
		if (result) return;
	}
	
	/* create an entry in the centers list for the initial cube */
	
	SetCenter(0, 0, 0);
	
	/* process all of the active cubes */
	
	if (!result)
		ProcessCubes(tetMode);

	/* check each blob to see if it has been processed */
	
	if (!result) {
		BlobPtr theBlob = firstBlob;
	
		while ((theBlob != NULL) && (!result)) {
			if (ProcessBlob(theBlob))
				ProcessCubes(tetMode);
			theBlob = theBlob->next;
		}
	}
	
	/* finish updating the status box */
	
	if (!result)
		UpdateStatus(1.0, cubeTotal);
}  /* Polygonize */

/* Try to locate a crossing cube starting at the center of the blob and moving in the +X direction */

static boolean ProcessBlob(BlobPtr theBlob)
{
	CubeRec newCube;
	boolean found;
	long count;
	
	/* use the center of the blob to find the starting location of the cube */
	
	newCube.i = floor(((theBlob->center.x - start.x) / cubeSize) + 0.5);
	newCube.j = floor(((theBlob->center.y - start.y) / cubeSize) + 0.5);
	newCube.k = floor(((theBlob->center.z - start.z) / cubeSize) + 0.5);

	/* go until 10 times the width of the blob */
	
	count = 10.0 * theBlob->radius / cubeSize;	
	found = 0;
	while ((count > 0) && (!found) && (!result)) {
		boolean pos;
		long n;
		
		/* set the 8 corners of the cube and watch for a change in the sign */
		
		newCube.corners[0] = SetCorner(newCube.i, newCube.j, newCube.k);
		pos = newCube.corners[0]->value > 0.0;
		for (n = 1; (n < 8) && (!result); n++) {
			newCube.corners[n] = SetCorner(newCube.i + TESTBIT(n, 2), newCube.j + TESTBIT(n, 1), newCube.k + TESTBIT(n, 0));
			found = found || (pos != (newCube.corners[n]->value > 0.0));
		}
		if (!found) {
			newCube.i += 1;
			count--;
		}
	}
	
	/* if a crossing cube was found and not previously encountered, add it to the list */
	
	if (found && (!result) && (!SetCenter(newCube.i, newCube.j, newCube.k))) {
		AddCube(&newCube);
		return !result;
	} else
		return 0;
}  /* ProcessBlob */

/* ProcessCubes: process active cubes till none left */

static void ProcessCubes(boolean tetMode)
{
	while ((cubes != NULL) && (!result)) {
		CubeRec c;
		CubeListPtr temp = cubes;
		c = cubes->cube;
		
		cubeCount++;
		if (StatusTime())
			SetupStatus(PS("Cube"), PS("Creating Surface"), 0, cubeTotal, (double)cubeCount / cubeTotal, cubeCount);
		if (tetMode) {

			/* either decompose into tetrahedra and polygonize: */

			DoTetrahedron(&c, LBN, LTN, RBN, LBF);
			DoTetrahedron(&c, RTN, LTN, LBF, RBN);
			DoTetrahedron(&c, RTN, LTN, LTF, LBF);
			DoTetrahedron(&c, RTN, RBN, LBF, RBF);
			DoTetrahedron(&c, RTN, LBF, LTF, RBF);
			DoTetrahedron(&c, RTN, LTF, RTF, RBF);
		} else

			/* or polygonize the cube directly: */

			DoCube(&c);

		if (!result) {
			
			/* pop current cube from stack */
			
			cubes = cubes->next;
			temp->next = freeCubes;
			freeCubes = temp;
			
			/* test six face directions, maybe add cubes to stack: */
			
			TestFace(c.i-1, c.j, c.k, &c, LLL, LBN, LBF, LTN, LTF);	/* Left */
			TestFace(c.i+1, c.j, c.k, &c, RRR, RBN, RBF, RTN, RTF);	/* Right */
			TestFace(c.i, c.j-1, c.k, &c, BBB, LBN, LBF, RBN, RBF);	/* Bottom */
			TestFace(c.i, c.j+1, c.k, &c, TTT, LTN, LTF, RTN, RTF);	/* Top */
			TestFace(c.i, c.j, c.k-1, &c, NNN, LBN, LTN, RBN, RTN);	/* Near */
			TestFace(c.i, c.j, c.k+1, &c, FFF, LBF, LTF, RBF, RTF);	/* Far */
		}
	}
}  /* ProcessCubes */

/* TestFace: given cube at lattice (i, j, k), and four corners of face,
 * if surface crosses face, compute other four corners of adjacent cube
 * and add new cube to cube stack */

static void TestFace(long i, long j, long k, CubePtr old, long face, long c1, long c2, long c3, long c4)
{
	CubeRec newCube;
	static long facebit[6] = {2, 2, 1, 1, 0, 0};
	long n, bit = facebit[face];
	boolean pos = old->corners[c1]->value > 0.0;
	
	/* test if no surface crossing or already visited: */
	
	if (result) return;
	if (((old->corners[c2]->value > 0.0) == pos) && ((old->corners[c3]->value > 0.0) == pos) && ((old->corners[c4]->value > 0.0) == pos)) return;
	
	/* If a cube at the current location already exists, return */
	
	if (SetCenter(i, j, k)) return;
	if (result) return;
	
	/* create new cube: */
	
	newCube.i = i;
	newCube.j = j;
	newCube.k = k;
	for (n = 0; n < 8; n++)
		newCube.corners[n] = NULL;
	newCube.corners[FLIP(c1, bit)] = old->corners[c1];
	newCube.corners[FLIP(c2, bit)] = old->corners[c2];
	newCube.corners[FLIP(c3, bit)] = old->corners[c3];
	newCube.corners[FLIP(c4, bit)] = old->corners[c4];
	for (n = 0; n < 8; n++)
		if (newCube.corners[n] == NULL) {
			newCube.corners[n] = SetCorner(i + TESTBIT(n, 2), j + TESTBIT(n, 1), k + TESTBIT(n, 0));
			if (result) return;
		}
	
	/* add cube to top of stack: */
	
	AddCube(&newCube);
}  /* TestFace */

/* AddCube: add cube to top of stack: */

static void AddCube(CubePtr newCube)
{
	CubeListPtr oldCubes = cubes;

	if (freeCubes != NULL) {
		cubes = freeCubes;
		freeCubes = freeCubes->next;
	} else {
		cubes = (CubeListPtr) AllocateMem(1, sizeof(CubeList));
		if (result) return;
	}
	cubes->cube = *newCube;
	cubes->next = oldCubes;
	cubeTotal++;
}  /* AddCube */

/* SetCorner: return corner with the given lattice location
   set (and cache) its function value */

static CornerPtr SetCorner(long i, long j, long k)
{
	/* for speed, do corner value caching here */
	CornerPtr c;
	long index = HASH(i, j, k);
	CornerListPtr l = corners[index];
	
	c = (CornerPtr) AllocateMem(1, sizeof(CornerRec));
	if (result) return NULL;
	
	c->i = i;
	c->j = j;
	c->k = k;
	c->x = start.x + ((double)i - 0.5) * cubeSize;
	c->y = start.y + ((double)j - 0.5) * cubeSize;
	c->z = start.z + ((double)k - 0.5) * cubeSize;
	for (; l != NULL; l = l->next)
		if ((l->i == i) && (l->j == j) && (l->k == k)) {
			c->value = l->value;
			return c;
		}
	l = (CornerListPtr) AllocateMem(1, sizeof(CornerList));
	if (result) return NULL;
	l->i = i;
	l->j = j;
	l->k = k;
	l->value = c->value = theFunctionRoutine(c->x, c->y, c->z);
	l->next = corners[index];
	corners[index] = l;
	return c;
}  /* SetCorner */


/* FindTest: search for point with value of given sign (0: neg, 1: pos) */

static void FindTest(TestPtr test, boolean sign, double x, double y, double z)
{
	long i;
	double range = cubeSize;
	test->ok = 1;
	for (i = 0; i < 100000; i++) {
		test->p.x = x + range * (RAND() - 0.5);
		test->p.y = y + range * (RAND() - 0.5);
		test->p.z = z + range * (RAND() - 0.5);
		test->value = theFunctionRoutine(test->p.x, test->p.y, test->p.z);
		if (sign == (test->value > 0.0)) return;
		range *= 1.005; /* slowly expand search outwards */
	}
	test->ok = 0;
}  /* FindTest */


/**** Tetrahedral Polygonization ****/


/* DoTetrahedron: triangulate the tetrahedron
 * b, c, d should appear clockwise when viewed from a
 * return 0 if client aborts, 1 otherwise */

static void DoTetrahedron(CubePtr cube, long c1, long c2, long c3, long c4)
{
	CornerPtr a = cube->corners[c1];
	CornerPtr b = cube->corners[c2];
	CornerPtr c = cube->corners[c3];
	CornerPtr d = cube->corners[c4];
	long index = 0, apos, bpos, cpos, dpos;
	VertexPtr e1, e2, e3, e4, e5, e6;
	
	if (result) return;
	apos = (a->value > 0.0);
	bpos = (b->value > 0.0);
	cpos = (c->value > 0.0);
	dpos = (d->value > 0.0);
	if (apos) index += 8;
	if (bpos) index += 4;
	if (cpos) index += 2;
	if (dpos) index += 1;
	
	/* index is now 4-bit number representing one of the 16 possible cases */

	if (apos != bpos) e1 = VertID(a, b);
	if (apos != cpos) e2 = VertID(a, c);
	if (apos != dpos) e3 = VertID(a, d);
	if (bpos != cpos) e4 = VertID(b, c);
	if (bpos != dpos) e5 = VertID(b, d);
	if (cpos != dpos) e6 = VertID(c, d);
	if (result) return;
	
	/* 14 productive tetrahedral cases (0000 and 1111 do not yield polygons */
	
	switch (index) {
		case 1:
			NewFacet(e5, e6, e3, NULL);
			break;
		case 2:
			NewFacet(e2, e6, e4, NULL);
			break;
		case 3:
			NewFacet(e3, e5, e4, e2);
			break;
		case 4:
			NewFacet(e1, e4, e5, NULL);
			break;
		case 5:
			NewFacet(e3, e1, e4, e6);
			break;
		case 6:
			NewFacet(e1, e2, e6, e5);
			break;
		case 7:
			NewFacet(e1, e2, e3, NULL);
			break;
		case 8:
			NewFacet(e1, e3, e2, NULL);
			break;
		case 9:
			NewFacet(e1, e5, e6, e2);
			break;
		case 10:
			NewFacet(e1, e3, e6, e4);
			break;
		case 11:
			NewFacet(e1, e5, e4, NULL);
			break;
		case 12:
			NewFacet(e3, e2, e4, e5);
			break;
		case 13:
			NewFacet(e6, e2, e4, NULL);
			break;
		case 14:
			NewFacet(e5, e3, e6, NULL);
			break;
	}
	return;
}  /* DoTetrahedron */


/* DoCube: triangulate the cube directly, without decomposition */

static void DoCube(CubePtr cube)
{
	IntListsPtr polys;
	long i, index = 0;

	for (i = 0; i < 8; i++)
		if (cube->corners[i]->value > 0.0)
			index += (1<<i);

	for (polys = cubetable[index]; polys != NULL; polys = polys->next) {
		IntListPtr edges;
		long count = 0;
		VertexPtr facet[4];
		
		for (edges = polys->list; edges != NULL; edges = edges->next) {
			CornerPtr c1 = cube->corners[corner1[edges->i]];
			CornerPtr c2 = cube->corners[corner2[edges->i]];
			facet[count++] = VertID(c1, c2);
			if (count == 4) {
				NewFacet(facet[0], facet[1], facet[2], facet[3]);
				facet[1] = facet[3];
				count = 2;
			}
			if (result) return;
		}
		if (count == 3)
			NewFacet(facet[0], facet[1], facet[2], NULL);
		CheckStatus();
		if (result) return;
	}
	return;
}  /* DoCube */


/* NextCWedge: return next clockwise edge from given edge around given face */

static long NextCWedge(long edge, long face)
{
	switch (edge) {
		case LB: return (face == LLL)? LF : BN;
		case LT: return (face == LLL)? LN : TF;
		case LN: return (face == LLL)? LB : TN;
		case LF: return (face == LLL)? LT : BF;
		case RB: return (face == RRR)? RN : BF;
		case RT: return (face == RRR)? RF : TN;
		case RN: return (face == RRR)? RT : BN;
		case RF: return (face == RRR)? RB : TF;
		case BN: return (face == BBB)? RB : LN;
		case BF: return (face == BBB)? LB : RF;
		case TN: return (face == TTT)? LT : RN;
		case TF: return (face == TTT)? RT : LF;
	}
	return 0;
}  /* NextCWedge */


/* OtherFace: return face adjoining edge that is not the given face */

static long OtherFace(long edge, long face)
{
    long other = leftface[edge];
    if (face == other)
    	return rightface[edge];
    else
    	return other;
}  /* OtherFace */


/* MakeCubeTable: create the 256 entry table for cubical polygonization */

static void MakeCubeTable(void)
{
	long i, e, c, done[12], pos[8];

	for (i = 0; i < 256; i++) {
		for (e = 0; e < 12; e++)
			done[e] = 0;
		for (c = 0; c < 8; c++)
			pos[c] = TESTBIT(i, c);
		for (e = 0; e < 12; e++)
			if ((!done[e]) && (pos[corner1[e]] != pos[corner2[e]])) {
				IntListPtr ints = NULL;
				IntListsPtr lists;
				long start = e, edge = e;
				long face;
				
				lists = (IntListsPtr) AllocateMem(1, sizeof(IntLists));
				if (result) return;
	
				/* get face that is to right of edge from pos to neg corner: */

				if (pos[corner1[e]])
					face = rightface[e];
				else
					face = leftface[e];
				while (1) {
					CheckStatus();
					if (result) return;
					edge = NextCWedge(edge, face);
					done[edge] = 1;
					if (pos[corner1[edge]] != pos[corner2[edge]]) {
						IntListPtr tmp = ints;
						ints = (IntListPtr) AllocateMem(1, sizeof(IntList));
						if (result) return;
						ints->i = edge;
						ints->next = tmp; /* add edge to head of list */
						if (edge == start) break;
						face = OtherFace(edge, face);
					}
				}
				lists->list = ints; /* add ints to head of table entry */
				lists->next = cubetable[i];
				cubetable[i] = lists;
			}
	}
}  /* MakeCubeTable */


/* SetCenter: set (i,j,k) entry of centers[]
 * return true if already set; otherwise, set and return false */

static boolean SetCenter(long i, long j, long k)
{
    long index = HASH(i, j, k);
    CenterListPtr newCENTERLIST, l, q = centers[index];

	/* Look for an existing cube at the same location and return true if one is found */
	
    for (l = q; l != NULL; l = l->next)
		if ((l->i == i) && (l->j == j) && (l->k == k)) return 1;
		
	/* Create a new cube center and return false */
	
    newCENTERLIST = (CenterListPtr) AllocateMem(1, sizeof(CenterList));
	if (result) return 0;
    newCENTERLIST->i = i;
    newCENTERLIST->j = j;
    newCENTERLIST->k = k;
    newCENTERLIST->next = q;
    centers[index] = newCENTERLIST;
    return 0;
}  /* SetCenter */


/* SetEdge: set vertex id for edge */

static void SetEdge(long i1, long j1, long k1, long i2, long j2, long k2, VertexPtr vid)
{
	ulong index;
	EdgeListPtr newEDGELIST;

	if (i1>i2 || (i1==i2 && (j1>j2 || (j1==j2 && k1>k2)))) {
		long t;
		t=i1;
		i1=i2;
		i2=t;
		t=j1;
		j1=j2;
		j2=t;
		t=k1;
		k1=k2;
		k2=t;
	}
	index = HASH(i1, j1, k1) + HASH(i2, j2, k2);
	if (index >= HASHSIZE2)
		index = 0;
	newEDGELIST = (EdgeListPtr) AllocateMem(1, sizeof(EdgeList));
	if (result) return;
	newEDGELIST->i1 = i1;
	newEDGELIST->j1 = j1;
	newEDGELIST->k1 = k1;
	newEDGELIST->i2 = i2;
	newEDGELIST->j2 = j2;
	newEDGELIST->k2 = k2;
	newEDGELIST->vid = vid;
	newEDGELIST->next = edges[index];
	edges[index] = newEDGELIST;
}  /* SetEdge */


/* GetEdge: return vertex id for edge; return -1 if not set */

static VertexPtr GetEdge(long i1, long j1, long k1, long i2, long j2, long k2)
{
	EdgeListPtr q;
	long index;

	if (i1>i2 || (i1==i2 && (j1>j2 || (j1==j2 && k1>k2)))) {
		long t;
		t=i1;
		i1=i2;
		i2=t;
		t=j1;
		j1=j2;
		j2=t;
		t=k1;
		k1=k2;
		k2=t;
	};
	index = HASH(i1, j1, k1) + HASH(i2, j2, k2);
	if (index >= HASHSIZE2)
		index = 0;
	q = edges[index];
	while (q != NULL) {
		if ((q->i1 == i1) && (q->j1 == j1) && (q->k1 == k1) && (q->i2 == i2) && (q->j2 == j2) && (q->k2 == k2))
			return q->vid;
		q = q->next;
	}
	return NULL;
}  /* GetEdge */


/**** Vertices ****/


/* VertID: return index for vertex on edge:
 * c1->value and c2->value are presumed of different sign
 * return saved index if any; else compute vertex and save */

static VertexPtr VertID(CornerPtr c1, CornerPtr c2)
{
	Vertex v;
	Coordinate a, b;
	VertexPtr vid;

	if (result) return 0;
	vid = GetEdge(c1->i, c1->j, c1->k, c2->i, c2->j, c2->k);
	if (result) return 0;
	if (vid != NULL) return vid;					/* previously computed */
	a.x = c1->x;
	a.y = c1->y;
	a.z = c1->z;
	b.x = c2->x;
	b.y = c2->y;
	b.z = c2->z;
	Converge(&a, &b, c1->value, &v.position);	/* position */
	VNormal(&v.position, &v.normal, &v.blur, &v.texture, &v.color);	/* normal */
	v.index = 0x80000000;						/* index */
	vid = NewVertex(&v);						/* save vertex */
	if (result) return 0;
	SetEdge(c1->i, c1->j, c1->k, c2->i, c2->j, c2->k, vid);
	return vid;
}  /* VertID */


/* VNormal: compute unit length surface normal at point */

static void VNormal(CoordPtr position, CoordPtr normal, CoordPtr blur, CoordPtr texture, ARGBPtr theColor)
{
	double factor;
	
	factor = theColorFunctionRoutine(position, blur, texture, theColor);
	if (BTST(theVertexFlags, pluginVertexNormalFlag)) {
		normal->x = theFunctionRoutine(position->x + normalDelta, position->y, position->z) - factor;
		normal->y = theFunctionRoutine(position->x, position->y + normalDelta, position->z) - factor;
		normal->z = theFunctionRoutine(position->x, position->y, position->z + normalDelta) - factor;
		factor = SQRT(SQR(normal->x) + SQR(normal->y) + SQR(normal->z));
		if (factor != 0.0) {
			factor = 1.0 / factor;
			normal->x *= factor;
			normal->y *= factor;
			normal->z *= factor;
		}
	}
}  /* VNormal */


/* Converge: from two points of differing sign, Converge to zero crossing */

static void Converge(Coordinate *p1, Coordinate *p2, double v, Coordinate *p)
{
	long i = 0;
	Coordinate pos, neg;
	
	if (v < 0) {
		pos.x = p2->x;
		pos.y = p2->y;
		pos.z = p2->z;
		neg.x = p1->x;
		neg.y = p1->y;
		neg.z = p1->z;
	}
	else {
		pos.x = p1->x;
		pos.y = p1->y;
		pos.z = p1->z;
		neg.x = p2->x;
		neg.y = p2->y;
		neg.z = p2->z;
	}
	while (1) {
		p->x = 0.5 * (pos.x + neg.x);
		p->y = 0.5 * (pos.y + neg.y);
		p->z = 0.5 * (pos.z + neg.z);
		if (i++ == RES) return;
		if ((theFunctionRoutine(p->x, p->y, p->z)) > 0.0) {
			pos.x = p->x;
			pos.y = p->y;
			pos.z = p->z;
		} else {
			neg.x = p->x;
			neg.y = p->y;
			neg.z = p->z;
		}
	}
}  /* Converge */
