/*
Copyright 1995-1998 Electric Image, Inc. This source code is intended for teaching
purposes only. It may not be included in any products (freeware, shareware, or
commercial) without the written permission of Electric Image, Inc. This source
code may be copied and modified but this copyright notice must be included in
its original form. This source code may not be offered for sale or distributed
by any entity without the written permission of Electric Image, Inc. Copies of
the original or modified code may be posted to the ElectricImage Forum on
America-on-line or eWorld to foster a better understanding of the Plug-In
Specification.
*/

/* This code presents an example of a polygon mesh generator model plugin. */

#include "EI_Dialog.h"
#include "EI_ResourceFile.h"
#include "MemoryUtils.h"
#include "PluginClient.h"
#include "PluginFPU.h"
#include "PluginUtils.h"
#include "PluginMatrix.h"
#include "PString.h"

#define pluginType			QUADBYTE('M', 'E', 'S', 'H')		/* The owner plugin's unique type */
#define pluginVersion		0x00000002				/* The owner plugin's version number 0:0:version:revision */
#define pluginName			PS("ElectricImage Mesh")
#define pluginCopyright		PS("Copyright  1990-1998 ElectricImage, Inc.")
#define pluginAuthor		PS("Mark Granger|Electric Image, Inc.|117 East Colorado Boulevard, Suite 300|Pasadena, Ca 91105|(818) 577-1627")
#define pluginDescription	PS("This model plugin can create several different animated meshes.  It has selectable mesh resolution and scales.")
#define pluginPermission	BIT(pluginSensitivityFlag) + BIT(pluginTimeSensitivityFlag)

#define meshFlat        		0
#define meshSineWave    		1
#define meshBreathingCylinder	2
#define meshScrewThread  		3
#define meshSphereWrap  		4

#define defFrameTimeDelta	0.03333333   /* 30fps */

#define normalDelta			0.0001

#define defMeshFlags		0
#define defMeshMode			meshSphereWrap
#define defMeshUPoints		20
#define defMeshVPoints		20

#define defMeshXScale		100.0
#define defMeshYScale		100.0
#define defMeshZScale		100.0
#define defMeshTimeScale	1.0

#define	oldPluginVersion	0x00000001
#define oldPluginType		QUADBYTE('M', 'E', 'S', 'H')
#define oldControlSize		60

#define oldOwnerType		0
#define oldOwnerVersion		4
#define oldMeshFlags		8
#define	oldMeshMode			12
#define oldMeshMainColor	16
#define	oldMeshUPoints		20
#define	oldMeshVPoints		24
#define oldMeshXScale		28
#define	oldMeshYScale		36
#define	oldMeshZScale		44
#define	oldMeshTimeScale	52

#define aboutDialogID   	16000
#define interfaceDialogID	16001

enum {
	cMeshTypePopup = 5,
	cUPointsET = 6,
	cVPointsET = 7,
	cLoopTimeET = 8,
	cXScaleET = 9,
	cYScaleET = 10,
	cZScaleET = 11
};

typedef struct PluginDataRec {
	double frameTime;
	double frameTimeDelta;
	double uStep, vStep;
	double t0, t1, t2;
	Matrix4 theMatrix;
	Matrix4 theRotMatrix;
	Matrix4 theBlurMatrix;
	long frameIndex;
	long vertexFlags;
} PluginDataRec, *PluginDataPtr;

typedef struct ControlDataRec {
	ulong ownerType;		/* The owner plugin's unique type */
	long ownerVersion;		/* The owner plugin's version number */
	double meshXScale;		/* The mesh's x scale */
	double meshYScale;		/* The mesh's y scale */
	double meshZScale;		/* The mesh's z scale */
	double meshTimeScale;	/* The time scale factor */
	long meshFlags;			/* The control flags for the mesh */
	long meshMode;			/* The type of mesh to be generated */
	ARGB meshMainColor;		/* The default mesh color */
	long meshUPoints;		/* The mesh's u  resolution */
	long meshVPoints;		/* The mesh's v resolution */
	long filler;			/* Filler bytes for 64 bit alignment */
} ControlDataRec, *ControlDataPtr;

static void InitGlobals(boolean initFlag);
static void FixControl(ControlDataPtr theControlRec);
static void UpdateControl(char* oldControl, ControlDataPtr newControl);
static void DefaultControl(ControlDataRec *theControlRec);
static void DefaultData(PluginDataRec *theData);
static long CheckData(PluginInfoRec *theInfo, boolean setControlData, boolean setPluginData);
static void SendModelReset(PluginInfoRec *theInfo);
static boolean StatusTime(void);
static void UpdateStatus(float thePos, long theValue);
static long AddVertex(ARGB color, XCoordPtr normal, XCoordPtr position, XCoordPtr blurPosition, XCoordPtr texturePosition);
static void AddFacet(ARGB color, short numVertices, long v0, long v1, long v2, long v3);
static void SetupTime(double *t);
static void SetupMesh(void);
static void CalculateMeshPosition(double t, double u, double v, XCoordinate *theCoordinate);
static void CalculateMeshTexturePosition(double t, double u, double v, XCoordinate *theCoordinate);
static long CalculateMesh(long uIndex, long vIndex);
static void OutputMesh(void);
static long DoInitialize(long theCommandRecSize, void* theCommandRec);
static long DoFinish(long theCommandRecSize, void* theCommandRec);
static long DoInformation(long theCommandRecSize, void* theCommandRec);
static long DoCheck(long theCommandRecSize, void* theCommandRec);
static long DoDescribe(long theCommandRecSize, void* theCommandRec);
static long DoInterface(long theCommandRecSize, void* theCommandRec);
static long DoSetup(long theCommandRecSize, void* theCommandRec);
static long DoGINF(long theCommandRecSize, void* theCommandRec);
static long DoGATR(long theCommandRecSize, void* theCommandRec);
static long DoGBLR(long theCommandRecSize, void* theCommandRec);
static long DoTMAP(long theCommandRecSize, void* theCommandRec);
static long DoBMAP(long theCommandRecSize, void* theCommandRec);
static long DoRMAP(long theCommandRecSize, void* theCommandRec);
static long DoGenerate(long theCommandRecSize, void* theCommandRec);

static ulong globalValidation1, globalValidation2;
static long result;
static PluginDataPtr theData;
static ControlDataPtr theControl;
static EIPluginRoutine hostProc;
static char* hostReference;
static HostStatusRec theStatusRec;

static void InitGlobals(boolean initFlag)
{
	if (initFlag || (globalValidation1 != pluginType) || (globalValidation2 != pluginModelType)) {
		globalValidation1 = pluginType;
		globalValidation2 = pluginModelType;
		InitUtils();
	}
}  /* InitGlobals */

static void FixControl(ControlDataPtr theControlRec)
{
	theControlRec->ownerType = pluginType;
	theControlRec->ownerVersion = pluginVersion;

	FixField(theControlRec->meshFlags);
	FixField(theControlRec->meshMode);
	FixField(theControlRec->meshUPoints);
	FixField(theControlRec->meshVPoints);
	FixField(theControlRec->meshXScale);
	FixField(theControlRec->meshYScale);
	FixField(theControlRec->meshZScale);
	FixField(theControlRec->meshTimeScale);
}  /* FixControl */

static void UpdateControl(char* oldControl, ControlDataPtr newControl)
{
	newControl->ownerType = pluginType;
	newControl->ownerVersion = pluginVersion;
	CopyField(oldControl + oldMeshFlags, newControl->meshFlags);
	CopyField(oldControl + oldMeshMode, newControl->meshMode);
	CopyField(oldControl + oldMeshMainColor, newControl->meshMainColor);
	CopyField(oldControl + oldMeshUPoints, newControl->meshUPoints);
	CopyField(oldControl + oldMeshVPoints, newControl->meshVPoints);
	CopyField(oldControl + oldMeshXScale, newControl->meshXScale);
	CopyField(oldControl + oldMeshYScale, newControl->meshYScale);
	CopyField(oldControl + oldMeshZScale, newControl->meshZScale);
	CopyField(oldControl + oldMeshTimeScale, newControl->meshTimeScale);
}  /* UpdateControl */

static void DefaultControl(ControlDataRec *theControlRec)
{
	theControlRec->ownerType = pluginType;
	theControlRec->ownerVersion = pluginVersion;
	theControlRec->meshFlags = defMeshFlags;
	theControlRec->meshMode = defMeshMode;
	theControlRec->meshMainColor.argb.a = defReferenceA;
	theControlRec->meshMainColor.argb.r = defReferenceR;
	theControlRec->meshMainColor.argb.g = defReferenceG;
	theControlRec->meshMainColor.argb.b = defReferenceB;
	theControlRec->meshUPoints = defMeshUPoints;
	theControlRec->meshVPoints = defMeshVPoints;
	theControlRec->meshXScale = defMeshXScale;
	theControlRec->meshYScale = defMeshYScale;
	theControlRec->meshZScale = defMeshZScale;
	theControlRec->meshTimeScale = defMeshTimeScale;
}  /* DefaultControl */

static void DefaultData(PluginDataRec *theData)
{
	theData->frameIndex = 0;
	theData->frameTime = 0.0;
	theData->frameTimeDelta = defFrameTimeDelta;
	theData->vertexFlags = 0;
	theData->uStep = 0.0;
	theData->vStep = 0.0;
	theData->t0 = 0.0;
	theData->t1 = 0.0;
	theData->t2 = 0.0;
	InitMatrix4(&theData->theMatrix);
	InitMatrix4(&theData->theRotMatrix);
	InitMatrix4(&theData->theBlurMatrix);
}  /* DefaultData */

static long CheckData(PluginInfoPtr theInfo, boolean setControlData, boolean setPluginData)
{
	HostAllocateRec theAllocate;
	long newPluginDataSize, newControlDataSize;
	boolean newPluginData, newControlData;
	ControlDataRec theControlRec;
	PluginDataRec theDataRec;
	ControlDataPtr theControl;
	long theResult;

	theResult = 0;

	/* Change the size of the plugin data block or control data block if the size is too small.  Old version */
	/* control records are updated to the current version and reverse byte ordering is corrected. */

	theControl = (ControlDataPtr)theInfo->theControlData;
	if ((!setControlData) && (theInfo->theControlDataSize == oldControlSize) && (theControl->ownerType == oldPluginType) && (theControl->ownerVersion == oldPluginVersion)) {
		UpdateControl((char*)theControl, &theControlRec);
		setControlData = 1;
	} else if ((!setControlData) && (theInfo->theControlDataSize == sizeof(ControlDataRec))
		&& ((theControl->ownerType == pluginType) || (theControl->ownerType == REVERSELONG(pluginType)))
		&& ((theControl->ownerVersion == pluginVersion) || (theControl->ownerVersion == REVERSELONG(pluginVersion)))) {
		if (theControl->ownerVersion == REVERSELONG(pluginVersion))
			FixControl(theControl);
		theControlRec = *theControl;
	} else {
		setControlData = 1;
		DefaultControl(&theControlRec);
	}
	if (setPluginData || (theInfo->thePluginDataSize < sizeof(PluginDataRec))) {
		setPluginData = 1;
		DefaultData(&theDataRec);
	} else
		theDataRec = *(PluginDataPtr)theInfo->thePluginData;

	newControlDataSize = sizeof(ControlDataRec);
	newControlData = (theInfo->theControlDataSize < newControlDataSize || theInfo->theControlDataSize > FixAlignment(newControlDataSize));
	newPluginDataSize = sizeof(PluginDataRec);
	newPluginData = (theInfo->thePluginDataSize < newPluginDataSize || theInfo->thePluginDataSize > FixAlignment(newPluginDataSize));

	if (newPluginData || newControlData) {
		theAllocate.theReference = theInfo->theHostReference;
		theAllocate.theControlDataSize = newControlDataSize;
		theAllocate.theControlData = theInfo->theControlData;
		theAllocate.thePluginDataSize = newPluginDataSize;
		theAllocate.thePluginData = theInfo->thePluginData;
		theResult = theInfo->theHostProc(hostAllocate, sizeof(HostAllocateRec), &theAllocate);
		theInfo->theControlData = theAllocate.theControlData;
		theInfo->theControlDataSize = theAllocate.theControlDataSize;
		theInfo->thePluginData = theAllocate.thePluginData;
		theInfo->thePluginDataSize = theAllocate.thePluginDataSize;
	}

	if (theResult == 0) {
	
		/* Setup the control data to the default values if an incorrect sized block was passed to the plugin */
	
		if (newControlData || setControlData)
			*(ControlDataPtr)theInfo->theControlData = theControlRec;
	
		/* Setup the plugin private data */
	
		if (newPluginData || setPluginData)
			*(PluginDataPtr)theInfo->thePluginData = theDataRec;
	}
	return theResult;
}  /* CheckData */

/* Reset the host's model data */

static void SendModelReset(PluginInfoRec *theInfo)
{
	HostModelResetRec theModelReset;
	long theResult;

	theModelReset.theReference = theInfo->theHostReference;
	theResult = theInfo->theHostProc(hostModelReset, sizeof(HostModelResetRec), &theModelReset);
}  /* SendModelReset */

static boolean StatusTime(void)
{
	return ((theStatusRec.theStatusTick == 0) || (TICKCOUNT() >= theStatusRec.theStatusTick));
}  /* StatusTime */

static void UpdateStatus(float thePos, long theValue)
{
	if (result == 0) {
		theStatusRec.theStatusValue = theValue;
		theStatusRec.theStatusPos = thePos;
		result = hostProc(hostStatus, sizeof(HostStatusRec), &theStatusRec);
	}
}  /* UpdateStatus */

static long AddVertex(ARGB color, XCoordPtr normal, XCoordPtr position, XCoordPtr blurPosition, XCoordPtr texturePosition)
{
	HostModelAddVertexRec theAddVertex;

	if (result == 0) {
		theAddVertex.theReference = hostReference;
		theAddVertex.theColor = color;
		TransformPoint(&normal->x, &normal->y, &normal->z, &theData->theRotMatrix);
		theAddVertex.theNormal.x = normal->x;
		theAddVertex.theNormal.y = normal->y;
		theAddVertex.theNormal.z = normal->z;
		TransformPoint(&position->x, &position->y, &position->z, &theData->theMatrix);
		theAddVertex.thePosition.x = position->x;
		theAddVertex.thePosition.y = position->y;
		theAddVertex.thePosition.z = position->z;
		TransformPoint(&blurPosition->x, &blurPosition->y, &blurPosition->z, &theData->theBlurMatrix);
		theAddVertex.theBlurPosition.x = blurPosition->x;
		theAddVertex.theBlurPosition.y = blurPosition->y;
		theAddVertex.theBlurPosition.z = blurPosition->z;
		theAddVertex.theTexturePosition.x = texturePosition->x;
		theAddVertex.theTexturePosition.y = texturePosition->y;
		theAddVertex.theTexturePosition.z = texturePosition->z;
		result = hostProc(hostModelAddVertex, sizeof(HostModelAddVertexRec), &theAddVertex);
	}
	if (result == 0)
		return (theAddVertex.theIndex);
	else
		return 0;
}  /* AddVertex */

static void AddFacet(ARGB color, short numVertices, long v0, long v1, long v2, long v3)
{
	HostModelAddFacetRec theAddFacet;

	if (result == 0) {
		theAddFacet.theReference = hostReference;
		theAddFacet.theColor = color;
		theAddFacet.theNumVertices = numVertices;
		theAddFacet.theVertices[0] = v0;
		theAddFacet.theVertices[1] = v1;
		theAddFacet.theVertices[2] = v2;
		theAddFacet.theVertices[3] = v3;
		result = hostProc(hostModelAddFacet, sizeof(HostModelAddFacetRec), &theAddFacet);
	}
}  /* AddFacet */

static void SetupTime(double *t)
{
	*t /= theControl->meshTimeScale;
	switch (theControl->meshMode) {

	  case meshSphereWrap:
		*t *= 0.5;
		*t = 2.0 * (*t - floor(*t));
		if (*t > 1.0)
			*t = 2.0 - *t;
		break;
	}
}  /* SetupTime */

static void SetupMesh(void)
{
	theData->uStep = 2.0 / (theControl->meshUPoints - 1);
	theData->vStep = 2.0 / (theControl->meshVPoints - 1);
	theData->t0 = theData->frameTime - theData->frameTimeDelta;
	theData->t1 = theData->frameTime;
	theData->t2 = theData->frameTime + theData->frameTimeDelta;
	SetupTime(&theData->t0);
	SetupTime(&theData->t1);
	SetupTime(&theData->t2);
}  /* SetupMesh */

static void CalculateMeshPosition(double t, double u, double v, XCoordinate *theCoordinate)
{
	double theta, theta2, r, squeeze;

	switch (theControl->meshMode) {

	  case meshSineWave:   /* Sinewave moving towards the center of a plane */
		theCoordinate->x = theControl->meshXScale * u;
		theCoordinate->y = theControl->meshYScale * v;
		theCoordinate->z = theControl->meshZScale * 0.1 * SIN((SQRT(u * u + v * v) + t) * PI2);
		break;

	  case meshBreathingCylinder:   /* Cylinder breathing in and out */
		theCoordinate->x = theControl->meshXScale * SIN(PI * u) * (1.0 + 0.5 * COS(PI05 * v) * COS(PI2 * t));
		theCoordinate->y = theControl->meshYScale * v;
		theCoordinate->z = theControl->meshZScale * COS(PI * u) * (1.0 + 0.5 * COS(PI05 * v) * COS(PI2 * t));
		break;

	  case meshScrewThread:   /* Screw thread reversing itself */
		theCoordinate->x = theControl->meshXScale * SIN(PI * u) * (1.0 + 0.25 * COS((4.0 * v + u) * PI) * COS(PI2 * t));
		theCoordinate->y = theControl->meshYScale * v;
		theCoordinate->z = theControl->meshZScale * COS(PI * u) * (1.0 + 0.25 * COS((4.0 * v + u) * PI) * COS(PI2 * t));
		break;

	  case meshSphereWrap:   /* Plane wrapping around a sphere */
		theta = PI * t;
		theta2 = 0.5 * theta;
		r = SIN(theta2);
		if (FABS(r) < 1.0e-10) {
			theCoordinate->x = theControl->meshXScale * 2.0 * u;
			theCoordinate->y = theControl->meshYScale * v;
			theCoordinate->z = 0.0;
		} else {
			r = 1.0 / r;   /* Keep the Y axis the same */
			squeeze = COS(v * theta2);
			theCoordinate->x = theControl->meshXScale * r * squeeze * SIN(u * theta);
			theCoordinate->y = theControl->meshYScale * r * SIN(v * theta2);
			theCoordinate->z = theControl->meshZScale * r * (1.0 - squeeze * COS(u * theta));
		}
		break;

	  default:	/* Flat mesh */
		theCoordinate->x = theControl->meshXScale * u;
		theCoordinate->y = theControl->meshYScale * v;
		theCoordinate->z = 0.0;
		break;
	}
}  /* CalculateMeshPosition */

static void CalculateMeshTexturePosition(double t, double u, double v, XCoordinate *theCoordinate)
{
	t = t;		/* Avoid an unused variable warning */
	switch (theControl->meshMode) {
	  case meshSineWave:   /* Sinewave moving towards the center of a plane */
		theCoordinate->x = theControl->meshXScale * u;
		theCoordinate->y = theControl->meshYScale * v;
		theCoordinate->z = 0.0;
		break;

	  case meshBreathingCylinder:   /* Cylinder breathing in and out */
		theCoordinate->x = theControl->meshXScale * SIN(PI * u);
		theCoordinate->y = theControl->meshYScale * v;
		theCoordinate->z = theControl->meshZScale * COS(PI * u);
		break;

	  case meshScrewThread:   /* Screw thread reversing itself */
		theCoordinate->x = theControl->meshXScale * SIN(PI * u);
		theCoordinate->y = theControl->meshYScale * v;
		theCoordinate->z = theControl->meshZScale * COS(PI * u);
		break;

	  case meshSphereWrap:   /* Plane wrapping around a sphere */
		theCoordinate->x = theControl->meshXScale * 2.0 * u;
		theCoordinate->y = theControl->meshYScale * v;
		theCoordinate->z = 0.0;
		break;
		
	  default:	/* Flat mesh */
		theCoordinate->x = theControl->meshXScale * u;
		theCoordinate->y = theControl->meshYScale * v;
		theCoordinate->z = 0.0;
		break;
	}
}  /* CalculateMeshTexturePosition */

static long CalculateMesh(long uIndex, long vIndex)
{
	double u, v, distance, a1, b1, c1, a2, b2, c2, aNum, bNum, cNum;
	XCoordinate position1, position2, position3, blurPosition, texturePosition, normal;

	/* Calculate the mesh u & v parameters */

	u = theData->uStep * uIndex - 1.0;
	v = theData->vStep * vIndex - 1.0;

	/* Calculate the mesh position */

	CalculateMeshPosition(theData->t1, u, v, &position1);

	/* Calculate the mesh motion blur position */

	if (BTST(theData->vertexFlags, pluginVertexBlurFlag)) {
		CalculateMeshPosition(theData->t0, u, v, &blurPosition);
		CalculateMeshPosition(theData->t2, u, v, &position2);
		
		/* The length of the motion blur vector is based upon the distance the point travels from the */
		/* current frame to the next frame.  The direction is based upoin the difference between the */
		/* position of the last frame and the position of the next frame */
		
		blurPosition.x -= position2.x;
		blurPosition.y -= position2.y;
		blurPosition.z -= position2.z;

		position2.x -= position1.x;
		position2.y -= position1.y;
		position2.z -= position1.z;

		distance = SQRT(SQR(blurPosition.x) + SQR(blurPosition.y) + SQR(blurPosition.z));
		if (distance == 0.0) {
			blurPosition.x = position1.x;
			blurPosition.y = position1.y;
			blurPosition.z = position1.z;
		} else {
			distance = SQRT(SQR(position2.x) + SQR(position2.y) + SQR(position2.z)) / distance;
			blurPosition.x = position1.x + distance * blurPosition.x;
			blurPosition.y = position1.y + distance * blurPosition.y;
			blurPosition.z = position1.z + distance * blurPosition.z;
		}
	} else
		blurPosition = position1;

	/* Calculate the mesh texture position */

	if (BTST(theData->vertexFlags, pluginVertexTextureFlag))
		CalculateMeshTexturePosition(theData->t1, u, v, &texturePosition);
	else
		texturePosition = position1;

	/* Calculate the mesh surface normal */

	if (BTST(theData->vertexFlags, pluginVertexNormalFlag)) {
		CalculateMeshPosition(theData->t1, u + normalDelta, v, &position2);
		CalculateMeshPosition(theData->t1, u, v + normalDelta, &position3);
		a1 = position1.x - position3.x;
		b1 = position1.y - position3.y;
		c1 = position1.z - position3.z;
		a2 = position2.x - position1.x;
		b2 = position2.y - position1.y;
		c2 = position2.z - position1.z;
		aNum = b1 * c2 - b2 * c1;
		bNum = c1 * a2 - c2 * a1;
		cNum = a1 * b2 - a2 * b1;
		distance = SQRT(SQR(aNum) + SQR(bNum) + SQR(cNum));
		if (distance == 0.0) {
			normal.x = 0.0;
			normal.y = 0.0;
			normal.z = 0.0;
			return (AddVertex(theControl->meshMainColor, &normal, &position1, &blurPosition, &texturePosition));
		}
		distance = 1.0 / distance;
		normal.x = aNum * distance;
		normal.y = bNum * distance;
		normal.z = cNum * distance;
		return (AddVertex(theControl->meshMainColor, &normal, &position1, &blurPosition, &texturePosition));
	}

	normal.x = 0.0;
	normal.y = 0.0;
	normal.z = 0.0;
	
	/* Create a new vertex */

	return (AddVertex(theControl->meshMainColor, &normal, &position1, &blurPosition, &texturePosition));
}  /* CalculateMesh */

static void OutputMesh(void)
{
	long uIndex, vIndex, theIndex, count, meshTotal;

	meshTotal = theControl->meshUPoints * theControl->meshVPoints;
	theStatusRec.theReference = hostReference;
	theStatusRec.theStatusPos = 0.0;
	PStrCpy(theStatusRec.theStatusUnit, PS("Point"));
	PStrCpy(theStatusRec.theStatusInfo, PS("Creating Mesh"));
	theStatusRec.theStatusMin = 0;
	theStatusRec.theStatusMax = meshTotal;
	theStatusRec.theStatusValue = 0;
	theStatusRec.theStatusTick = 0;
	UpdateStatus(0.0, 0);

	/* Convert mesh to facets */

	count = 0;
	vIndex = 0;
	while ((vIndex < theControl->meshVPoints) && (result == 0)) {
		uIndex = 0;
		while ((uIndex < theControl->meshUPoints) && (result == 0)) {
			count++;
			theIndex = CalculateMesh(uIndex, vIndex);
			if (uIndex > 0 && vIndex > 0)
				AddFacet(theControl->meshMainColor, 4, theIndex - theControl->meshUPoints - 1, theIndex - theControl->meshUPoints, theIndex, theIndex - 1);
			uIndex++;
			if (StatusTime())
				UpdateStatus((double)count / meshTotal, count);
		}
		vIndex++;
	}
	UpdateStatus(1.0, meshTotal);
}  /* OutputMesh */

/* Initialize the plugin's data structures and perform memory allocation */

static long DoInitialize(long theCommandRecSize, void* theCommandRec)
{
	PluginInitializePtr theParameters;
	long theResult;

	InitGlobals(1);
	theParameters = (PluginInitializePtr)theCommandRec;
	if ((theCommandRecSize < sizeof(PluginInitializeRec)) || (theParameters->theInfo.theHostProc == NULL))
		theResult = pluginParameterErr;
	else
		theResult = CheckData(&theParameters->theInfo, 0, 1);
	return theResult;
}  /* DoInitialize */

/* Finish up anything which needs it.  This is the last time that the plugin will be called */

static long DoFinish(long theCommandRecSize, void* theCommandRec)
{
	PluginFinishPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginFinishPtr)theCommandRec;
	if ((theCommandRecSize < sizeof(PluginFinishRec)) || (theParameters->theInfo.theHostProc == NULL))
		theResult = pluginParameterErr;
	FinishUtils();
	return theResult;
}  /* DoFinish */

/* Get information about the plugin */

static long DoInformation(long theCommandRecSize, void* theCommandRec)
{
	PluginInformationPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginInformationPtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginInformationRec))
		theResult = pluginParameterErr;
	else {
		theParameters->thePluginType = pluginType;
		theParameters->thePluginVersion = pluginVersion;
		PStrCpy(theParameters->thePluginName, pluginName);
		PStrCpy(theParameters->thePluginCopyright, pluginCopyright);
		PStrCpy(theParameters->thePluginAuthor, pluginAuthor);
		PStrCpy(theParameters->thePluginDescription, pluginDescription);
		theParameters->thePluginPermission = pluginPermission;
	}
	return theResult;
}  /* DoInformation */

/* Check to see if a particular command is supported */

static long DoCheck(long theCommandRecSize, void* theCommandRec)
{
	PluginCheckPtr theParameters;
	
	theParameters = (PluginCheckPtr)theCommandRec;
	if (theCommandRecSize >= sizeof(PluginCheckRec))
		switch (theParameters->theCommand) {
			case pluginInitialize:
			case pluginFinish:
			case pluginInformation:
			case pluginCheck:
			case pluginDescribe:
			case pluginInterface:
			case pluginModelSetup:
			case pluginModelGINF:
			case pluginModelGATR:
			case pluginModelGBLR:
			case pluginModelTMAP:
			case pluginModelBMAP:
			case pluginModelRMAP:
			case pluginModelGenerate:
				return pluginNoErr;
		}
	return pluginParameterErr;
}  /* DoCheck */

/* Describe an error number */

static long DoDescribe(long theCommandRecSize, void* theCommandRec)
{
	PluginDescribePtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginDescribePtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginDescribeRec))
		theResult = pluginParameterErr;
	else
		DescribePluginError(theParameters->theResult, theParameters->theDescription);
	return theResult;
}  /* DoDescribe */

/* Display the plugin's user interface and edit parameters */

static long DoInterface(long theCommandRecSize, void* theCommandRec)
{
	EI_ResourceFile			*resourceFile;
	EI_Dialog				*theDialog;
	long					temp;
	boolean					changeFlag;
	ControlDataPtr			theControl;
	PluginInterfacePtr		theParameters;
	long					theResult;

	theResult = pluginNoErr;
	theParameters = (PluginInterfacePtr)theCommandRec;
	if ((theCommandRecSize < sizeof(PluginInterfaceRec)) || (theParameters->theInfo.theControlData == NULL) || (theParameters->theInfo.theControlDataSize < sizeof(ControlDataRec)))
		theResult = pluginParameterErr;
	else
		theResult = CheckData(&theParameters->theInfo, 0, 0);
	if (theResult == 0) {
		resourceFile = OpenResourceFile("Mesh.rsc");
		if (resourceFile == NULL)
			theResult = pluginResourceErr;
	}
	if (theResult == 0) {
		theControl = (ControlDataPtr)theParameters->theInfo.theControlData;
		theDialog = EI_MakeDialog(resourceFile, interfaceDialogID);
		if (theDialog != NULL) {
			SetTheINum(theDialog, cUPointsET, theControl->meshUPoints);
			SetTheINum(theDialog, cVPointsET, theControl->meshVPoints);
			SetTheIReal(theDialog, cXScaleET, theControl->meshXScale, 2);
			SetTheIReal(theDialog, cYScaleET, theControl->meshYScale, 2);
			SetTheIReal(theDialog, cZScaleET, theControl->meshZScale, 2);
			SetTheIReal(theDialog, cLoopTimeET, theControl->meshTimeScale, 2);
			SetTheIPopUp(theDialog, cMeshTypePopup, theControl->meshMode);
			if (EI_ExecuteDialog(theDialog)) {
				changeFlag = 0;
				temp = GetTheINum(theDialog, cUPointsET);
				if (temp < 2)
					temp = 2;
				SetLongInt(&theControl->meshUPoints, temp, &changeFlag);
				temp = GetTheINum(theDialog, cVPointsET);
				if (temp < 2)
					temp = 2;
				SetLongInt(&theControl->meshVPoints, temp, &changeFlag);
				SetDouble(&theControl->meshXScale, GetTheIReal(theDialog, cXScaleET), &changeFlag);
				SetDouble(&theControl->meshYScale, GetTheIReal(theDialog, cYScaleET), &changeFlag);
				SetDouble(&theControl->meshZScale, GetTheIReal(theDialog, cZScaleET), &changeFlag);
				SetDouble(&theControl->meshTimeScale, GetTheIReal(theDialog, cLoopTimeET), &changeFlag);
				SetLongInt(&theControl->meshMode, GetTheIPopUp(theDialog, cMeshTypePopup), &changeFlag);
				if (changeFlag)
					SendModelReset(&theParameters->theInfo);
			} else
				theResult = pluginCancelErr;
			EI_DestroyDialog(theDialog);
		}
		EI_CloseResourceFile(resourceFile);
	}
	return theResult;
}  /* DoInterface */

/* Setup the model generation variables */

static long DoSetup(long theCommandRecSize, void* theCommandRec)
{
	PluginModelSetupPtr theParameters;
	PluginInfoPtr theInfo;
	ControlDataPtr theControl;
	PluginDataPtr theData;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelSetupPtr)theCommandRec;
	theInfo = &theParameters->theInfo;
	if ((theCommandRecSize < sizeof(PluginModelSetupRec)) || (theInfo->theHostProc == NULL))
		theResult = pluginParameterErr;
	else
		theResult = CheckData(&theParameters->theInfo, 0, 0);
	if (theResult == 0) {
		theControl = (ControlDataPtr)theInfo->theControlData;
		theData = (PluginDataPtr)theInfo->thePluginData;
		theData->frameIndex = theParameters->theFrameIndex;
		theData->frameTime = theParameters->theFrameTime;
		theData->frameTimeDelta = theParameters->theFrameTimeDelta;
		theData->theMatrix = theParameters->theGroupMatrix;
		theData->theRotMatrix = theParameters->theGroupRotMatrix;
		theData->theBlurMatrix = theParameters->theGroupBlurMatrix;
		SendModelReset(&theParameters->theInfo);
	}
	return theResult;
}  /* DoSetup */

/* Setup the group GINF record */

static long DoGINF(long theCommandRecSize, void* theCommandRec)
{
	PluginModelGINFPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelGINFPtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginModelGINFRec))
		theResult = pluginParameterErr;
	else
		theParameters->theChangeFlag = 0;
	return theResult;
}  /* DoGINF */

/* Setup the group GATR record */

static long DoGATR(long theCommandRecSize, void* theCommandRec)
{
	PluginModelGATRPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelGATRPtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginModelGATRRec))
		theResult = pluginParameterErr;
	else
		theParameters->theChangeFlag = 0;
	return theResult;
}  /* DoGATR */

/* Setup the group GBLR record */

static long DoGBLR(long theCommandRecSize, void* theCommandRec)
{
	PluginModelGBLRPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelGBLRPtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginModelGBLRRec))
		theResult = pluginParameterErr;
	else
		theParameters->theChangeFlag = 0;
	return theResult;
}  /* DoGBLR */

/* Setup the group TMAP record */

static long DoTMAP(long theCommandRecSize, void* theCommandRec)
{
	PluginModelTMAPPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelTMAPPtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginModelTMAPRec))
		theResult = pluginParameterErr;
	else
		theParameters->theChangeFlag = 0;
	return theResult;
}  /* DoTMAP */

/* Setup the group BMAP record */

static long DoBMAP(long theCommandRecSize, void* theCommandRec)
{
	PluginModelBMAPPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelBMAPPtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginModelBMAPRec))
		theResult = pluginParameterErr;
	else
		theParameters->theChangeFlag = 0;
	return theResult;
}  /* DoBMAP */

/* Setup the group RMAP record */

static long DoRMAP(long theCommandRecSize, void* theCommandRec)
{
	PluginModelRMAPPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelRMAPPtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginModelRMAPRec))
		theResult = pluginParameterErr;
	else
		theParameters->theChangeFlag = 0;
	return theResult;
}  /* DoRMAP */

/* Generate the coordinates and polygons for the group */

static long DoGenerate(long theCommandRecSize, void* theCommandRec)
{
	PluginModelGeneratePtr theParameters;
	PluginInfoPtr theInfo;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelGeneratePtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginModelGenerateRec))
		theResult = pluginParameterErr;
	else {
		theInfo = &theParameters->theInfo;
		theResult = CheckData(&theParameters->theInfo, 0, 0);
		if (theResult == 0) {
			theControl = (ControlDataRec *)theInfo->theControlData;
			theData = (PluginDataRec *)theInfo->thePluginData;
			hostProc = theInfo->theHostProc;
			hostReference = theInfo->theHostReference;
			theData->vertexFlags = theParameters->theVertexFlags;
		
			/* Setup the mesh as a function of the present time */
		
			result = theResult;
			SetupMesh();
		
			/* Write out the mesh vertices and facets */
		
			OutputMesh();
			theResult = result;
		}
	}
	return theResult;
}  /* DoGenerate */

/* The main function is called by the host application for all plugin commands */

EXPDLL long Main(ulong theCommand, long theCommandRecSize, void* theCommandRec)
{
	long theResult;
	
	/* Initialize the plugin's global variables if they have not been initialize previously */

	InitGlobals(0);

	/* Perform the requested command */

	switch (theCommand) {
		case pluginInitialize:		theResult = DoInitialize(theCommandRecSize, theCommandRec);		break;
		case pluginFinish:			theResult = DoFinish(theCommandRecSize, theCommandRec);			break;
		case pluginInformation:		theResult = DoInformation(theCommandRecSize, theCommandRec);	break;
		case pluginCheck:			theResult = DoCheck(theCommandRecSize, theCommandRec);			break;
		case pluginDescribe:		theResult = DoDescribe(theCommandRecSize, theCommandRec);		break;
		case pluginInterface:		theResult = DoInterface(theCommandRecSize, theCommandRec);		break;
		case pluginModelSetup:		theResult = DoSetup(theCommandRecSize, theCommandRec);			break;
		case pluginModelGINF:		theResult = DoGINF(theCommandRecSize, theCommandRec);			break;
		case pluginModelGATR:		theResult = DoGATR(theCommandRecSize, theCommandRec);			break;
		case pluginModelGBLR:		theResult = DoGBLR(theCommandRecSize, theCommandRec);			break;
		case pluginModelTMAP:		theResult = DoTMAP(theCommandRecSize, theCommandRec);			break;
		case pluginModelBMAP:		theResult = DoBMAP(theCommandRecSize, theCommandRec);			break;
		case pluginModelRMAP:		theResult = DoRMAP(theCommandRecSize, theCommandRec);			break;
		case pluginModelGenerate:	theResult = DoGenerate(theCommandRecSize, theCommandRec);		break;
		default:					theResult = pluginCommandErr;									break;
	}
	return theResult;
}  /* Mesh */
