/*
Copyright 1995-1998 Electric Image, Inc. This source code may be included in
commercial, freeware, or shareware products.  This copyright notice must be
included in its original form in all such products. This source code may not be
offered for sale or distributed by any entity without the written permission of
Electric Image, Inc.
*/

/* This code presents an example of an implicit surface model plugin. */

#include "EI_Dialog.h"
#include "EI_FileRef.h"
#include "EI_ResourceFile.h"
#include "PluginClient.h"
#include "PluginFPU.h"
#include "PluginUtils.h"
#include "PluginMatrix.h"
#include "PString.h"
#include "Implicit.h"

#include <stddef.h>

#define pluginType			QUADBYTE('B', 'L', 'O', 'B')		/* The owner plugin's unique type */
#define pluginVersion		0x00000001				/* The owner plugin's version number 0:0:version:revision */
#define pluginName			PS("Mr. Blobby")
#define pluginCopyright		PS("Copyright  1993-1998 ElectricImage, Inc.")
#define pluginAuthor		PS("Mark Granger|Electric Image, Inc.|117 East Colorado Boulevard, Suite 300|Pasadena, Ca 91105|(818) 577-1627")
#define pluginDescription	PS("This model plugin generates an implicit surface.")
#define pluginPermission	BIT(pluginSensitivityFlag) + BIT(pluginChildSensitivityFlag)

#define interfaceDialogID	16001

enum {
	cTextureMapRB = 13,
	cInterpolatedColorRB = 14,
	cMeshWithRB = 15,
	cGenerateRB = 16,
	cBlobThresholdET = 17,
	cVertexBlobSizeET = 18,
	cFacetBlobScaleET = 19,
	cEIMeshResolutionET = 21,
	cCameraMeshResolution = 22
};

/*
#define okItem				1
#define cancelItem			2
#define cubesItem			6
#define tetrahedronsItem	7
#define sphereItem			9
#define torusItem			10
#define spheresItem			11
#define groupsItem			12
#define verticesItem		13
#define facetsItem			14
#define vertexSizeItem		15
#define facetScaleItem		16
#define blobThresholdItem	17
#define EIResolutionItem	18
#define CameraResolutionItem	19
#define RGBItem				29
#define HSVItem				30
#define textureBlobItem		31
#define textureModelItem	32
*/

#define sphereSurface		0
#define spheresSurface		1
#define torusSurface		2
#define groupsSurface		3
#define verticesSurface		4
#define facetsSurface		5

#define tetrahedronsFlag	31
#define hsvFlag				30
#define textureModelFlag	29

#define defSurfaceType		groupsSurface
#define defMode				0
#define defResolutionEI		5
#define defResolutionCamera	10
#define defVertexSize		10.0
#define defFacetScale		1.0
#define defBlobThreshold	0.5

typedef struct PluginDataRec {
	Matrix4 theMatrix;
	Matrix4 theNormalMatrix;
	Matrix4 theBlurMatrix;
	Matrix4 theInvMatrix;
	long facetCount;
	long firstFacet;
	long theDataSize;
} PluginDataRec, *PluginDataPtr;

typedef struct ControlDataRec {
	unsigned long ownerType;			/* The owner plugin's unique type */
	long ownerVersion;		/* The owner plugin's version number */
	long surfaceType;		/* The implicit surface to generate */
	ulong mode;				/* The tesselation & color mode */
	long resolutionEI;		/* The tesselation resolution */
	long resolutionCamera;	/* The tesselation resolution */
	float vertexSize;			/* The size of the vertex blobs */
	float facetScale;			/* The scale of the facet blobs */
	float blobThreshold;			/* The blob value theshold */
} ControlDataRec, *ControlDataPtr;

typedef struct FacetRec {
	long next;
	long v1;
	long v2;
	long v3;
} FacetRec, *FacetPtr;

static char* AllocateBotMem(long nBytes);
static double Torus(double x, double y, double z);
static double ColorTorus(CoordPtr position, CoordPtr blur, CoordPtr texture, ARGBPtr theColor);
static double Sphere(double radius, double x, double y, double z);
static double ASphere(double x, double y, double z);
static double ColorASphere(CoordPtr position, CoordPtr blur, CoordPtr texture, ARGBPtr theColor);
static double ThreeSpheres(double x, double y, double z);
static double ColorThreeSpheres(CoordPtr position, CoordPtr blur, CoordPtr texture, ARGBPtr theColor);
static double Blobs(double x, double y, double z);
static double ColorBlobs(CoordPtr position, CoordPtr blur, CoordPtr texture, ARGBPtr theColor);
static FacetPtr GetNewFacet(void);
static ARGB HSVToARGB(ARGB HSVvalues);
static ARGB ARGBToHSV(ARGB RGBvalues);
static unsigned long GetHostCreator(void);
static boolean GetGroup(long groupIndex, char* *groupReference, ulong *groupVertexFlags, long *groupVertices, long *groupFacets);
static void DeleteGroup(char* groupReference);
static void GetVertex(char* groupReference, long theVertex, DCoordPtr thePosition, DCoordPtr theBlurPosition, ARGBPtr theColor);
static void GetFacet(char* groupReference, long facetIndex, long *numVertices, Vertices theVertices, ARGBPtr theColor);
static void BuildGroupsBlobList(float *xc, float *yc, float *zc, float *scale);
static void BuildVerticesBlobList(float *xc, float *yc, float *zc, float *scale);
static void BuildFacetsBlobList(float *xc, float *yc, float *zc, float *scale);
static void InitGlobals(boolean initFlag);
static void FixControl(ControlDataPtr theControlRec);
static void DefaultControl(ControlDataPtr theControlRec);
static void DefaultData(PluginDataPtr theData);
static void ResetData(PluginDataPtr theData);
static long GetFreeSize(PluginInfoPtr theInfo, long *theResult);
static long CheckData(PluginInfoRec *theInfo, boolean setControlData, boolean setPluginData, boolean bigPluginData);
static void SendModelReset(PluginInfoRec *theInfo);
static long AddVertex(long v, ARGBPtr theColor);
static void AddFacet(long v1, long v2, long v3);
static void OutputSurface(void);
static long DoInitialize(long theCommandRecSize, void* theCommandRec);
static long DoFinish(long theCommandRecSize, void* theCommandRec);
static long DoInformation(long theCommandRecSize, void* theCommandRec);
static long DoCheck(long theCommandRecSize, void* theCommandRec);
static long DoDescribe(long theCommandRecSize, void* theCommandRec);
static long DoInterface(long theCommandRecSize, void* theCommandRec);
static long DoSetup(long theCommandRecSize, void* theCommandRec);
static long DoGINF(long theCommandRecSize, void* theCommandRec);
static long DoGATR(long theCommandRecSize, void* theCommandRec);
static long DoGBLR(long theCommandRecSize, void* theCommandRec);
static long DoTMAP(long theCommandRecSize, void* theCommandRec);
static long DoBMAP(long theCommandRecSize, void* theCommandRec);
static long DoRMAP(long theCommandRecSize, void* theCommandRec);
static long DoGenerate(long theCommandRecSize, void* theCommandRec);

static unsigned long globalValidation1, globalValidation2;
static long maxDataSize, topDataSize;
static PluginDataPtr theData;
static ControlDataPtr theControl;
static EIPluginRoutine hostProc;
static char* hostReference;
static HostStatusRec theStatusRec;
static FacetPtr lastFacet;
static long groupCount;

BlobPtr firstBlob;
long blobCount, result;
ulong theVertexFlags;
FunctionRoutine theFunctionRoutine;
ColorFunctionRoutine theColorFunctionRoutine;

boolean StatusTime(void)
{
	return ((theStatusRec.theStatusTick == 0) || (TICKCOUNT() >= theStatusRec.theStatusTick));
}  /* StatusTime */

void UpdateStatus(float thePos, long theValue)
{
	if (result == 0) {
		theStatusRec.theStatusValue = theValue;
		theStatusRec.theStatusPos = thePos;
		result = hostProc(hostStatus, sizeof(HostStatusRec), &theStatusRec);
	}
}  /* UpdateStatus */

void CheckStatus(void)
{
	if ((result == 0) && ((theStatusRec.theStatusTick == 0) || (TICKCOUNT() >= theStatusRec.theStatusTick)))
		result = hostProc(hostStatus, sizeof(HostStatusRec), &theStatusRec);
}  /* CheckStatus */

void SetupStatus(PStrPtr theUnit, PStrPtr theInfo, long min, long max, float thePos, long theValue)
{
	theStatusRec.theReference = hostReference;
	theStatusRec.theStatusPos = thePos;
	PStrCpy(theStatusRec.theStatusUnit, theUnit);
	PStrCpy(theStatusRec.theStatusInfo, theInfo);
	theStatusRec.theStatusMin = min;
	theStatusRec.theStatusMax = max;
	theStatusRec.theStatusValue = theValue;
	theStatusRec.theStatusTick = 0;
	UpdateStatus(thePos, theValue);
}  /* SetupStatus */

char* AllocateMem(long nitems, long nbytes)
{
	long newSize = nitems * FixAlignment(nbytes);
	if (newSize > maxDataSize) {
		result = pluginMemoryErr;
		return NULL;
	} else {
		topDataSize -= newSize;
		maxDataSize -= newSize;
		return (char*)theData + topDataSize;
	}
}  /* AllocateMem */

static char* AllocateBotMem(long nbytes)
{
	long newSize = FixAlignment(nbytes);
	if (newSize > maxDataSize) {
		result = pluginMemoryErr;
		return NULL;
	} else {
		char* thePtr = (char*)theData + theData->theDataSize;
		theData->theDataSize += newSize;
		maxDataSize -= newSize;
		return thePtr;
	}
}  /* AllocateBotMem */

/* Torus: a torus with major, minor radii = 100.0, 20.0, try size = 10.0 */

static double Torus(double x, double y, double z)
{
	double x2 = SQR(x);
	double y2 = SQR(y);
	double z2 = SQR(z);
    double a = x2 + y2 + z2 + (100.0 * 100.0) - (20.0 * 20.0);
	return 4.0 * (100.0 * 100.0) * (y2 + z2) - SQR(a);
}  /* Torus */

static double ColorTorus(CoordPtr position, CoordPtr blur, CoordPtr texture, ARGBPtr theColor)
{
	double x2 = SQR(position->x);
	double y2 = SQR(position->y);
	double z2 = SQR(position->z);
	double a = x2 + y2 + z2 + (100.0 * 100.0) - (20.0 * 20.0);
	
	if (BTST(theVertexFlags, pluginVertexBlurFlag))
		*blur = *position;
	if (BTST(theVertexFlags, pluginVertexTextureFlag))
		*texture = *position;
	if (BTST(theControl->mode, hsvFlag)) {
		theColor->argb.a = 255;
		theColor->argb.r = 255;
		theColor->argb.g = 0;
		theColor->argb.b = 255;
	} else {
		theColor->argb.a = 255;
		theColor->argb.r = 255;
		theColor->argb.g = 255;
		theColor->argb.b = 255;
	}
	return 4.0 * (100.0 * 100.0) * (y2 + z2) - SQR(a);
}  /* ColorTorus */

/* Sphere: an inverse square function (always positive) */

static double Sphere(double radius, double x, double y, double z)
{
    double rsq = x * x + y * y + z * z;
    return radius / (rsq < 0.00001 ? 0.00001 : rsq);
}  /* Sphere */

/* ASphere: a sphere function, try size = 10.0 */

static double ASphere(double x, double y, double z)
{
    return Sphere(10000.0, x, y, z) - 1.0;
}  /* ASphere */

static double ColorASphere(CoordPtr position, CoordPtr blur, CoordPtr texture, ARGBPtr theColor)
{
	double sum = 0.0, hSum = 0.0;
	
	if (BTST(theVertexFlags, pluginVertexBlurFlag))
		*blur = *position;
	if (BTST(theVertexFlags, pluginVertexTextureFlag))
		*texture = *position;
	if (BTST(theControl->mode, hsvFlag)) {
		theColor->argb.a = 255;
		theColor->argb.r = 255;
		theColor->argb.g = 0;
		theColor->argb.b = 255;
	} else {
		theColor->argb.a = 255;
		theColor->argb.r = 255;
		theColor->argb.g = 255;
		theColor->argb.b = 255;
	}
    return Sphere(10000.0, position->x, position->y, position->z) - 1.0;
}  /* ColorASphere */

/* ThreeSpheres: a three-pole blend function, try size = 10.0 */

static double ThreeSpheres(double x, double y, double z)
{
    return Sphere(10000.0, x + 100.0, y, z) + Sphere(10000.0, x, y + 100.0, z) + Sphere(10000.0, x, y, z + 100.0) - 4.0;
}  /* ThreeSpheres */

static double ColorThreeSpheres(CoordPtr position, CoordPtr blur, CoordPtr texture, ARGBPtr theColor)
{
	double sum = 0.0, hSum = 0.0;
	double b1 = Sphere(10000.0, position->x + 100.0, position->y, position->z);
	double b2 = Sphere(10000.0, position->x, position->y + 100.0, position->z);
	double b3 = Sphere(10000.0, position->x, position->y, position->z + 100.0);
	double factor;
	
	sum = b1 + b2 + b3;	
	if (BTST(theVertexFlags, pluginVertexBlurFlag))
		*blur = *position;
	if (BTST(theVertexFlags, pluginVertexTextureFlag)) {
		double factor = 100.0 / sum;
		texture->x = position->x + factor * b1;
		texture->y = position->y + factor * b2;
		texture->z = position->z + factor * b3;
	}
	factor = 255.0 / sum;
	if (BTST(theControl->mode, hsvFlag)) {
		theColor->argb.a = 255;
		theColor->argb.r = (0.333333 * b2 + 0.666666 * b3) * factor + 0.5;
		theColor->argb.g = 255;
		theColor->argb.b = 255;
	} else {
		theColor->argb.a = 255;
		theColor->argb.r = b1 * factor + 0.5;
		theColor->argb.g = b2 * factor + 0.5;
		theColor->argb.b = b3 * factor + 0.5;
	}

    return sum - 4.0;
}  /* ColorThreeSpheres */

static double Blobs(double x, double y, double z)
{
	BlobPtr theBlob = firstBlob;
	double sum = 0.0;
	
	while (theBlob != NULL) {
		double xp = x - theBlob->center.x;
		double yp = y - theBlob->center.y;
		double zp = z - theBlob->center.z;
		double r2 = SQR(xp) + SQR(yp) + SQR(zp);
		if (r2 <= theBlob->n1) {
			double r4 = SQR(r2);
			double r6 = r2 * r4;
			sum += theBlob->n2 * r2 + theBlob->n3 * r4 + theBlob->n4 * r6 + 1.0;
		}
		theBlob = theBlob->next;
	}
	return sum - theControl->blobThreshold;
}  /* Blobs */

static double ColorBlobs(CoordPtr position, CoordPtr blur, CoordPtr texture, ARGBPtr theColor)
{
	BlobPtr theBlob = firstBlob;
	double bXSum = 0.0, bYSum = 0.0, bZSum = 0.0;
	double tXSum = 0.0, tYSum = 0.0, tZSum = 0.0;
	double aSum = 0.0, rSum = 0.0, gSum = 0.0, bSum = 0.0;
	double sum = 0.0, factor;
	
	while (theBlob != NULL) {
		double xp = position->x - theBlob->center.x;
		double yp = position->y - theBlob->center.y;
		double zp = position->z - theBlob->center.z;
		double r2 = SQR(xp) + SQR(yp) + SQR(zp);
		if (r2 <= theBlob->n1) {
			double r4 = SQR(r2);
			double r6 = r2 * r4;
			double blobVal = theBlob->n2 * r2 + theBlob->n3 * r4 + theBlob->n4 * r6 + 1.0;

			sum += blobVal;
			if (BTST(theVertexFlags, pluginVertexBlurFlag)) {
				bXSum += theBlob->blur.x * blobVal;
				bYSum += theBlob->blur.y * blobVal;
				bZSum += theBlob->blur.z * blobVal;
			}
			
			if (BTST(theVertexFlags, pluginVertexTextureFlag)) {
				tXSum += theBlob->center.x * blobVal;
				tYSum += theBlob->center.y * blobVal;
				tZSum += theBlob->center.z * blobVal;
			}
			
			aSum += theBlob->color.argb.a * blobVal;
			rSum += theBlob->color.argb.r * blobVal;
			gSum += theBlob->color.argb.g * blobVal;
			bSum += theBlob->color.argb.b * blobVal;
		}
		theBlob = theBlob->next;
	}
	factor = 1.0 / sum;
	
	if (BTST(theVertexFlags, pluginVertexBlurFlag)) {
		blur->x = bXSum * factor;
		blur->y = bYSum * factor;
		blur->z = bZSum * factor;
	}
	
	if (BTST(theVertexFlags, pluginVertexTextureFlag)) {
		texture->x = position->x - tXSum * factor;
		texture->y = position->y - tYSum * factor;
		texture->z = position->z - tZSum * factor;
	}
	
	theColor->argb.a = (long)(aSum * factor + 0.5);
	theColor->argb.r = (long)(rSum * factor + 0.5);
	theColor->argb.g = (long)(gSum * factor + 0.5);
	theColor->argb.b = (long)(bSum * factor + 0.5);
	return sum - theControl->blobThreshold;
}  /* ColorBlobs */

VertexPtr NewVertex(VertexPtr v)
{
	VertexPtr theVertex;
	
	if (result) return NULL;
	theVertex = (VertexPtr)AllocateBotMem(sizeof(Vertex));
	if (result) return NULL;
	*theVertex = *v;
	return theVertex;
}  /* NewVertex */

static FacetPtr GetNewFacet(void)
{
	FacetPtr theFacet;
	long theOffset;

	if (result)
		return NULL;
	theFacet = (FacetPtr)AllocateBotMem(sizeof(FacetRec));
	if (result)
		return NULL;
	theOffset = (long)theFacet - (long)theData;
	if (theData->facetCount == 0)
		theData->firstFacet = theOffset;
	else
		lastFacet->next = theOffset;
	lastFacet = theFacet;
	theFacet->next = 0;
	theData->facetCount++;
	return theFacet;
}  /* GetNewFacet */

void NewFacet(VertexPtr v1, VertexPtr v2, VertexPtr v3, VertexPtr v4)
{
	FacetPtr theFacet = GetNewFacet();

	if (theFacet) {
		theFacet->v1 = (long)v1 - (long)theData;
		theFacet->v2 = (long)v2 - (long)theData;
		theFacet->v3 = (long)v3 - (long)theData;
		if (v4) {
			theFacet = GetNewFacet();
			if (theFacet) {
				theFacet->v1 = (long)v1 - (long)theData;
				theFacet->v2 = (long)v3 - (long)theData;
				theFacet->v3 = (long)v4 - (long)theData;
			}
		}
	}
}  /* NewFacet */

/* r is hue where 0 is 0 degree and 256 is 360 degrees */
/* g is saturation where 0 is no saturation (white) and 255 is full saturation */
/* b is value where 0 is no value (black) and 255 is full value */
/* Output ranges for r, g and b are 0 to 255 */
static ARGB HSVToARGB(ARGB HSVvalues)
{
	long hue, sat, val, f;
	ARGB RGBvalues;
	
	RGBvalues.argb.a = HSVvalues.argb.a;		/* preserve alpha */
	if (HSVvalues.argb.b == 0) {
		RGBvalues.argb.r = 0;
		RGBvalues.argb.g = 0;
		RGBvalues.argb.b = 0;
	} else if (HSVvalues.argb.g == 0) {
		RGBvalues.argb.r = HSVvalues.argb.b;
		RGBvalues.argb.g = HSVvalues.argb.b;
		RGBvalues.argb.b = HSVvalues.argb.b;
	} else {
		hue = BSR((long)HSVvalues.argb.r * 6144, 8);
		f = BAND(hue, 1023);
		sat = BSL(HSVvalues.argb.g, 6);
		val = BSL(HSVvalues.argb.b, 6);
		switch (BSR(hue, 10)) {
		case 0: 
			RGBvalues.argb.r = HSVvalues.argb.b;
			RGBvalues.argb.g = val * (16320 - BSR(sat * (1024 - f), 10)) / 1044480;
			RGBvalues.argb.b = val * (16320 - sat) / 1044480;
			break;
		case 1: 
			RGBvalues.argb.r = val * (16320 - BSR(sat * f, 10)) / 1044480;
			RGBvalues.argb.g = HSVvalues.argb.b;
			RGBvalues.argb.b = val * (16320 - sat) / 1044480;
			break;
		case 2: 
			RGBvalues.argb.r = val * (16320 - sat) / 1044480;
			RGBvalues.argb.g = HSVvalues.argb.b;
			RGBvalues.argb.b = val * (16320 - BSR(sat * (1024 - f), 10)) / 1044480;
			break;
		case 3: 
			RGBvalues.argb.r = val * (16320 - sat) / 1044480;
			RGBvalues.argb.g = val * (16320 - BSR(sat * f, 10)) / 1044480;
			RGBvalues.argb.b = HSVvalues.argb.b;
			break;
		case 4: 
			RGBvalues.argb.r = val * (16320 - BSR(sat * (1024 - f), 10)) / 1044480;
			RGBvalues.argb.g = val * (16320 - sat) / 1044480;
			RGBvalues.argb.b = HSVvalues.argb.b;
			break;
		default:	/* 5 */
			RGBvalues.argb.r = HSVvalues.argb.b;
			RGBvalues.argb.g = val * (16320 - sat) / 1044480;
			RGBvalues.argb.b = val * (16320 - BSR(sat * f, 10)) / 1044480;
			break;
		}
	}
	return RGBvalues;
}  /* HSVToARGB */

/* r is hue where 0 is 0 degree and 256 is 360 degrees */
/* g is saturation where 0 is no saturation (white) and 255 is full saturation */
/* b is value where 0 is no value (black) and 255 is full value */
/* Input ranges for r, g and b are 0 to 255 */
static ARGB ARGBToHSV(ARGB RGBvalues)
{
	long gVal;
	long min, max;
	ARGB HSVvalues;
	
	HSVvalues.argb.a = RGBvalues.argb.a;		/* Preserve alpha */
	max = RGBvalues.argb.r;
	min = RGBvalues.argb.r;
	if (RGBvalues.argb.g > max)
		max = RGBvalues.argb.g;
	else if (RGBvalues.argb.g < min)
		min = RGBvalues.argb.g;
	if (RGBvalues.argb.b > max)
		max = RGBvalues.argb.b;
	else if (RGBvalues.argb.b < min)
		min = RGBvalues.argb.b;
	
	if (max == 0) {
		HSVvalues.argb.g = 0;
		HSVvalues.argb.r = 0;
		HSVvalues.argb.b = 0;
	} else {
		gVal = (255 * (max - min)) / max;
		if (gVal == 0)
			HSVvalues.argb.r = 0;
		else if (RGBvalues.argb.r == max)
			if (RGBvalues.argb.b < RGBvalues.argb.g)
				HSVvalues.argb.r = (BSL(max - RGBvalues.argb.b, 10) / (max - min) - BSL(max - RGBvalues.argb.g, 10) / (max - min)) * 255 / 6144;
			else
				HSVvalues.argb.r = (6144 + BSL(max - RGBvalues.argb.b, 10) / (max - min) - BSL(max - RGBvalues.argb.g, 10) / (max - min)) * 255 / 6144;
		else if (RGBvalues.argb.g == max)
			HSVvalues.argb.r = (2048 + BSL(max - RGBvalues.argb.r, 10) / (max - min) - BSL(max - RGBvalues.argb.b, 10) / (max - min)) * 255 / 6144;
		else
			HSVvalues.argb.r = (4096 + BSL(max - RGBvalues.argb.g, 10) / (max - min) - BSL(max - RGBvalues.argb.r, 10) / (max - min)) * 255 / 6144;
		HSVvalues.argb.g = gVal;
		HSVvalues.argb.b = max;
	}
	return HSVvalues;
}  /* ARGBToHSV */

static unsigned long GetHostCreator(void)
{
	HostCreatorRec theHostCreatorRec;
	long theResult;
	
	theHostCreatorRec.theReference = hostReference;
	theHostCreatorRec.theCreator = anyCreator;
	theResult = hostProc(hostCreator, sizeof(HostCreatorRec), &theHostCreatorRec);
	return theHostCreatorRec.theCreator;
}  /* GetHostCreator */

static boolean GetGroup(long groupIndex, char* *groupReference, ulong *groupVertexFlags, long *groupVertices, long *groupFacets)
{
	HostModelGetGroupRec theGetGroup;
	long theResult;

	theGetGroup.theReference = hostReference;
	theGetGroup.theGroupIndex = groupIndex;
	theGetGroup.theGroup = NULL;
	theGetGroup.theGroupVertexFlags = 0;
	theGetGroup.theGroupVertices = 0;
	theGetGroup.theGroupFacets = 0;
	theResult = hostProc(hostModelGetGroup, sizeof(HostModelGetGroupRec), &theGetGroup);
	*groupReference = theGetGroup.theGroup;
	*groupVertexFlags = theGetGroup.theGroupVertexFlags;
	*groupVertices = theGetGroup.theGroupVertices;
	*groupFacets = theGetGroup.theGroupFacets;
	return theResult == 0;
}  /* GetGroup */

static void DeleteGroup(char* groupReference)
{
	HostModelDeleteGroupRec theDeleteGroup;

	if (result == 0) {
		theDeleteGroup.theReference = hostReference;
		theDeleteGroup.theGroup = groupReference;
		result = hostProc(hostModelDeleteGroup, sizeof(HostModelDeleteGroupRec), &theDeleteGroup);
	}
}  /* DeleteGroup */

static void GetVertex(char* groupReference, long theVertex, DCoordPtr thePosition, DCoordPtr theBlurPosition, ARGBPtr theColor)
{
	HostModelGetVertexRec theGetVertex;

	if (result == 0) {
		theGetVertex.theReference = hostReference;
		theGetVertex.theGroup = groupReference;
		theGetVertex.theVertexIndex = theVertex;
		result = hostProc(hostModelGetVertex, sizeof(HostModelGetVertexRec), &theGetVertex);
		if (result == 0) {
			*theColor = theGetVertex.theColor;
			*thePosition = theGetVertex.thePosition;
			TransformDCoordinate(thePosition, &theData->theInvMatrix);
			if (BTST(theVertexFlags, pluginVertexBlurFlag)) {
				*theBlurPosition = theGetVertex.theBlurPosition;
				theBlurPosition->x -= theGetVertex.thePosition.x;
				theBlurPosition->y -= theGetVertex.thePosition.y;
				theBlurPosition->z -= theGetVertex.thePosition.z;
			} else {
				theBlurPosition->x = 0.0;
				theBlurPosition->y = 0.0;
				theBlurPosition->z = 0.0;
			}
		}
	}
}  /* GetVertex */

static void GetFacet(char* groupReference, long facetIndex, long *numVertices, Vertices theVertices, ARGBPtr theColor)
{
	short i;
	HostModelGetFacetRec theGetFacet;

	if (result == 0) {
		theGetFacet.theReference = hostReference;
		theGetFacet.theGroup = groupReference;
		theGetFacet.theFacetIndex = facetIndex;
		theGetFacet.theNumVertices = 0;
		result = hostProc(hostModelGetFacet, sizeof(HostModelGetFacetRec), &theGetFacet);
		if ((result != 0) || (theGetFacet.theNumVertices <= 0))
			*numVertices = 0;
		else {
			*theColor = theGetFacet.theColor;
			*numVertices = theGetFacet.theNumVertices;
			for (i = 0; i < *numVertices; i++)
				theVertices[i] = theGetFacet.theVertices[i];
		}
	}
}  /* GetFacet */

/* BuildGroupBlobList will turn each group into a blob of the group's approximate size */

static void BuildGroupsBlobList(float *xc, float *yc, float *zc, float *scale)
{
	char* groupReference;
	long groupIndex, groupVertices, groupFacets, i;
	ulong groupVertexFlags;
	double worldMax = 0.0;
	BlobPtr lastBlob;
	
	lastBlob = NULL;
	groupIndex = 1;
	while ((result == 0) && GetGroup(groupIndex++, &groupReference, &groupVertexFlags, &groupVertices, &groupFacets)) {
		if (groupVertices > 0) {
			DCoordinate vertexPosition;
			DCoordinate vertexBlurPosition;
			double xp, yp, zp;
			double xb, yb, zb;
			double xMin, yMin, zMin;
			double xMax, yMax, zMax;
			ARGB vertexColor;
			
			/* Find the group's extents */
			
			GetVertex(groupReference, 1, &vertexPosition, &vertexBlurPosition, &vertexColor);
			if (BTST(theControl->mode, hsvFlag))
				vertexColor = ARGBToHSV(vertexColor);
			xMin = vertexPosition.x;
			yMin = vertexPosition.y;
			zMin = vertexPosition.z;
			xMax = xMin;
			yMax = yMin;
			zMax = zMin;
			xb = vertexBlurPosition.x;
			yb = vertexBlurPosition.y;
			zb = vertexBlurPosition.z;
			for (i = 2; (i < groupVertices) && (result == 0); i++) {
				ARGB color;
				GetVertex(groupReference, i, &vertexPosition, &vertexBlurPosition, &color);
				if (vertexPosition.x < xMin)
					xMin = vertexPosition.x;
				else if (vertexPosition.x > xMax)
					xMax = vertexPosition.x;
				if (vertexPosition.y < yMin)
					yMin = vertexPosition.y;
				else if (vertexPosition.y > yMax)
					yMax = vertexPosition.y;
				if (vertexPosition.z < zMin)
					zMin = vertexPosition.z;
				else if (vertexPosition.z > zMax)
					zMax = vertexPosition.z;
				xb += vertexBlurPosition.x;
				yb += vertexBlurPosition.y;
				zb += vertexBlurPosition.z;
					
			}
			
			/* Calculate the group's center */
			
			xp = 0.5 * (xMin + xMax);
			yp = 0.5 * (yMin + yMax);
			zp = 0.5 * (zMin + zMax);
			
			/* Calculate the group's motion blur vector */
			
			xb /= groupVertices;
			yb /= groupVertices;
			zb /= groupVertices;
			
			/* Calculate the group's radius */
			
			xMax -= xMin;
			yMax -= yMin;
			zMax -= zMin;
			if (xMax < yMax)
				xMax = yMax;
			if (xMax < zMax)
				xMax = zMax;
				
			/* If the group has a scale, add it to the list */
			
			if (xMax > 0.0) {
				BlobPtr theBlob = (BlobPtr)AllocateMem(1, sizeof(BlobRec));
				if (result == 0) {
					double n2 = SQR(xMax);
					double n4 = SQR(n2);
					double n6 = n2 * n4;
					
					/* Add the blob to the list */

					blobCount++;
					if (lastBlob == NULL)
						firstBlob = theBlob;
					else
						lastBlob->next = theBlob;
					lastBlob = theBlob;
					theBlob->next = NULL;
							
					/* Set the blob color */
					
					theBlob->color = vertexColor;
					
					/* Set the blob center */
					
					theBlob->center.x = xp;
					theBlob->center.y = yp;
					theBlob->center.z = zp;
					
					/* Set the blob motion blur vector */
					
					theBlob->blur.x = xb;
					theBlob->blur.y = yb;
					theBlob->blur.z = zb;
					
					/* Set the blob radius values */
					
					theBlob->radius = xMax;
					theBlob->n1 = n2;
					theBlob->n2 = -2.4444 / n2;
					theBlob->n3 = 1.8888 / n4;
					theBlob->n4 = -0.4444 / n6;
					
					/* Update the world scale */
					
					if (worldMax < xMax)
						worldMax = xMax;
				}
			}
		}
		groupCount++;
	}
		
	/* Calculate the world center and scale */
	
	if ((result == 0) & (blobCount > 0)) {
		*xc = firstBlob->center.x;
		*yc = firstBlob->center.y;
		*zc = firstBlob->center.z;
		*scale = worldMax;
	}
}  /* BuildGroupBlobList */

/* BuildVerticesBlobList will convert each vertex of each attached group into a blob */

static void BuildVerticesBlobList(float *xc, float *yc, float *zc, float *scale)
{
	char* groupReference;
	long groupIndex, groupVertices, groupFacets, i;
	ulong groupVertexFlags;
	BlobPtr lastBlob;
	
	/* Precompute the blob radius values */
	
	double radius = theControl->vertexSize;
	double n1 = SQR(radius);
	double n2 = -2.4444 / n1;
	double n3 = SQR(n1);
	double n4 = -0.4444 / (n1 * n3);
	n3 = 1.8888 / n3;

	lastBlob = NULL;
	groupIndex = 1;
	while ((result == 0) && GetGroup(groupIndex++, &groupReference, &groupVertexFlags, &groupVertices, &groupFacets)) {
		if (groupVertices > 0)
			
			/* Convert each vertex into a blob */
			
			for (i = 1; (i <= groupVertices) && (result == 0); i++) {
				DCoordinate vertexPosition;
				DCoordinate vertexBlurPosition;
				BlobPtr theBlob;
				ARGB vertexColor;
				
				GetVertex(groupReference, i, &vertexPosition, &vertexBlurPosition, &vertexColor);

				if (result == 0)
					theBlob = (BlobPtr)AllocateMem(1, sizeof(BlobRec));
				if (result == 0) {
					
					/* Add the blob to the list */

					blobCount++;
					if (lastBlob == NULL)
						firstBlob = theBlob;
					else
						lastBlob->next = theBlob;
					lastBlob = theBlob;
					theBlob->next = NULL;
								
					/* Set the blob color */
					
					if (BTST(theControl->mode, hsvFlag))
						vertexColor = ARGBToHSV(vertexColor);
					theBlob->color = vertexColor;
					theBlob->blur.x = vertexBlurPosition.x;
					theBlob->blur.y = vertexBlurPosition.y;
					theBlob->blur.z = vertexBlurPosition.z;
					
					/* Set the blob center */
					
					theBlob->center.x = vertexPosition.x;
					theBlob->center.y = vertexPosition.y;
					theBlob->center.z = vertexPosition.z;
					
					/* Set the blob radius values */
					
					theBlob->radius = radius;
					theBlob->n1 = n1;
					theBlob->n2 = n2;
					theBlob->n3 = n3;
					theBlob->n4 = n4;
				}
			}
		groupCount++;
	}
		
	/* Calculate the world center */
	
	if ((result == 0) & (blobCount > 0)) {
		*xc = firstBlob->center.x;
		*yc = firstBlob->center.y;
		*zc = firstBlob->center.z;
		*scale = theControl->vertexSize;
	}
}  /* BuildVerticesBlobList */

/* BuildFacetsBlobList will convert each facet of each attached group into a blob */

static void BuildFacetsBlobList(float *xc, float *yc, float *zc, float *scale)
{
	char* groupReference;
	long groupIndex, groupVertices, groupFacets, i;
	ulong groupVertexFlags;
	double worldMax = 0.0;
	BlobPtr lastBlob;
	
	lastBlob = NULL;
	groupIndex = 1;
	while ((result == 0) && GetGroup(groupIndex++, &groupReference, &groupVertexFlags, &groupVertices, &groupFacets)) {
		if (groupFacets > 0)
			
			/* Convert each vertex into a blob */
			
			for (i = 1; (i <= groupFacets) && (result == 0); i++) {
				long numVertices;
				Vertices theVertices;
				ARGB facetColor;
				
				/* Get the facet */
				
				GetFacet(groupReference, i, &numVertices, theVertices, &facetColor);
				if ((result == 0) && (numVertices > 0)) {
					DCoordinate vertexPosition[4];
					DCoordinate vertexBlurPosition;
					double xr = 0.0, yr = 0.0, zr = 0.0;
					double xb = 0.0, yb = 0.0, zb = 0.0;
					ARGB vertexColor;
					long j;
					
					/* Get the facet's vertices and sum their values */
					
					for (j = 0; (j < numVertices) && (result == 0); j++) {
						GetVertex(groupReference, theVertices[j], &vertexPosition[j], &vertexBlurPosition, &vertexColor);
						xr += vertexPosition[j].x;
						yr += vertexPosition[j].y;
						zr += vertexPosition[j].z;
						xb += vertexBlurPosition.x;
						yb += vertexBlurPosition.y;
						zb += vertexBlurPosition.z;
					}
					if (result == 0) {
						double radius = 0.0;
						double factor = 1.0 / numVertices;
					
						/* Locate the facet's center */
					
						xr *= factor;
						yr *= factor;
						zr *= factor;
						
						/* Calculate the facet's motion blur vector */
						
						xb *= factor;
						yb *= factor;
						zb *= factor;
						
						/* Find the largest distance from a facet vertex to the facet's center */

						if (numVertices >= 3) {						
							for (j = 0; j < numVertices; j++) {
								double x = xr - vertexPosition[j].x;
								double y = yr - vertexPosition[j].y;
								double z = zr - vertexPosition[j].z;
								double dist = SQR(x) + SQR(y) + SQR(z);
								if (radius < dist)
									radius = dist;
							}
							radius = 2.0 * theControl->facetScale * SQRT(radius);
						} else
							radius = theControl->vertexSize;
						
						/* Create a blob with the facet's radius */
						
						if (radius > 0.0) {
							BlobPtr theBlob = (BlobPtr)AllocateMem(1, sizeof(BlobRec));
							if (result == 0) {
								double n2 = SQR(radius);
								double n4 = SQR(n2);
								double n6 = n2 * n4;
								
								/* Add the blob to the list */
			
								blobCount++;
								if (lastBlob == NULL)
									firstBlob = theBlob;
								else
									lastBlob->next = theBlob;
								lastBlob = theBlob;
								theBlob->next = NULL;
								
								/* Set the blob color */
								
								if (BTST(theControl->mode, hsvFlag))
									facetColor = ARGBToHSV(facetColor);
								theBlob->color = facetColor;
								
								/* Set the blob center */
								
								theBlob->center.x = xr;
								theBlob->center.y = yr;
								theBlob->center.z = zr;
								
								/* Set the blob motion blur vector */
								
								theBlob->blur.x = xb;
								theBlob->blur.y = yb;
								theBlob->blur.z = zb;
								
								/* Set the blob radius values */
								
								theBlob->radius = radius;
								theBlob->n1 = n2;
								theBlob->n2 = -2.4444 / n2;
								theBlob->n3 = 1.8888 / n4;
								theBlob->n4 = -0.4444 / n6;
								
								/* Update the world scale */
								
								if (worldMax < radius)
									worldMax = radius;
							}
						}
					}
				}
			}
		groupCount++;
	}
		
	/* Calculate the world center */
	
	if ((result == 0) & (blobCount > 0)) {
		*xc = firstBlob->center.x;
		*yc = firstBlob->center.y;
		*zc = firstBlob->center.z;
		*scale = worldMax;
	}
}  /* BuildFacetsBlobList */

static void InitGlobals(boolean initFlag)
{
	if (initFlag || (globalValidation1 != pluginType) || (globalValidation2 != pluginModelType)) {
		globalValidation1 = pluginType;
		globalValidation2 = pluginModelType;
		InitUtils();
	}
}  /* InitGlobals */

static void FixControl(ControlDataPtr theControlRec)
{
	theControlRec->ownerType = pluginType;
	theControlRec->ownerVersion = pluginVersion;

	FixField(theControlRec->surfaceType);
	FixField(theControlRec->mode);
	FixField(theControlRec->resolutionEI);
	FixField(theControlRec->resolutionCamera);
	FixField(theControlRec->vertexSize);
	FixField(theControlRec->facetScale);
	FixField(theControlRec->blobThreshold);
}  /* FixControl */

static void DefaultControl(ControlDataPtr theControl)
{
	theControl->ownerType = pluginType;
	theControl->ownerVersion = pluginVersion;
	theControl->surfaceType = defSurfaceType;
	theControl->mode = defMode;
	theControl->resolutionEI = defResolutionEI;
	theControl->resolutionCamera = defResolutionCamera;
	theControl->vertexSize = defVertexSize;
	theControl->facetScale = defFacetScale;
	theControl->blobThreshold = defBlobThreshold;
}  /* DefaultControl */

static void DefaultData(PluginDataPtr theData)
{
	theData->theDataSize = sizeof(PluginDataRec);
	InitMatrix4(&theData->theMatrix);
	InitMatrix4(&theData->theNormalMatrix);
	InitMatrix4(&theData->theBlurMatrix);
	InitMatrix4(&theData->theInvMatrix);
	theData->facetCount = 0;
	theData->firstFacet = 0;
}  /* DefaultData */

static void ResetData(PluginDataPtr theData)
{
	theData->theDataSize = sizeof(PluginDataRec);
	theData->facetCount = 0;
	theData->firstFacet = 0;
}  /* ResetData */

static long GetFreeSize(PluginInfoPtr theInfo, long *theResult)
{
	HostFreeRec theFree;

	theFree.theReference = theInfo->theHostReference;
	theFree.theFreeSize = 0;
	*theResult = theInfo->theHostProc(hostFree, sizeof(HostFreeRec), &theFree);
	if (*theResult == 0)
		return theFree.theFreeSize;
	else
		return 0;
}  /* GetFreeSize */ 

static long CheckData(PluginInfoPtr theInfo, boolean setControlData, boolean setPluginData, boolean bigPluginData)
{
	HostAllocateRec theAllocate;
	long newPluginDataSize, newControlDataSize;
	boolean newPluginData, newControlData;
	ControlDataRec theControlRec;
	PluginDataRec theDataRec;
	ControlDataPtr theControl;
	long theResult;

	theResult = 0;

	/* Change the size of the plugin data block or control data block if the size is too small.*/
	/* Reverse byte ordering is corrected. */

	theControl = (ControlDataPtr)theInfo->theControlData;
	if (!setControlData && (theInfo->theControlDataSize >= sizeof(ControlDataRec))
		&& ((theControl->ownerType == pluginType) || (theControl->ownerType == REVERSELONG(pluginType)))
		&& ((theControl->ownerVersion == pluginVersion) || (theControl->ownerVersion == REVERSELONG(pluginVersion)))) {
		if (theControl->ownerVersion == REVERSELONG(pluginVersion))
			FixControl(theControl);
		theControlRec = *theControl;
	} else {
		setControlData = 1;
		DefaultControl(&theControlRec);
	}

	if (setPluginData || (theInfo->thePluginDataSize < sizeof(PluginDataRec))) {
		setPluginData = 1;
		DefaultData(&theDataRec);
	} else
		theDataRec = *(PluginDataPtr)theInfo->thePluginData;

	newControlDataSize = sizeof(ControlDataRec);
	newControlData = (theInfo->theControlDataSize < newControlDataSize || theInfo->thePluginDataSize > FixAlignment(newControlDataSize));
	if (bigPluginData) {
		newPluginDataSize = GetFreeSize(theInfo, &theResult);
		if (newPluginDataSize < theDataRec.theDataSize)
			newPluginDataSize = theDataRec.theDataSize;
		if (newPluginDataSize < sizeof(PluginDataRec))
			newPluginDataSize = sizeof(PluginDataRec);
	} else
		newPluginDataSize = theDataRec.theDataSize;
	newPluginData = (theInfo->thePluginDataSize < newPluginDataSize || theInfo->thePluginDataSize > FixAlignment(newPluginDataSize));

	if (newPluginData || newControlData) {
		theAllocate.theReference = theInfo->theHostReference;
		theAllocate.theControlDataSize = newControlDataSize;
		theAllocate.theControlData = theInfo->theControlData;
		theAllocate.thePluginDataSize = newPluginDataSize;
		theAllocate.thePluginData = theInfo->thePluginData;
		theResult = theInfo->theHostProc(hostAllocate, sizeof(HostAllocateRec), &theAllocate);
		theInfo->theControlData = theAllocate.theControlData;
		theInfo->theControlDataSize = theAllocate.theControlDataSize;
		theInfo->thePluginData = theAllocate.thePluginData;
		theInfo->thePluginDataSize = theAllocate.thePluginDataSize;
	}

	if (theResult == 0) {
	
		/* Setup the control data to the default values if an incorrect sized block was passed to the plugin */
	
		if (newControlData || setControlData)
			*(ControlDataPtr)theInfo->theControlData = theControlRec;
	
		/* Setup the plugin private data */
	
		if (newPluginData || setPluginData)
			*(PluginDataPtr)theInfo->thePluginData = theDataRec;
	}
	return theResult;
}  /* CheckData */

/* Reset the host's model data */

static void SendModelReset(PluginInfoRec *theInfo)
{
	HostModelResetRec theModelReset;
	long theResult;

	theModelReset.theReference = theInfo->theHostReference;
	theResult = theInfo->theHostProc(hostModelReset, sizeof(HostModelResetRec), &theModelReset);
}  /* SendModelReset */

static long AddVertex(long v, ARGBPtr theColor)
{
	HostModelAddVertexRec theAddVertex;
	VertexPtr theVertex;
	
	if ((result == 0) && (v > 0)) {
		theVertex = (VertexPtr)((char*)theData + v);
		*theColor = theVertex->color;
		if (theVertex->index == 0x80000000) {
			theAddVertex.theReference = hostReference;

			if (BTST(theControl->mode, hsvFlag))
				theAddVertex.theColor = HSVToARGB(theVertex->color);
			else
				theAddVertex.theColor = theVertex->color;

			if (BTST(theVertexFlags, pluginVertexNormalFlag)) {
				theAddVertex.theNormal.x = theVertex->normal.x;
				theAddVertex.theNormal.y = theVertex->normal.y;
				theAddVertex.theNormal.z = theVertex->normal.z;
				TransformNormal(&theAddVertex.theNormal.x, &theAddVertex.theNormal.y, &theAddVertex.theNormal.z, &theData->theNormalMatrix);
			} else {
				theAddVertex.theNormal.x = 0.0;
				theAddVertex.theNormal.y = 0.0;
				theAddVertex.theNormal.z = 0.0;
			}

			theAddVertex.thePosition.x = theVertex->position.x;
			theAddVertex.thePosition.y = theVertex->position.y;
			theAddVertex.thePosition.z = theVertex->position.z;
			TransformDCoordinate(&theAddVertex.thePosition, &theData->theMatrix);

			if (BTST(theVertexFlags, pluginVertexBlurFlag)) 
				if (theControl->surfaceType >= groupsSurface) {
					theAddVertex.theBlurPosition.x = theVertex->blur.x + theAddVertex.thePosition.x;
					theAddVertex.theBlurPosition.y = theVertex->blur.y + theAddVertex.thePosition.y;
					theAddVertex.theBlurPosition.z = theVertex->blur.z + theAddVertex.thePosition.z;
				} else {
					theAddVertex.theBlurPosition.x = theVertex->blur.x;
					theAddVertex.theBlurPosition.y = theVertex->blur.y;
					theAddVertex.theBlurPosition.z = theVertex->blur.z;
					TransformDCoordinate(&theAddVertex.theBlurPosition, &theData->theBlurMatrix);
			} else
				theAddVertex.theBlurPosition = theAddVertex.thePosition;

			if (BTST(theVertexFlags, pluginVertexTextureFlag)) {
				theAddVertex.theTexturePosition.x = theVertex->texture.x;
				theAddVertex.theTexturePosition.y = theVertex->texture.y;
				theAddVertex.theTexturePosition.z = theVertex->texture.z;
			} else {
				theAddVertex.theTexturePosition.x = theVertex->position.x;
				theAddVertex.theTexturePosition.y = theVertex->position.y;
				theAddVertex.theTexturePosition.z = theVertex->position.z;
			}
			result = hostProc(hostModelAddVertex, sizeof(HostModelAddVertexRec), &theAddVertex);
			if (result == 0) {
				theVertex->index = theAddVertex.theIndex;
				return theVertex->index;
			} else
				return 0;
		} else
			return theVertex->index;
	} else
		return 0;
}  /* AddVertex */

static void AddFacet(long v1, long v2, long v3)
{
	if (result == 0) {
		HostModelAddFacetRec theAddFacet;
		ARGB c1, c2, c3;

		theAddFacet.theReference = hostReference;
		theAddFacet.theNumVertices = 3;
		theAddFacet.theVertices[0] = AddVertex(v1, &c1);
		theAddFacet.theVertices[1] = AddVertex(v2, &c2);
		theAddFacet.theVertices[2] = AddVertex(v3, &c3);
		theAddFacet.theVertices[3] = 0;
		if (result == 0) {
			ARGB color;
		
			/* Calculate the facet color as the average of its vertex colors */
			
			if ((c1.colorValue == c2.colorValue) && (c1.colorValue == c3.colorValue))
				color = c1;
			else {
				long aSum = (long)c1.argb.a + (long)c2.argb.a + (long)c3.argb.a;
				long rSum = (long)c1.argb.r + (long)c2.argb.r + (long)c3.argb.r;
				long gSum = (long)c1.argb.g + (long)c2.argb.g + (long)c3.argb.g;
				long bSum = (long)c1.argb.b + (long)c2.argb.b + (long)c3.argb.b;

				color.argb.a = (aSum * 5462) >> 14;
				color.argb.r = (rSum * 5462) >> 14;
				color.argb.g = (gSum * 5462) >> 14;
				color.argb.b = (bSum * 5462) >> 14;
			}
			
			/* Convert the color to RGB */
			
			if (BTST(theControl->mode, hsvFlag))
				theAddFacet.theColor = HSVToARGB(color);
			else
				theAddFacet.theColor = color;
				
			/* Add the facet to the host's facet list */
			
			result = hostProc(hostModelAddFacet, sizeof(HostModelAddFacetRec), &theAddFacet);
		}
	}
}  /* AddFacet */

static void OutputSurface(void)
{
	long i;
	FacetPtr theFacet;

	SetupStatus(PS("Facet"), PS("Creating Facets"), 0, theData->facetCount, 0.0, 0);
		
	/* Output the surface facets */

	theFacet = (FacetPtr)((char*)theData + theData->firstFacet);
	i = 0;
	while ((i < theData->facetCount) && (result == 0)) {
		AddFacet(theFacet->v1, theFacet->v2, theFacet->v3);
		theFacet = (FacetPtr)((char*)theData + theFacet->next);
		i++;
		if (StatusTime())
			UpdateStatus((double)i / theData->facetCount, i);
	}
	UpdateStatus(1.0, theData->facetCount);
}  /* OutputSurface */

/* Initialize the plugin's data structures and perform memory allocation */

static long DoInitialize(long theCommandRecSize, void* theCommandRec)
{
	PluginInitializePtr theParameters;
	long theResult;

	InitGlobals(1);
	theParameters = (PluginInitializePtr)theCommandRec;
	if ((theCommandRecSize < sizeof(PluginInitializeRec)) || (theParameters->theInfo.theHostProc == NULL))
		theResult = pluginParameterErr;
	else
		theResult = CheckData(&theParameters->theInfo, 0, 1, 0);
	if (theResult == 0)
		SendModelReset(&theParameters->theInfo);
	return theResult;
}  /* DoInitialize */

/* Finish up anything which needs it.  This is the last time that the plugin will be called */

static long DoFinish(long theCommandRecSize, void* theCommandRec)
{
	PluginFinishPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginFinishPtr)theCommandRec;
	if ((theCommandRecSize < sizeof(PluginFinishRec)) || (theParameters->theInfo.theHostProc == NULL))
		theResult = pluginParameterErr;
	FinishUtils();
	return theResult;
}  /* DoFinish */

/* Get information about the plugin */

static long DoInformation(long theCommandRecSize, void* theCommandRec)
{
	PluginInformationPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginInformationPtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginInformationRec))
		theResult = pluginParameterErr;
	else {
		theParameters->thePluginType = pluginType;
		theParameters->thePluginVersion = pluginVersion;
		PStrCpy(theParameters->thePluginName, pluginName);
		PStrCpy(theParameters->thePluginCopyright, pluginCopyright);
		PStrCpy(theParameters->thePluginAuthor, pluginAuthor);
		PStrCpy(theParameters->thePluginDescription, pluginDescription);
		theParameters->thePluginPermission = pluginPermission;
	}
	return theResult;
}  /* DoInformation */

/* Check to see if a particular command is supported */

static long DoCheck(long theCommandRecSize, void* theCommandRec)
{
	PluginCheckPtr theParameters;
	
	theParameters = (PluginCheckPtr)theCommandRec;
	if (theCommandRecSize >= sizeof(PluginCheckRec))
		switch (theParameters->theCommand) {
			case pluginInitialize:
			case pluginFinish:
			case pluginInformation:
			case pluginCheck:
			case pluginDescribe:
			case pluginInterface:
			case pluginModelSetup:
			case pluginModelGINF:
			case pluginModelGATR:
			case pluginModelGBLR:
			case pluginModelTMAP:
			case pluginModelBMAP:
			case pluginModelRMAP:
			case pluginModelGenerate:
				return pluginNoErr;
		}
	return pluginParameterErr;
}  /* DoCheck */

/* Describe an error number */

static long DoDescribe(long theCommandRecSize, void* theCommandRec)
{
	PluginDescribePtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginDescribePtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginDescribeRec))
		theResult = pluginParameterErr;
	else
		DescribePluginError(theParameters->theResult, theParameters->theDescription);
	return theResult;
}  /* DoDescribe */

/* Display the plugin's user interface and edit parameters */

static long DoInterface(long theCommandRecSize, void* theCommandRec)
{
	EI_ResourceFile			*resourceFile;
	EI_Dialog				*dialog;
	ControlDataPtr			theControl;
	PluginDataPtr			theData;
	PluginInterfacePtr		theParameters;
	long					theResult;
	
	theResult = pluginNoErr;
	theParameters = (PluginInterfacePtr)theCommandRec;
	if ((theCommandRecSize < sizeof(PluginInterfaceRec)) || (theParameters->theInfo.theControlData == NULL) || (theParameters->theInfo.theControlDataSize < sizeof(ControlDataRec)))
		theResult = pluginParameterErr;
	else
		theResult = CheckData(&theParameters->theInfo, 0, 0, 0);
	if (theResult == 0)
		resourceFile = OpenResourceFile("MrBlobby.rsc");
	if (resourceFile == NULL)
		theResult = pluginResourceErr;
	if (theResult == 0) {
		theControl = (ControlDataPtr)theParameters->theInfo.theControlData;
		theData = (PluginDataPtr)theParameters->theInfo.thePluginData;
		dialog = EI_MakeDialog(resourceFile, interfaceDialogID);
		if (dialog != NULL) {
			SetButton(dialog, cMeshWithRB, BTST(theControl->mode, tetrahedronsFlag));
			switch (theControl->surfaceType) {
				case sphereSurface:
					SetButton(dialog, cGenerateRB, 0);
					break;
				case torusSurface:
					SetButton(dialog, cGenerateRB, 1);
					break;
				case spheresSurface:
					SetButton(dialog, cGenerateRB, 2);
					break;
				case groupsSurface:
					SetButton(dialog, cGenerateRB, 3);
					break;
				case verticesSurface:
					SetButton(dialog, cGenerateRB, 4);
					break;
				case facetsSurface:
					SetButton(dialog, cGenerateRB, 5);
					break;
			}
			SetButton(dialog, cInterpolatedColorRB, BTST(theControl->mode, hsvFlag));
			SetButton(dialog, cTextureMapRB, BTST(theControl->mode, textureModelFlag));
			SetTheIReal(dialog, cEIMeshResolutionET, theControl->resolutionEI, 4);
			SetTheIReal(dialog, cCameraMeshResolution, theControl->resolutionCamera, 4);
			SetTheIReal(dialog, cVertexBlobSizeET, theControl->vertexSize, 4);
			SetTheIReal(dialog, cFacetBlobScaleET, theControl->facetScale, 4);
			SetTheIReal(dialog, cBlobThresholdET, theControl->blobThreshold, 4);
			
			if (EI_ExecuteDialog(dialog)) {
				long temp;
				float rTemp;
				boolean changeFlag = 0;
				
				SetBooleanFlag(&theControl->mode, tetrahedronsFlag, GetButton(dialog, cMeshWithRB), &changeFlag);
				switch (GetButton(dialog, cGenerateRB)) {
					case 0:
						SetLongInt(&theControl->surfaceType, sphereSurface, &changeFlag);
						break;
					case 1:
						SetLongInt(&theControl->surfaceType, torusSurface, &changeFlag);
						break;
					case 2:
						SetLongInt(&theControl->surfaceType, spheresSurface, &changeFlag);
						break;
					case 3:
						SetLongInt(&theControl->surfaceType, groupsSurface, &changeFlag);
						break;
					case 4:
						SetLongInt(&theControl->surfaceType, verticesSurface, &changeFlag);
						break;
					case 5:
						SetLongInt(&theControl->surfaceType, facetsSurface, &changeFlag);
						break;
				}
				SetBooleanFlag(&theControl->mode, hsvFlag, GetButton(dialog, cInterpolatedColorRB), &changeFlag);
				SetBooleanFlag(&theControl->mode, textureModelFlag, GetButton(dialog, cTextureMapRB), &changeFlag);
				
				temp = GetTheIReal(dialog, cEIMeshResolutionET) + 0.5;
				if (temp >= 1)
					SetLongInt(&theControl->resolutionEI, temp, &changeFlag);
				
				temp = GetTheIReal(dialog, cCameraMeshResolution) + 0.5;
				if (temp >= 1)
					theControl->resolutionCamera = temp;
				
				rTemp = GetTheIReal(dialog, cVertexBlobSizeET);
				if (rTemp > 0.0)
					SetReal(&theControl->vertexSize, rTemp, &changeFlag);
				
				rTemp = GetTheIReal(dialog, cFacetBlobScaleET);
				if (rTemp > 0.0)
					SetReal(&theControl->facetScale, rTemp, &changeFlag);
				
				SetReal(&theControl->blobThreshold, GetTheIReal(dialog, cBlobThresholdET), &changeFlag);

				if (changeFlag) {
					ResetData(theData);
					SendModelReset(&theParameters->theInfo);
				}
			} else
				theResult = pluginCancelErr;
			EI_DestroyDialog(dialog);
		}
		EI_CloseResourceFile(resourceFile);
	}
	return theResult;
}  /* DoInterface */

/* Setup the model generation variables */

static long DoSetup(long theCommandRecSize, void* theCommandRec)
{
	PluginModelSetupPtr theParameters;
	PluginInfoPtr theInfo;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelSetupPtr)theCommandRec;
	theInfo = &theParameters->theInfo;
	if ((theCommandRecSize < sizeof(PluginModelSetupRec)) || (theInfo->theHostProc == NULL))
		theResult = pluginParameterErr;
	else
		theResult = CheckData(theInfo, 0, 0, 0);
	if (theResult == 0) {
		char* groupReference;
		long groupVertices, groupFacets;

		theControl = (ControlDataPtr)theInfo->theControlData;
		theData = (PluginDataPtr)theInfo->thePluginData;
		theData->theMatrix = theParameters->theGroupMatrix;
		theData->theNormalMatrix = theParameters->theGroupMatrix;
		NormalMatrix4(&theData->theNormalMatrix);
		theData->theBlurMatrix = theParameters->theGroupBlurMatrix;
		InvertMatrix4(&theData->theInvMatrix, theData->theMatrix);

		if ((theData->theDataSize <= sizeof(PluginDataRec)) || (theControl->surfaceType >= groupsSurface)) {
			float xc = 0.0, yc = 0.0, zc = 0.0, scale = 100.0, cubeSize;

			result = CheckData(theInfo, 0, 0, 1);
			hostProc = theInfo->theHostProc;
			hostReference = theInfo->theHostReference;
			theControl = (ControlDataPtr)theInfo->theControlData;
			theData = (PluginDataPtr)theInfo->thePluginData;
			ResetData(theData);
			maxDataSize = theInfo->thePluginDataSize;
			topDataSize = maxDataSize;
			groupCount = 0;
			blobCount = 0;
			firstBlob = NULL;
			if (!GetGroup(0, &groupReference, &theVertexFlags, &groupVertices, &groupFacets))
				theVertexFlags = 0;
			if (BTST(theControl->mode, textureModelFlag))
				BCLR(theVertexFlags, pluginVertexTextureFlag);
	
			/* Setup the implicit surface functions and blob lists */
			
			theFunctionRoutine = ASphere;
			theColorFunctionRoutine = ColorASphere;
			if (theControl->surfaceType == spheresSurface) {
				theFunctionRoutine = ThreeSpheres;
				theColorFunctionRoutine = ColorThreeSpheres;
			} else if (theControl->surfaceType == torusSurface) {
				theFunctionRoutine = Torus;
				theColorFunctionRoutine = ColorTorus;
			} else if (theControl->surfaceType == groupsSurface) {
				BuildGroupsBlobList(&xc, &yc, &zc, &scale);
				if (groupCount > 0) {
					theFunctionRoutine = Blobs;
					theColorFunctionRoutine = ColorBlobs;
				}
			} else if (theControl->surfaceType == verticesSurface) {
				BuildVerticesBlobList(&xc, &yc, &zc, &scale);
				if (groupCount > 0) {
					theFunctionRoutine = Blobs;
					theColorFunctionRoutine = ColorBlobs;
				}
			} else if (theControl->surfaceType == facetsSurface) {
				BuildFacetsBlobList(&xc, &yc, &zc, &scale);
				if (groupCount > 0) {
					theFunctionRoutine = Blobs;
					theColorFunctionRoutine = ColorBlobs;
				}
			}
			
			/* Calculate the cubic mesh size based upon whether or not the host app is Camera */
			
			if (GetHostCreator() == QUADBYTE('E', 'I', 'A', 'R'))	/* Camera's creator type */
				cubeSize = scale / theControl->resolutionCamera;
			else
				cubeSize = scale / theControl->resolutionEI;
			
			/* Generate the implicit surface polygon mesh */
			
			Polygonize(cubeSize, xc, yc, zc, BTST(theControl->mode, tetrahedronsFlag));
			
			/* Release the unused memory */
			
			if (result == 0)
				result = CheckData(theInfo, 0, 0, 0);
				
			/* Tell the host to send this plug-in a pluginModelGenerate command */
			
			theResult = result;
		}
		if (theResult == 0)
			SendModelReset(theInfo);
	}
	return theResult;
}  /* DoSetup */

/* Setup the group GINF record */

static long DoGINF(long theCommandRecSize, void* theCommandRec)
{
	PluginModelGINFPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelGINFPtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginModelGINFRec))
		theResult = pluginParameterErr;
	else
		theParameters->theChangeFlag = 0;
	return theResult;
}  /* DoGINF */

/* Setup the group GATR record */

static long DoGATR(long theCommandRecSize, void* theCommandRec)
{
	PluginModelGATRPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelGATRPtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginModelGATRRec))
		theResult = pluginParameterErr;
	else
		theParameters->theChangeFlag = 0;
	return theResult;
}  /* DoGATR */

/* Setup the group GBLR record */

static long DoGBLR(long theCommandRecSize, void* theCommandRec)
{
	PluginModelGBLRPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelGBLRPtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginModelGBLRRec))
		theResult = pluginParameterErr;
	else
		theParameters->theChangeFlag = 0;
	return theResult;
}  /* DoGBLR */

/* Setup the group TMAP record */

static long DoTMAP(long theCommandRecSize, void* theCommandRec)
{
	PluginModelTMAPPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelTMAPPtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginModelTMAPRec))
		theResult = pluginParameterErr;
	else
		theParameters->theChangeFlag = 0;
	return theResult;
}  /* DoTMAP */

/* Setup the group BMAP record */

static long DoBMAP(long theCommandRecSize, void* theCommandRec)
{
	PluginModelBMAPPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelBMAPPtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginModelBMAPRec))
		theResult = pluginParameterErr;
	else
		theParameters->theChangeFlag = 0;
	return theResult;
}  /* DoBMAP */

/* Setup the group RMAP record */

static long DoRMAP(long theCommandRecSize, void* theCommandRec)
{
	PluginModelRMAPPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelRMAPPtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginModelRMAPRec))
		theResult = pluginParameterErr;
	else
		theParameters->theChangeFlag = 0;
	return theResult;
}  /* DoRMAP */

/* Generate the coordinates and polygons for the group */

static long DoGenerate(long theCommandRecSize, void* theCommandRec)
{
	PluginModelGeneratePtr theParameters;
	PluginInfoPtr theInfo;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelGeneratePtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginModelGenerateRec))
		theResult = pluginParameterErr;
	else {
		theInfo = &theParameters->theInfo;
		theResult = CheckData(&theParameters->theInfo, 0, 0, 0);
		if (theResult == 0) {
			result = theResult;

			theControl = (ControlDataRec *)theInfo->theControlData;
			theData = (PluginDataRec *)theInfo->thePluginData;
			hostProc = theInfo->theHostProc;
			hostReference = theInfo->theHostReference;
		
			/* Write out the mesh vertices and facets */
		
			OutputSurface();
			
			/* Force the surface data to regenerate in the next setup command */

			ResetData(theData);
			
			/* Delete all of the child groups */
			
			if ((result == 0) && ((theControl->surfaceType == groupsSurface) || (theControl->surfaceType == verticesSurface) || (theControl->surfaceType == facetsSurface))) {
				long i, groupVertices, groupFacets;
				ulong groupVertexFlags;
				char* groupReference;
				
				for (i = 1; (i <= theParameters->theChildGroups) && (result == 0); i++)
					if (GetGroup(i, &groupReference, &groupVertexFlags, &groupVertices, &groupFacets))
						DeleteGroup(groupReference);
			}
			theResult = result;
		}
	}
	return theResult;
}  /* DoGenerate */

/* The main function is called by the host application for all plugin commands */

EXPDLL long Main(unsigned long theCommand, long theCommandRecSize, void* theCommandRec)
{
	long theResult;
	
	/* Initialize the plugin's global variables if they have not been initialize previously */

	InitGlobals(0);

	/* Perform the requested command */

	switch (theCommand) {
		case pluginInitialize:		theResult = DoInitialize(theCommandRecSize, theCommandRec);		break;
		case pluginFinish:			theResult = DoFinish(theCommandRecSize, theCommandRec);			break;
		case pluginInformation:		theResult = DoInformation(theCommandRecSize, theCommandRec);	break;
		case pluginCheck:			theResult = DoCheck(theCommandRecSize, theCommandRec);			break;
		case pluginDescribe:		theResult = DoDescribe(theCommandRecSize, theCommandRec);		break;
		case pluginInterface:		theResult = DoInterface(theCommandRecSize, theCommandRec);		break;
		case pluginModelSetup:		theResult = DoSetup(theCommandRecSize, theCommandRec);			break;
		case pluginModelGINF:		theResult = DoGINF(theCommandRecSize, theCommandRec);			break;
		case pluginModelGATR:		theResult = DoGATR(theCommandRecSize, theCommandRec);			break;
		case pluginModelGBLR:		theResult = DoGBLR(theCommandRecSize, theCommandRec);			break;
		case pluginModelTMAP:		theResult = DoTMAP(theCommandRecSize, theCommandRec);			break;
		case pluginModelBMAP:		theResult = DoBMAP(theCommandRecSize, theCommandRec);			break;
		case pluginModelRMAP:		theResult = DoRMAP(theCommandRecSize, theCommandRec);			break;
		case pluginModelGenerate:	theResult = DoGenerate(theCommandRecSize, theCommandRec);		break;
		default:					theResult = pluginCommandErr;									break;
	}

	/* Restore the host's A4 register */

	return theResult;
}  /* Main */
