/*
Copyright 1995-1998 Electric Image, Inc. This source code is intended for teaching
purposes only. It may not be included in any products (freeware, shareware, or
commercial) without the written permission of Electric Image, Inc. This source
code may be copied and modified but this copyright notice must be included in
its original form. This source code may not be offered for sale or distributed
by any entity without the written permission of Electric Image, Inc. Copies of
the original or modified code may be posted to the ElectricImage Forum on
America-on-line or eWorld to foster a better understanding of the Plug-In
Specification.
*/

/* This code presents an example of a surface patch generator model plugin. */

#include "EI_Control.h"
#include "EI_Dialog.h"
#include "EI_ResourceFile.h"
#include "MemoryUtils.h"
#include "PluginClient.h"
#include "PluginFPU.h"
#include "PluginUtils.h"
#include "PluginMatrix.h"
#include "PString.h"

#define pluginType			QUADBYTE('T', 'E', 'A', 'P')	/* The owner plugin's unique type */
#define pluginVersion		0x00000001			/* The owner plugin's version number 0:0:version:revision */
#define pluginName			PS("ElectricImage Teapot")
#define pluginCopyright		PS("Copyright  1995 ElectricImage, Inc.")
#define pluginAuthor		PS("Mark Granger|Electric Image, Inc.|117 East Colorado Boulevard, Suite 300|Pasadena, Ca 91105|(818) 577-1627")
#define pluginDescription	PS("This model plugin will generate the famous Newell teapot.  It has selectable patch resolution and color.")
#define pluginPermission	BIT(pluginSensitivityFlag)

#define normalDelta			0.0001

#define defTeapotFlags		0x00000000	/* Texture by facet, generate a teapot */
#define defTeapotPoints		20
#define defTeapotScale		100.0

#define texturePatchFlag	31
#define generatePatchFlag	30

#define aboutDialogID   	16000
#define interfaceDialogID	16001

enum {
	cPatchPointsET = 5,
	cPatchScaleET = 6,
	cPatchColor = 8,
	cTextureByPatchCheck = 9,
	cGenerateRB = 11
};
/*
#define okItem          	1
#define cancelItem     		2
#define patchPointsItem   	3
#define patchScaleItem    	4
#define colorItem			5
#define texturePatchItem	6
#define generateTeapotItem	7
#define generatePatchItem	8
*/

typedef struct PluginDataRec {
	Matrix4 theMatrix;
	Matrix4 theRotMatrix;
	Matrix4 theBlurMatrix;
	long vertexFlags;
} PluginDataRec, *PluginDataPtr;

typedef struct ControlDataRec {
	ulong ownerType;		/* The owner plugin's unique type */
	long ownerVersion;	/* The owner plugin's version number */
	ulong teapotFlags;	/* The control flags for the teapot */
	double teapotScale;		/* The overall scale of the teapot */
	ARGB teapotColor;		/* The default surface color */
	long teapotPoints;	/* The patch density of the teapot */
	long filler;			/* Filler bytes for 64 bit alignment */
} ControlDataRec, *ControlDataPtr;

typedef float PointType[3];
typedef long PatchType[4][4];

static ulong globalValidation1, globalValidation2;
static long result;
static PluginDataPtr theData;
static ControlDataPtr theControl;
static EIPluginRoutine hostProc;
static char* hostReference;
static HostStatusRec theStatusRec;
static ARGB theTeapotColor;

static void InitGlobals(boolean initFlag)
{
	if (initFlag || (globalValidation1 != pluginType) || (globalValidation2 != pluginModelType)) {
		globalValidation1 = pluginType;
		globalValidation2 = pluginModelType;
		InitUtils();
	}
}  /* InitGlobals */

static void FixControl(ControlDataPtr theControlRec)
{
	theControlRec->ownerType = pluginType;
	theControlRec->ownerVersion = pluginVersion;

	FixField(theControlRec->teapotScale);
	FixField(theControlRec->teapotFlags);
	FixField(theControlRec->teapotPoints);
}  /* FixControl */

static void DefaultControl(ControlDataRec *theControlRec)
{
	theControlRec->ownerType = pluginType;
	theControlRec->ownerVersion = pluginVersion;
	theControlRec->teapotScale = defTeapotScale;
	theControlRec->teapotFlags = defTeapotFlags;
	theControlRec->teapotColor.argb.a = defReferenceA;
	theControlRec->teapotColor.argb.r = defReferenceR;
	theControlRec->teapotColor.argb.g = defReferenceG;
	theControlRec->teapotColor.argb.b = defReferenceB;
	theControlRec->teapotPoints = defTeapotPoints;
}  /* DefaultControl */

static void DefaultData(PluginDataRec *theData)
{
	theData->vertexFlags = 0;
	InitMatrix4(&theData->theMatrix);
	InitMatrix4(&theData->theRotMatrix);
	InitMatrix4(&theData->theBlurMatrix);
}  /* DefaultData */

static long CheckData(PluginInfoPtr theInfo, boolean setControlData, boolean setPluginData)
{
	HostAllocateRec theAllocate;
	long newPluginDataSize, newControlDataSize;
	boolean newPluginData, newControlData;
	ControlDataRec theControlRec;
	PluginDataRec theDataRec;
	ControlDataPtr theControl;
	long theResult;

	theResult = 0;

	/* Change the size of the plugin data block or control data block if the size is too small.  Old version */
	/* control records are updated to the current version and reverse byte ordering is corrected. */

	theControl = (ControlDataPtr)theInfo->theControlData;
	if (!setControlData && (theInfo->theControlDataSize >= sizeof(ControlDataRec))
		&& ((theControl->ownerType == pluginType) || (theControl->ownerType == REVERSELONG(pluginType)))
		&& ((theControl->ownerVersion == pluginVersion) || (theControl->ownerVersion == REVERSELONG(pluginVersion)))) {
		if (theControl->ownerVersion == REVERSELONG(pluginVersion))
			FixControl(theControl);
		theControlRec = *theControl;
	} else {
		setControlData = 1;
		DefaultControl(&theControlRec);
	}

	if (setPluginData || (theInfo->thePluginDataSize < sizeof(PluginDataRec))) {
		setPluginData = 1;
		DefaultData(&theDataRec);
	} else
		theDataRec = *(PluginDataPtr)theInfo->thePluginData;

	newControlDataSize = sizeof(ControlDataRec);
	newControlData = (theInfo->theControlDataSize < newControlDataSize || theInfo->thePluginDataSize > FixAlignment(newControlDataSize));
	newPluginDataSize = sizeof(PluginDataRec);
	newPluginData = (theInfo->thePluginDataSize < newPluginDataSize || theInfo->thePluginDataSize > FixAlignment(newPluginDataSize));

	if (newPluginData || newControlData) {
		theAllocate.theReference = theInfo->theHostReference;
		theAllocate.theControlDataSize = newControlDataSize;
		theAllocate.theControlData = theInfo->theControlData;
		theAllocate.thePluginDataSize = newPluginDataSize;
		theAllocate.thePluginData = theInfo->thePluginData;
		theResult = theInfo->theHostProc(hostAllocate, sizeof(HostAllocateRec), &theAllocate);
		theInfo->theControlData = theAllocate.theControlData;
		theInfo->theControlDataSize = theAllocate.theControlDataSize;
		theInfo->thePluginData = theAllocate.thePluginData;
		theInfo->thePluginDataSize = theAllocate.thePluginDataSize;
	}

	if (theResult == 0) {
	
		/* Setup the control data to the default values if an incorrect sized block was passed to the plugin */
	
		if (newControlData || setControlData)
			*(ControlDataPtr)theInfo->theControlData = theControlRec;
	
		/* Setup the plugin private data */
	
		if (newPluginData || setPluginData)
			*(PluginDataPtr)theInfo->thePluginData = theDataRec;
	}
	return theResult;
}  /* CheckData */

/* Reset the host's model data */

static void SendModelReset(PluginInfoRec *theInfo)
{
	HostModelResetRec theModelReset;
	long theResult;

	theModelReset.theReference = theInfo->theHostReference;
	theResult = theInfo->theHostProc(hostModelReset, sizeof(HostModelResetRec), &theModelReset);
}  /* SendModelReset */

static boolean StatusTime(void)
{
	return ((theStatusRec.theStatusTick == 0) || (TICKCOUNT() >= theStatusRec.theStatusTick));
}  /* StatusTime */

static void UpdateStatus(float thePos, long theValue)
{
	if (result == 0) {
		theStatusRec.theStatusValue = theValue;
		theStatusRec.theStatusPos = thePos;
		result = hostProc(hostStatus, sizeof(HostStatusRec), &theStatusRec);
	}
}  /* UpdateStatus */

static long AddVertex(ARGB color, PointType normal, PointType position, PointType blurPosition, PointType texturePosition)
{
	HostModelAddVertexRec theAddVertex;

	if (result == 0) {
		theAddVertex.theReference = hostReference;
		theAddVertex.theColor = color;
		theAddVertex.theNormal.x = normal[0];
		theAddVertex.theNormal.y = normal[1];
		theAddVertex.theNormal.z = normal[2];
		TransformDCoordinate(&theAddVertex.theNormal, &theData->theRotMatrix);
		theAddVertex.thePosition.x = position[0];
		theAddVertex.thePosition.y = position[1];
		theAddVertex.thePosition.z = position[2];
		TransformDCoordinate(&theAddVertex.thePosition, &theData->theMatrix);
		theAddVertex.theBlurPosition.x = blurPosition[0];
		theAddVertex.theBlurPosition.y = blurPosition[1];
		theAddVertex.theBlurPosition.z = blurPosition[2];
		TransformDCoordinate(&theAddVertex.theBlurPosition, &theData->theBlurMatrix);
		theAddVertex.theTexturePosition.x = texturePosition[0];
		theAddVertex.theTexturePosition.y = texturePosition[1];
		theAddVertex.theTexturePosition.z = texturePosition[2];
		result = hostProc(hostModelAddVertex, sizeof(HostModelAddVertexRec), &theAddVertex);
	}
	if (result == 0)
		return (theAddVertex.theIndex);
	else
		return 0;
}  /* AddVertex */

static void AddFacet(ARGB color, short numVertices, long v0, long v1, long v2, long v3)
{
	HostModelAddFacetRec theAddFacet;

	if (result == 0) {
		theAddFacet.theReference = hostReference;
		theAddFacet.theColor = color;
		theAddFacet.theNumVertices = numVertices;
		theAddFacet.theVertices[0] = v0;
		theAddFacet.theVertices[1] = v1;
		theAddFacet.theVertices[2] = v2;
		theAddFacet.theVertices[3] = v3;
		result = hostProc(hostModelAddFacet, sizeof(HostModelAddFacetRec), &theAddFacet);
	}
}  /* AddFacet */

static void BlendVector(PointType d0, PointType d1, PointType d2, PointType d3, double t, PointType result)
{
	double t1 = 1.0 - t;
	double t2 = t1 * t1;
	double t3 = 3.0 * t;
	double a = t2 * t1;
	double b = t3 * t2;
	double c = t3 * t * t1;
	double d = t * t * t;
	
	result[0] = d0[0] * a + d1[0] * b + d2[0] * c + d3[0] * d;
	result[1] = d0[1] * a + d1[1] * b + d2[1] * c + d3[1] * d;
	result[2] = d0[2] * a + d1[2] * b + d2[2] * c + d3[2] * d;
}  /* BlendVector */

static void PickPatch(PointType *points, PatchType patch, double u, double v, PointType thePoint)
{
	PointType d0, d1, d2, d3;

	/* Splines for constant U */

	BlendVector(*(points + patch[0][0]), *(points + patch[0][1]), *(points + patch[0][2]), *(points + patch[0][3]), u, d0);
	BlendVector(*(points + patch[1][0]), *(points + patch[1][1]), *(points + patch[1][2]), *(points + patch[1][3]), u, d1);
	BlendVector(*(points + patch[2][0]), *(points + patch[2][1]), *(points + patch[2][2]), *(points + patch[2][3]), u, d2);
	BlendVector(*(points + patch[3][0]), *(points + patch[3][1]), *(points + patch[3][2]), *(points + patch[3][3]), u, d3);

	/* Point for constant V */

	BlendVector(d0, d1, d2, d3, v, thePoint);
}  /* PickPatch */

static boolean PickNormal(PointType p0, PointType p1, PointType p2, PointType theNormal)
{
	double a1 = p1[0] - p0[0];
	double b1 = p1[1] - p0[1];
	double c1 = p1[2] - p0[2];
	
	double a2 = p0[0] - p2[0];
	double b2 = p0[1] - p2[1];
	double c2 = p0[2] - p2[2];
	
	double aNum = b1 * c2 - b2 * c1;
	double bNum = c1 * a2 - c2 * a1;
	double cNum = a1 * b2 - a2 * b1;
	
	double distance = SQR(aNum) + SQR(bNum) + SQR(cNum);
	
	if (distance > 0.0) {
		distance = 1.0 / SQRT(distance);
		theNormal[0] = aNum * distance;
		theNormal[1] = bNum * distance;
		theNormal[2] = cNum * distance;
		return 1;
	} else
		return 0;
}  /* PickNormal */

static void PickNPatch(PointType *points, PatchType patch, double u, double v, long corner, PointType thePosition, PointType theNormal)
{
	PointType p1, p2, p3;
	
	PickPatch(points, patch, u, v, thePosition);
	switch (corner) {
		case 0:	/* Top left corner */
			PickPatch(points, patch, u, v + normalDelta, p1);
			PickPatch(points, patch, u + normalDelta, v, p2);
			break;
		case 1:	/* Top right corner */
			PickPatch(points, patch, u - normalDelta, v, p1);
			PickPatch(points, patch, u, v + normalDelta, p2);
			break;
		case 2:	/* Bottom left corner */
			PickPatch(points, patch, u + normalDelta, v, p1);
			PickPatch(points, patch, u, v - normalDelta, p2);
			break;
		case 3:	/* Bottom right corner */
			PickPatch(points, patch, u, v - normalDelta, p1);
			PickPatch(points, patch, u - normalDelta, v, p2);
			break;
	}
	
	if (!PickNormal(thePosition, p1, p2, theNormal)) {
		switch (corner) {
			case 0:	/* Top left corner */
				PickPatch(points, patch, u + normalDelta, v + normalDelta, p3);
				break;
			case 1:	/* Top right corner */
				PickPatch(points, patch, u - normalDelta, v + normalDelta, p3);
				break;
			case 2:	/* Bottom left corner */
				PickPatch(points, patch, u + normalDelta, v - normalDelta, p3);
				break;
			case 3:	/* Bottom right corner */
				PickPatch(points, patch, u - normalDelta, v - normalDelta, p3);
				break;
		}
		if (!PickNormal(p1, p3, p2, theNormal)) {
			theNormal[0] = 0.0;
			theNormal[1] = 0.0;
			theNormal[2] = 0.0;
		}
	}
}  /* PickNPatch */

static void BuildTeapot(void)
{
	#define pointCount	307			/* The number of control points */
	#define patchCount	32			/* The number of surface patches */
	
	PointType points[pointCount] = {
		{0.0, 0.0, 0.0},			/* Dummy point because the indices are one based */
		{1.4, 0.0, 2.4},
		{1.4, -0.784, 2.4},
		{0.784, -1.4, 2.4},
		{0.0, -1.4, 2.4},
		{1.3375, 0.0, 2.53125},
		{1.3375, -0.749, 2.53125},
		{0.749, -1.3375, 2.53125},
		{0.0, -1.3375, 2.53125},
		{1.4375, 0.0, 2.53125},
		{1.4375, -0.805, 2.53125},
		{0.805, -1.4375, 2.53125},
		{0.0, -1.4375, 2.53125},
		{1.5, 0.0, 2.4},
		{1.5, -0.84, 2.4},
		{0.84, -1.5, 2.4},
		{0.0, -1.5, 2.4},
		{-0.784, -1.4, 2.4},
		{-1.4, -0.784, 2.4},
		{-1.4, 0.0, 2.4},
		{-0.749, -1.3375, 2.53125},
		{-1.3375, -0.749, 2.53125},
		{-1.3375, 0.0, 2.53125},
		{-0.805, -1.4375, 2.53125},
		{-1.4375, -0.805, 2.53125},
		{-1.4375, 0.0, 2.53125},
		{-0.84, -1.5, 2.4},
		{-1.5, -0.84, 2.4},
		{-1.5, 0.0, 2.4},
		{-1.4, 0.784, 2.4},
		{-0.784, 1.4, 2.4},
		{0.0, 1.4, 2.4},
		{-1.3375, 0.749, 2.53125},
		{-0.749, 1.3375, 2.53125},
		{0.0, 1.3375, 2.53125},
		{-1.4375, 0.805, 2.53125},
		{-0.805, 1.4375, 2.53125},
		{0.0, 1.4375, 2.53125},
		{-1.5, 0.84, 2.4},
		{-0.84, 1.5, 2.4},
		{0.0, 1.5, 2.4},
		{0.784, 1.4, 2.4},
		{1.4, 0.784, 2.4},
		{0.749, 1.3375, 2.53125},
		{1.3375, 0.749, 2.53125},
		{0.805, 1.4375, 2.53125},
		{1.4375, 0.805, 2.53125},
		{0.84, 1.5, 2.4},
		{1.5, 0.84, 2.4},
		{1.75, 0.0, 1.875},
		{1.75, -0.98, 1.875},
		{0.98, -1.75, 1.875},
		{0.0, -1.75, 1.875},
		{2.0, 0.0, 1.35},
		{2.0, -1.12, 1.35},
		{1.12, -2.0, 1.35},
		{0.0, -2.0, 1.35},
		{2.0, 0.0, 0.9},
		{2.0, -1.12, 0.9},
		{1.12, -2.0, 0.9},
		{0.0, -2.0, 0.9},
		{-0.98, -1.75, 1.875},
		{-1.75, -0.98, 1.875},
		{-1.75, 0.0, 1.875},
		{-1.12, -2.0, 1.35},
		{-2.0, -1.12, 1.35},
		{ -2.0, 0.0, 1.35},
		{-1.12, -2.0, 0.9},
		{-2.0, -1.12, 0.9},
		{-2.0, 0.0, 0.9},
		{-1.75, 0.98, 1.875},
		{-0.98, 1.75, 1.875},
		{0.0, 1.75, 1.875},
		{-2.0, 1.12, 1.35},
		{-1.12, 2.0, 1.35},
		{0.0, 2.0, 1.35},
		{-2.0, 1.12, 0.9},
		{-1.12, 2.0, 0.9},
		{0.0, 2.0, 0.9},
		{0.98, 1.75, 1.875},
		{1.75, 0.98, 1.875},
		{1.12, 2.0, 1.35},
		{2.0, 1.12, 1.35},
		{1.12, 2.0, 0.9},
		{2.0, 1.12, 0.9},
		{2.0, 0.0, 0.45},
		{2.0, -1.12, 0.45},
		{1.12, -2.0, 0.45},
		{0.0, -2.0, 0.45},
		{1.5, 0.0, 0.225},
		{1.5, -0.84, 0.225},
		{0.84, -1.5, 0.225},
		{0.0, -1.5, 0.225},
		{1.5, 0.0, 0.15},
		{1.5, -0.84, 0.15},
		{0.84, -1.5, 0.15},
		{0.0, -1.5, 0.15},
		{-1.12, -2.0, 0.45},
		{-2.0, -1.12, 0.45},
		{-2.0, 0.0, 0.45},
		{-0.84, -1.5, 0.225},
		{-1.5, -0.84, 0.225},
		{-1.5, 0.0, 0.225},
		{-0.84, -1.5, 0.15},
		{-1.5, -0.84, 0.15},
		{-1.5, 0.0, 0.15},
		{-2.0, 1.12, 0.45},
		{-1.12, 2.0, 0.45},
		{0.0, 2.0, 0.45},
		{-1.5, 0.84, 0.225},
		{-0.84, 1.5, 0.225},
		{0.0, 1.5, 0.225},
		{-1.5, 0.84, 0.15},
		{-0.84, 1.5, 0.15},
		{0.0, 1.5, 0.15},
		{1.12, 2.0, 0.45},
		{2.0, 1.12, 0.45},
		{0.84, 1.5, 0.225},
		{1.5, 0.84, 0.225},
		{0.84, 1.5, 0.15},
		{1.5, 0.84, 0.15},
		{-1.6, 0.0, 2.025},
		{-1.6, -0.3, 2.025},
		{-1.5, -0.3, 2.25},
		{-1.5, 0.0, 2.25},
		{-2.3, 0.0, 2.025},
		{-2.3, -0.3, 2.025},
		{-2.5, -0.3, 2.25},
		{-2.5, 0.0, 2.25},
		{-2.7, 0.0, 2.025},
		{-2.7, -0.3, 2.025},
		{-3.0, -0.3, 2.25},
		{-3.0, 0.0, 2.25},
		{-2.7, 0.0, 1.8},
		{-2.7, -0.3, 1.8},
		{-3.0, -0.3, 1.8},
		{-3.0, 0.0, 1.8},
		{-1.5, 0.3, 2.25},
		{-1.6, 0.3, 2.025},
		{-2.5, 0.3, 2.25},
		{-2.3, 0.3, 2.025},
		{-3.0, 0.3, 2.25},
		{-2.7, 0.3, 2.025},
		{-3.0, 0.3, 1.8},
		{-2.7, 0.3, 1.8},
		{-2.7, 0.0, 1.575},
		{-2.7, -0.3, 1.575},
		{-3.0, -0.3, 1.35},
		{-3.0, 0.0, 1.35},
		{-2.5, 0.0, 1.125},
		{-2.5, -0.3, 1.125},
		{-2.65, -0.3, 0.9375},
		{-2.65, 0.0, 0.9375},
		{-2.0, -0.3, 0.9},
		{-1.9, -0.3, 0.6},
		{-1.9, 0.0, 0.6},
		{-3.0, 0.3, 1.35},
		{-2.7, 0.3, 1.575},
		{-2.65, 0.3, 0.9375},
		{-2.5, 0.3, 1.125},
		{-1.9, 0.3, 0.6},
		{-2.0, 0.3, 0.9},
		{1.7, 0.0, 1.425},
		{1.7, -0.66, 1.425},
		{1.7, -0.66, 0.6},
		{1.7, 0.0, 0.6},
		{2.6, 0.0, 1.42},
		{2.6, -0.66, 1.425},
		{3.1, -0.66, 0.825},
		{3.1, 0.0, 0.825},
		{2.3, 0.0, 2.1},
		{2.3, -0.25, 2.1},
		{2.4, -0.25, 2.025},
		{2.4, 0.0, 2.025},
		{2.7, 0.0, 2.4},
		{2.7, -0.25, 2.4},
		{3.3, -0.25, 2.4},
		{3.3, 0.0, 2.4},
		{1.7, 0.66, 0.6},
		{1.7, 0.66, 1.425},
		{3.1, 0.66, 0.825},
		{2.6, 0.66, 1.425},
		{2.4, 0.25, 2.025},
		{2.3, 0.25, 2.1},
		{3.3, 0.25, 2.4},
		{2.7, 0.25, 2.4},
		{2.8, 0.0, 2.475},
		{2.8, -0.25, 2.475},
		{3.525, -0.25, 2.49375},
		{3.525, 0.0, 2.49375},
		{2.9, 0.0, 2.475},
		{2.9, -0.15, 2.475},
		{3.45, -0.15, 2.5125},
		{3.45, 0.0, 2.5125},
		{2.8, 0.0, 2.4},
		{2.8, -0.15, 2.4},
		{3.2, -0.15, 2.4},
		{3.2, 0.0, 2.4},
		{3.525, 0.25, 2.49375},
		{2.8, 0.25, 2.475},
		{3.45, 0.15, 2.5125},
		{2.9, 0.15, 2.475},
		{3.2, 0.15, 2.4},
		{2.8, 0.15, 2.4},
		{0.0, 0.0, 3.15},
		{0.0, -0.002, 3.15},
		{0.002, 0.0, 3.15},
		{0.8, 0.0, 3.15},
		{0.8, -0.45, 3.15},
		{0.45, -0.8, 3.15},
		{0.0, -0.8, 3.15},
		{0.0, 0.0, 2.85},
		{0.2, 0.0, 2.7},
		{0.2, -0.112, 2.7},
		{0.112, -0.2, 2.7},
		{0.0, -0.2, 2.7},
		{-0.002, 0.0, 3.15},
		{-0.45, -0.8, 3.15},
		{-0.8, -0.45, 3.15},
		{-0.8, 0.0, 3.15},
		{-0.112, -0.2, 2.7},
		{-0.2, -0.112, 2.7},
		{-0.2, 0.0, 2.7},
		{0.0, 0.002, 3.15},
		{-0.8, 0.45, 3.15},
		{-0.45, 0.8, 3.15},
		{0.0, 0.8, 3.15},
		{-0.2, 0.112, 2.7},
		{-0.112, 0.2, 2.7},
		{0.0, 0.2, 2.7},
		{0.45, 0.8, 3.15},
		{0.8, 0.45, 3.15},
		{0.112, 0.2, 2.7},
		{0.2, 0.112, 2.7},
		{0.4, 0.0, 2.55},
		{0.4, -0.224, 2.55},
		{0.224, -0.4, 2.55},
		{0.0, -0.4, 2.55},
		{1.3, 0.0, 2.55},
		{1.3, -0.728, 2.55},
		{0.728, -1.3, 2.55},
		{0.0, -1.3, 2.55},
		{1.3, 0.0, 2.4},
		{1.3, -0.728, 2.4},
		{0.728, -1.3, 2.4},
		{0.0, -1.3, 2.4},
		{-0.224, -0.4, 2.55},
		{-0.4, -0.224, 2.55},
		{-0.4, 0.0, 2.55},
		{-0.728, -1.3, 2.55},
		{-1.3, -0.728, 2.55},
		{-1.3, 0.0, 2.55},
		{-0.728, -1.3, 2.4},
		{-1.3, -0.728, 2.4},
		{-1.3, 0.0, 2.4},
		{-0.4, 0.224, 2.55},
		{-0.224, 0.4, 2.55},
		{0.0, 0.4, 2.55},
		{-1.3, 0.728, 2.55},
		{-0.728, 1.3, 2.55},
		{0.0, 1.3, 2.55},
		{-1.3, 0.728, 2.4},
		{-0.728, 1.3, 2.4},
		{0.0, 1.3, 2.4},
		{0.224, 0.4, 2.55},
		{0.4, 0.224, 2.55},
		{0.728, 1.3, 2.55},
		{1.3, 0.728, 2.55},
		{0.728, 1.3, 2.4},
		{1.3, 0.728, 2.4},
		{0.0, 0.0, 0.0},
		{1.5, 0.0, 0.15},
		{1.5, 0.84, 0.15},
		{0.84, 1.5, 0.15},
		{0.0, 1.5, 0.15},
		{1.5, 0.0, 0.075},
		{1.5, 0.84, 0.075},
		{0.84, 1.5, 0.075},
		{0.0, 1.5, 0.075},
		{1.425, 0.0, 0.0},
		{1.425, 0.798, 0.0},
		{0.798, 1.425, 0.0},
		{0.0, 1.425, 0.0},
		{-0.84, 1.5, 0.15},
		{-1.5, 0.84, 0.15},
		{-1.5, 0.0, 0.15},
		{-0.84, 1.5, 0.075},
		{-1.5, 0.84, 0.075},
		{-1.5, 0.0, 0.075},
		{-0.798, 1.425, 0.0},
		{-1.425, 0.798, 0.0},
		{-1.425, 0.0, 0.0},
		{-1.5, -0.84, 0.15},
		{-0.84, -1.5, 0.15},
		{0.0, -1.5, 0.15},
		{-1.5, -0.84, 0.075},
		{-0.84, -1.5, 0.075},
		{0.0, -1.5, 0.075},
		{-1.425, -0.798, 0.0},
		{-0.798, -1.425, 0.0},
		{0.0, -1.425, 0.0},
		{0.84, -1.5, 0.15},
		{1.5, -0.84, 0.15},
		{0.84, -1.5, 0.075},
		{1.5, -0.84, 0.075},
		{0.798, -1.425, 0.0},
		{1.425, -0.798, 0.0},
	};
	
	PatchType patches[patchCount] = {
		{{1, 2, 3, 4},			{5, 6, 7, 8},			{9, 10, 11, 12},		{13, 14, 15, 16}},
		{{4, 17, 18, 19},		{8, 20, 21, 22},		{12, 23, 24, 25},		{16, 26, 27, 28}},
		{{19, 29, 30, 31},		{22, 32, 33, 34},		{25, 35, 36, 37},		{28, 38, 39, 40}},
		{{31, 41, 42, 1},		{34, 43, 44, 5},		{37, 45, 46, 9},		{40, 47, 48, 13}},
		{{13, 14, 15, 16},		{49, 50, 51, 52},		{53, 54, 55, 56},		{57, 58, 59, 60}},
		{{16, 26, 27, 28},		{52, 61, 62, 63},		{56, 64, 65, 66},		{60, 67, 68, 69}},
		{{28, 38, 39, 40},		{63, 70, 71, 72},		{66, 73, 74, 75},		{69, 76, 77, 78}},
		{{40, 47, 48, 13},		{72, 79, 80, 49},		{75, 81, 82, 53},		{78, 83, 84, 57}},
		{{57, 58, 59, 60},		{85, 86, 87, 88},		{89, 90, 91, 92},		{93, 94, 95, 96}},
		{{60, 67, 68, 69},		{88, 97, 98, 99},		{92, 100, 101, 102},	{96, 103, 104, 105}},
		{{69, 76, 77, 78},		{99, 106, 107, 108},	{102, 109, 110, 111},	{105, 112, 113, 114}},
		{{78, 83, 84, 57},		{108, 115, 116, 85},	{111, 117, 118, 89},	{114, 119, 120, 93}},
		{{121, 122, 123, 124},	{125, 126, 127, 128},	{129, 130, 131, 132},	{133, 134, 135, 136}},
		{{124, 137, 138, 121},	{128, 139, 140, 125},	{132, 141, 142, 129},	{136, 143, 144, 133}},
		{{133, 134, 135, 136},	{145, 146, 147, 148},	{149, 150, 151, 152},	{69, 153, 154, 155}},
		{{136, 143, 144, 133},	{148, 156, 157, 145},	{152, 158, 159, 149},	{155, 160, 161, 69}},
		{{162, 163, 164, 165},	{166, 167, 168, 169},	{170, 171, 172, 173},	{174, 175, 176, 177}},
		{{165, 178, 179, 162},	{169, 180, 181, 166},	{173, 182, 183, 170},	{177, 184, 185, 174}},
		{{174, 175, 176, 177},	{186, 187, 188, 189},	{190, 191, 192, 193},	{194, 195, 196, 197}},
		{{177, 184, 185, 174},	{189, 198, 199, 186},	{193, 200, 201, 190},	{197, 202, 203, 194}},
		{{204, 204, 204, 204},	{207, 208, 209, 210},	{211, 211, 211, 211},	{212, 213, 214, 215}},
		{{204, 204, 204, 204},	{210, 217, 218, 219},	{211, 211, 211, 211},	{215, 220, 221, 222}},
		{{204, 204, 204, 204},	{219, 224, 225, 226},	{211, 211, 211, 211},	{222, 227, 228, 229}},
		{{204, 204, 204, 204},	{226, 230, 231, 207},	{211, 211, 211, 211},	{229, 232, 233, 212}},
		{{212, 213, 214, 215},	{234, 235, 236, 237},	{238, 239, 240, 241},	{242, 243, 244, 245}},
		{{215, 220, 221, 222},	{237, 246, 247, 248},	{241, 249, 250, 251},	{245, 252, 253, 254}},
		{{222, 227, 228, 229},	{248, 255, 256, 257},	{251, 258, 259, 260},	{254, 261, 262, 263}},
		{{229, 232, 233, 212},	{257, 264, 265, 234},	{260, 266, 267, 238},	{263, 268, 269, 242}},
		{{270, 270, 270, 270},	{279, 280, 281, 282},	{275, 276, 277, 278},	{271, 272, 273, 274}},
		{{270, 270, 270, 270},	{282, 289, 290, 291},	{278, 286, 287, 288},	{274, 283, 284, 285}},
		{{270, 270, 270, 270},	{291, 298, 299, 300},	{288, 295, 296, 297},	{285, 292, 293, 294}},
		{{270, 270, 270, 270},	{300, 305, 306, 279},	{297, 303, 304, 275},	{294, 301, 302, 271}},
	};
	
	long uvPoints = theControl->teapotPoints;
	double uvStep = 1.0 / (uvPoints - 1);
	long patchTotal = uvPoints * uvPoints * patchCount;
	long patchIndex = 0;
	long count = 0;
	long i;
	
	/* Setup the status information */
	
	theStatusRec.theReference = hostReference;
	theStatusRec.theStatusPos = 0.0;
	PStrCpy(theStatusRec.theStatusUnit, PS("Point"));
	PStrCpy(theStatusRec.theStatusInfo, PS("Building Teapot"));
	theStatusRec.theStatusMin = 0;
	theStatusRec.theStatusMax = patchTotal;
	theStatusRec.theStatusValue = 0;
	theStatusRec.theStatusTick = 0;
	UpdateStatus(0.0, 0);

	/* Fix the patch data */
	
	for (i = 0; i < pointCount; i++) {
		
		/* Swqp the y and z coordinates */
		
		float temp = points[i][1];
		points[i][1] = points[i][2];
		points[i][2] = temp;
		
		/* Scale the patch */
		
		points[i][0] *= theControl->teapotScale;
		points[i][1] *= theControl->teapotScale;
		points[i][2] *= theControl->teapotScale;
	}

	/* Convert patch to facets */

	count = 0;
	patchIndex = 0;
	while ((patchIndex < patchCount) && (result == 0)) {
		PointType p00, p10, p01, p11;
		PointType n00, n10, n01, n11;
		boolean pinchU0, pinchV0, pinchU1, pinchV1;
		long vIndex = 0;
		double v = 0.0;
		
		PickNPatch(points, patches[patchIndex], 0.0, 0.0, 0, p00, n00);
		PickNPatch(points, patches[patchIndex], 1.0, 0.0, 1, p10, n10);
		PickNPatch(points, patches[patchIndex], 0.0, 1.0, 2, p01, n01);
		PickNPatch(points, patches[patchIndex], 1.0, 1.0, 3, p11, n11);
		
		pinchU0 = (p01[0] == p00[0]) && (p01[1] == p00[1]) && (p01[2] == p00[2]);
		pinchV0 = (p00[0] == p10[0]) && (p00[1] == p10[1]) && (p00[2] == p10[2]);
		pinchU1 = (p10[0] == p11[0]) && (p10[1] == p11[1]) && (p10[2] == p11[2]);
		pinchV1 = (p11[0] == p01[0]) && (p11[1] == p01[1]) && (p11[2] == p01[2]);
		
		/* Code needed here to fix the pinched corners of a triangular patch */
		
		while ((vIndex < uvPoints) && (result == 0)) {
			double u = 0.0;
			long uIndex = 0;
			long vCorner;
			
			vIndex++;
			if (vIndex == uvPoints) {
				v = 1.0;
				vCorner = 2;
			} else
				vCorner = 0;
			while ((uIndex < uvPoints) && (result == 0)) {
				PointType thePosition, theNormal, theBlurPosition, theTexturePosition;
				long theCorner, theIndex;
				
				uIndex++;
				if (uIndex == uvPoints) {
					u = 1.0;
					theCorner = vCorner + 1;
				} else
					theCorner = vCorner;
				count++;
				PickNPatch(points, patches[patchIndex], u, v, theCorner, thePosition, theNormal);
				
				theBlurPosition[0] = thePosition[0];
				theBlurPosition[1] = thePosition[1];
				theBlurPosition[2] = thePosition[2];
				if (BTST(theControl->teapotFlags, texturePatchFlag)) {
					theTexturePosition[0] = u;
					theTexturePosition[1] = 1.0 - v;
					theTexturePosition[2] = 0.0;
				} else {
					theTexturePosition[0] = thePosition[0];
					theTexturePosition[1] = thePosition[1];
					theTexturePosition[2] = thePosition[2];
				}
				theIndex = AddVertex(theControl->teapotColor, theNormal, thePosition, theBlurPosition, theTexturePosition);
				if ((uIndex > 1) && (vIndex > 1))
					AddFacet(theControl->teapotColor, 4, theIndex - uvPoints - 1, theIndex - uvPoints, theIndex, theIndex - 1);
				u += uvStep;
				if (StatusTime())
					UpdateStatus((double)count / patchTotal, count);
			}
			v += uvStep;
		}
		patchIndex++;
	}
	UpdateStatus(1.0, patchTotal);
}  /* BuildTeapot */

static void BuildPatch(void)
{	
	long uvPoints = theControl->teapotPoints;
	double uvStep = 1.0 / (uvPoints - 1);
	long patchTotal = uvPoints * uvPoints;
	long count = 0;
	long vIndex = 0;
	double v = 0.0;

	theStatusRec.theReference = hostReference;
	theStatusRec.theStatusPos = 0.0;
	PStrCpy(theStatusRec.theStatusUnit, PS("Point"));
	PStrCpy(theStatusRec.theStatusInfo, PS("Building Patch"));
	theStatusRec.theStatusMin = 0;
	theStatusRec.theStatusMax = patchTotal;
	theStatusRec.theStatusValue = 0;
	theStatusRec.theStatusTick = 0;
	UpdateStatus(0.0, 0);

	/* Convert patch to facets */

	count = 0;
	while ((vIndex < uvPoints) && (result == 0)) {
		double u = 0.0;
		long uIndex = 0;
		
		vIndex++;
		if (vIndex == uvPoints)
			v = 1.0;
		while ((uIndex < uvPoints) && (result == 0)) {
			PointType thePosition, theNormal, theBlurPosition, theTexturePosition;
			long theIndex;
			
			uIndex++;
			if (uIndex == uvPoints)
				u = 1.0;
			count++;
			thePosition[0] = u;
			thePosition[1] = 1.0 - v;
			thePosition[2] = 0.0;
			theNormal[0] = 0.0;
			theNormal[1] = 0.0;
			theNormal[2] = 1.0;
			theBlurPosition[0] = thePosition[0];
			theBlurPosition[1] = thePosition[1];
			theBlurPosition[2] = thePosition[2];
			theTexturePosition[0] = thePosition[0];
			theTexturePosition[1] = thePosition[1];
			theTexturePosition[2] = thePosition[2];
			theIndex = AddVertex(theControl->teapotColor, theNormal, thePosition, theBlurPosition, theTexturePosition);
			if ((uIndex > 1) && (vIndex > 1))
				AddFacet(theControl->teapotColor, 4, theIndex - uvPoints - 1, theIndex - uvPoints, theIndex, theIndex - 1);
			u += uvStep;
			if (StatusTime())
				UpdateStatus((double)count / patchTotal, count);
		}
		v += uvStep;
	}
	UpdateStatus(1.0, patchTotal);
}  /* BuildPatch */

/* Initialize the plugin's data structures and perform memory allocation */

static long DoInitialize(long theCommandRecSize, void* theCommandRec)
{
	PluginInitializePtr theParameters;
	long theResult;

	InitGlobals(1);
	theParameters = (PluginInitializePtr)theCommandRec;
	if ((theCommandRecSize < sizeof(PluginInitializeRec)) || (theParameters->theInfo.theHostProc == NULL))
		theResult = pluginParameterErr;
	else
		theResult = CheckData(&theParameters->theInfo, 0, 1);
	return theResult;
}  /* DoInitialize */

/* Finish up anything which needs it., This is the last time that the plugin will be called */

static long DoFinish(long theCommandRecSize, void* theCommandRec)
{
	PluginFinishPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginFinishPtr)theCommandRec;
	if ((theCommandRecSize < sizeof(PluginFinishRec)) || (theParameters->theInfo.theHostProc == NULL))
		theResult = pluginParameterErr;
	FinishUtils();
	return theResult;
}  /* DoFinish */

/* Get information about the plugin */

static long DoInformation(long theCommandRecSize, void* theCommandRec)
{
	PluginInformationPtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginInformationPtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginInformationRec))
		theResult = pluginParameterErr;
	else {
		theParameters->thePluginType = pluginType;
		theParameters->thePluginVersion = pluginVersion;
		PStrCpy(theParameters->thePluginName, pluginName);
		PStrCpy(theParameters->thePluginCopyright, pluginCopyright);
		PStrCpy(theParameters->thePluginAuthor, pluginAuthor);
		PStrCpy(theParameters->thePluginDescription, pluginDescription);
		theParameters->thePluginPermission = pluginPermission;
	}
	return theResult;
}  /* DoInformation */

/* Check to see if a particular command is supported */

static long DoCheck(long theCommandRecSize, void* theCommandRec)
{
	PluginCheckPtr theParameters;
	
	theParameters = (PluginCheckPtr)theCommandRec;
	if (theCommandRecSize >= sizeof(PluginCheckRec))
		switch (theParameters->theCommand) {
			case pluginInitialize:
			case pluginFinish:
			case pluginInformation:
			case pluginCheck:
			case pluginDescribe:
			case pluginInterface:
			case pluginModelSetup:
			case pluginModelGenerate:
				return pluginNoErr;
		}
	return pluginParameterErr;
}  /* DoCheck */

/* Describe an error number */

static long DoDescribe(long theCommandRecSize, void* theCommandRec)
{
	PluginDescribePtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginDescribePtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginDescribeRec))
		theResult = pluginParameterErr;
	else
		DescribePluginError(theParameters->theResult, theParameters->theDescription);
	return theResult;
}  /* DoDescribe */

static void RadioHitFunction(EI_Control *iControl, EI_DrawContext *iContext)

{
	iContext;
	
	if (EI_GetControlValue(iControl))
		SetButton(EI_GetControlDialog(iControl), cTextureByPatchCheck, 1);
}

/* Display the plugin's user interface and edit parameters */

static long DoInterface(long theCommandRecSize, void* theCommandRec)
{
	EI_ResourceFile		*resourceFile;
	EI_Dialog			*theDialog;
	long temp;
	boolean changeFlag;
	ControlDataPtr theControl;
	PluginInterfacePtr theParameters;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginInterfacePtr)theCommandRec;
	if ((theCommandRecSize < sizeof(PluginInterfaceRec)) || (theParameters->theInfo.theControlData == NULL) || (theParameters->theInfo.theControlDataSize < sizeof(ControlDataRec)))
		theResult = pluginParameterErr;
	else
		theResult = CheckData(&theParameters->theInfo, 0, 0);
	if (theResult == 0) {
		resourceFile = OpenResourceFile("Teapot.rsc");
		if (resourceFile == NULL)
			theResult = pluginResourceErr;
	}
	if (theResult == 0) {
		theControl = (ControlDataPtr)theParameters->theInfo.theControlData;
		theDialog = EI_MakeDialog(resourceFile, interfaceDialogID);
		if (theDialog != NULL) {
			SetTheINum(theDialog, cPatchPointsET, theControl->teapotPoints);
			SetTheIReal(theDialog, cPatchScaleET, theControl->teapotScale, 2);
			SetButton(theDialog, cTextureByPatchCheck, BTST(theControl->teapotFlags, texturePatchFlag));
			SetButton(theDialog, cGenerateRB, BTST(theControl->teapotFlags, generatePatchFlag));
			SetColorButtonIDARGB(theDialog, cPatchColor, &theControl->teapotColor);
			SetControlIDHitFunction(theDialog, cGenerateRB, RadioHitFunction);
			if (EI_ExecuteDialog(theDialog)) {
				ARGB		color;
				
				changeFlag = 0;
				temp = GetTheINum(theDialog, cPatchPointsET);
				if (temp < 2)
					temp = 2;
				SetLongInt(&theControl->teapotPoints, temp, &changeFlag);
				SetDouble(&theControl->teapotScale, GetTheIReal(theDialog, cPatchScaleET), &changeFlag);
				GetColorButtonIDARGB(theDialog, cPatchColor, &color);
				if (color.colorValue != theControl->teapotColor.colorValue) {
					theControl->teapotColor = color;
					changeFlag = 1;
				}
				SetBooleanFlag(&theControl->teapotFlags, texturePatchFlag, GetButton(theDialog, cTextureByPatchCheck), &changeFlag);
				SetBooleanFlag(&theControl->teapotFlags, generatePatchFlag, GetButton(theDialog, cGenerateRB), &changeFlag);
				if (changeFlag)
					SendModelReset(&theParameters->theInfo);
			} else
				theResult = pluginCancelErr;
			
/*
			do {
				ModalDialog(theStandardFilterProc, &item);
				switch (item) {
					case colorItem:
						if (TrackColorSelector(theDialog, item))
							SetTeapotColor(theDialog, item);
						break;
					case texturePatchItem:
						ToggleButton(theDialog, texturePatchItem);
						break;
					case generateTeapotItem:
						SetButton(theDialog, generateTeapotItem, 1);
						SetButton(theDialog, generatePatchItem, 0);
						break;
					case generatePatchItem:
						SetButton(theDialog, generateTeapotItem, 0);
						SetButton(theDialog, generatePatchItem, 1);
						SetButton(theDialog, texturePatchItem, 1);	/* Shortcut * /
						break;
					default:
						break;
				}
			} while ((item != okItem) && (item != cancelItem));
			if (item == cancelItem)
				theResult = pluginCancelErr;
			else {
				changeFlag = 0;
				temp = GetTheINum(theDialog, patchPointsItem);
				if (temp < 2)
					temp = 2;
				SetLongInt(&theControl->teapotPoints, temp, &changeFlag);
				SetDouble(&theControl->teapotScale, GetTheIReal(theDialog, patchScaleItem), &changeFlag);
				SetLongInt((long*)&theControl->teapotColor.colorValue, theTeapotColor.colorValue, &changeFlag);
				SetBooleanFlag(&theControl->teapotFlags, texturePatchFlag, GetButton(theDialog, texturePatchItem), &changeFlag);
				SetBooleanFlag(&theControl->teapotFlags, generatePatchFlag, GetButton(theDialog, generatePatchItem), &changeFlag);
				if (changeFlag)
					SendModelReset(&theParameters->theInfo);
			}
			DisposeRoutineDescriptor(theStandardFilterProc);
			DisposeRoutineDescriptor(theColorSelectorProc);
			DisposeDialog(theDialog);
*/
			EI_DestroyDialog(theDialog);
		}
		EI_CloseResourceFile(resourceFile);
	}
	return theResult;
}  /* DoInterface */

/* Setup the model generation variables */

static long DoSetup(long theCommandRecSize, void* theCommandRec)
{
	PluginModelSetupPtr theParameters;
	PluginInfoPtr theInfo;
	ControlDataPtr theControl;
	PluginDataPtr theData;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelSetupPtr)theCommandRec;
	theInfo = &theParameters->theInfo;
	if ((theCommandRecSize < sizeof(PluginModelSetupRec)) || (theInfo->theHostProc == NULL))
		theResult = pluginParameterErr;
	else
		theResult = CheckData(&theParameters->theInfo, 0, 0);
	if (theResult == 0) {
		theControl = (ControlDataPtr)theInfo->theControlData;
		theData = (PluginDataPtr)theInfo->thePluginData;
		theData->theMatrix = theParameters->theGroupMatrix;
		theData->theRotMatrix = theParameters->theGroupRotMatrix;
		theData->theBlurMatrix = theParameters->theGroupBlurMatrix;
		SendModelReset(&theParameters->theInfo);
	}
	return theResult;
}  /* DoSetup */

/* Generate the coordinates and polygons for the group */

static long DoGenerate(long theCommandRecSize, void* theCommandRec)
{
	PluginModelGeneratePtr theParameters;
	PluginInfoPtr theInfo;
	long theResult;

	theResult = pluginNoErr;
	theParameters = (PluginModelGeneratePtr)theCommandRec;
	if (theCommandRecSize < sizeof(PluginModelGenerateRec))
		theResult = pluginParameterErr;
	else {
		theInfo = &theParameters->theInfo;
		theResult = CheckData(&theParameters->theInfo, 0, 0);
		if (theResult == 0) {
			theControl = (ControlDataRec *)theInfo->theControlData;
			theData = (PluginDataRec *)theInfo->thePluginData;
			hostProc = theInfo->theHostProc;
			hostReference = theInfo->theHostReference;
			theData->vertexFlags = theParameters->theVertexFlags;
		
			/* Build the teapot model */
		
			result = theResult;
			if (BTST(theControl->teapotFlags, generatePatchFlag))
				BuildPatch();
			else
				BuildTeapot();
			theResult = result;
		}
	}
	return theResult;
}  /* DoGenerate */

/* The main function is called by the host application for all plugin commands */

EXPDLL long Main(unsigned long theCommand, long theCommandRecSize, void* theCommandRec)
{
	long theResult;
	
	/* Initialize the plugin's global variables if they have not been initialize previously */

	InitGlobals(0);

	/* Perform the requested command */

	switch (theCommand) {
		case pluginInitialize:		theResult = DoInitialize(theCommandRecSize, theCommandRec);		break;
		case pluginFinish:			theResult = DoFinish(theCommandRecSize, theCommandRec);			break;
		case pluginInformation:		theResult = DoInformation(theCommandRecSize, theCommandRec);	break;
		case pluginCheck:			theResult = DoCheck(theCommandRecSize, theCommandRec);			break;
		case pluginDescribe:		theResult = DoDescribe(theCommandRecSize, theCommandRec);		break;
		case pluginInterface:		theResult = DoInterface(theCommandRecSize, theCommandRec);		break;
		case pluginModelSetup:		theResult = DoSetup(theCommandRecSize, theCommandRec);			break;
		case pluginModelGenerate:	theResult = DoGenerate(theCommandRecSize, theCommandRec);		break;
		default:					theResult = pluginCommandErr;									break;
	}
	return theResult;
}  /* Main */
